// ----------------------------------------------------------------------------
// Source.java
// Handling of source files
// $Id: Source.java,v 1.13 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import java.io.*;
import java.util.*;
import fmrisc.ProgramExplorer.*;

/**
 * 
 * Handling of source files.
 *
 */
public final class Source
{
  private File file;         // the file holding the text
  private long modified;     // the time stamp of the last modification
  private String text;       // null or the text of the file
  private int[] lines;       // null or the start positions of the line
  private ErrorStream error; // the stream for printing error messages
  
  // the set of positions attached to this source
  private Vector<SourcePosition> positions = new Vector<SourcePosition>();
  
  // the line separator for the current system
  public static final String NL = System.getProperty("line.separator");
  
  // all sources available in the system
  private static MapClass<String, Source> sources = new MapClass<String, Source>();
  
  /***************************************************************************
   * Construct a source file.
   * @param file the file denoting the source.
   * @param error the stream for printing error messages.
   * @return the source file.
   **************************************************************************/
  public static Source create(File file, ErrorStream error)
  {
    // take care to follow symbolic links
    String path = null;
    try
    {
      path = file.toPath().toRealPath().toString();
    }
    catch(IOException e)
    {
      File parent = file.getParentFile();
      try
      {
        File file0 = new File(parent.toPath().toRealPath().toString(), file.getName());
        path = file0.getAbsolutePath();
      }
      catch(IOException e0)
      {
        Main.internalError("cannot access path: " + e0.getMessage());
      }
    }
    Source source = sources.get(path);
    if (source != null) return source;
    source = new Source(file, error);
    sources.put(path, source);
    return source;
  }
  
  /***************************************************************************
   * Construct a source file.
   * @param file the file denoting the source.
   * @param error the stream for printing error messages.
   **************************************************************************/
  private Source(File file, ErrorStream error)
  {
    this.file = file;
    this.text = null;
    this.lines = null;
    this.error = error;
  }
  
  /***************************************************************************
   * Determine whether file is a source file.
   * @param file the file.
   * @return true iff file is a source file.
   **************************************************************************/
  public static boolean isSource(File file)
  {
    return isClassSource(file) || isTheorySource(file);
  }
  
  /***************************************************************************
   * Determine whether file is a source file for a class.
   * @param file the file.
   * @return true iff file is a source file for a class.
   **************************************************************************/
  public static boolean isClassSource(File file)
  {
    return file.isFile() && file.getName().endsWith(".java");
  }

  /***************************************************************************
   * Determine whether file is a source file for a class.
   * @param file the file.
   * @return true iff file is a source file for a class.
   **************************************************************************/
  public static boolean isTheorySource(File file)
  {
    return file.isFile() && file.getName().endsWith(".theory");
  }
  
  /***************************************************************************
   * Get the file denoting the source.
   * @return the file.
   **************************************************************************/
  public File getFile()
  {
    return file;
  }
  
  /****************************************************************************
   * Return true if text has been modified and should be retrieved.
   * @return true if trext should be retrieved by getText().
   ***************************************************************************/
  public boolean isModified()
  {
    return modified != file.lastModified();
  }
  
  /****************************************************************************
   * Get the text of the source.
   * @return the text
   ***************************************************************************/
  public String getText()
  {
    if (text != null && !isModified()) 
      return text;
    try
    {
      StringBuffer buffer = new StringBuffer();
      BufferedReader reader = new BufferedReader(new FileReader(file));
      int pos = 0;
      Vector<Integer> lines0 = new Vector<Integer>();
      while (true)
      {
        String line = reader.readLine();
        if (line == null) break;
        buffer.append(line);
        buffer.append(NL);
        lines0.add(pos);
        pos = pos+line.length()+1;
      }
      text = buffer.toString();
      int n = lines0.size();
      lines = new int[n];
      for (int i=0; i<n; i++)
        lines[i] = lines0.elementAt(i);
      modified = file.lastModified();
      reader.close();
      return text;
    }
    catch(IOException e)
    {
      error.newError("error in reading file " + file.getAbsolutePath() + ":" +
          e.getMessage());
      return null;
    }
  }
  
  /***************************************************************************
   * Get size of source.
   * @return the number of characters in the source.
   **************************************************************************/
  public int getSize()
  {
    return text.length();
  }
  
  /***************************************************************************
   * Get number of lines of source.
   * @return the number of lines in the source.
   **************************************************************************/
  public int getLines()
  {
    return lines.length;
  }
  
  /***************************************************************************
   * Get a line from the source.
   * @param i the index of a line in the source.
   * @return the content of the line.
   **************************************************************************/
  public String getLine(int i)
  {
    int from = lines[i];
    int to = text.indexOf(NL, from);
    if (to == -1) to = text.length();
    return text.substring(from, to);
  }
  
  /***************************************************************************
   * Construct position in this source.
   * @param line the line number.
   * @param column the column number.
   * @return the position.
   **************************************************************************/
  public SourcePosition newPosition(int line, int column)
  {
    // the parser generates position (-1,-1) on EOF
    if (line < 0)
    {
      int last = text.length()-1;
      for (int i=last; i>=0; i--)
      {
        if (!Character.isWhitespace(text.charAt(i)))
        {
          for (int j=lines.length-1; j>=0; j--)
          {
            if (lines[j] <= i)
            {
              // apparently (?) annotation cannot be generated in last line
              if (j == lines.length-1 && j > 0)
                j = j-1;
              return new SourcePositionClass(this, j, 0);
            }
          }
        }
      }
      return new SourcePositionClass(this, 0, 0);
    }
    else
      return new SourcePositionClass(this, line, column);
  }
  
  /****************************************************************************
   * Register position for this source.
   * @param pos a position for this source.
   ***************************************************************************/
  public void register(SourcePosition pos)
  {
    if (pos.getSource() != this)
      Main.internalError("invalid position registration");
    positions.add(pos);
  }
  
  /***************************************************************************
   * Get positions for this source.
   * @return the positions created for this source.
   **************************************************************************/
  public SourcePosition[] getPositions()
  {
    int n = positions.size();
    SourcePosition[] p = new SourcePosition[n];
    positions.copyInto(p);
    return p;
  }
  
  /****************************************************************************
   * Clear positions for this source.
   ***************************************************************************/
  public void clearPositions()
  {
    positions = new Vector<SourcePosition>();
  }
  
  /****************************************************************************
   * Get offset of denoted position
   * @param line line index 
   * @param column column index
   * @return the number of characters up to (excluding) the position denoted
   * by the line and column index.
   ***************************************************************************/
  public int getOffset(int line, int column)
  {
    return lines[line]+column;
  }
  
  /**************************************************************************
   * Get number of characters from current position to end of line.
   * @param line line index 
   * @param column column index
   * @return the number of characters.
   *************************************************************************/
  public int getLength(int line, int column)
  {
    if (line+1 < lines.length)
      return lines[line+1]-getOffset(line, column);
    else
      return getSize()-getOffset(line, column);
  }
}
