// ----------------------------------------------------------------------------
// SourceAnnotation.java
// An annotation of the source code.
// $Id: SourceAnnotation.java,v 1.3 2010/03/18 17:15:59 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import java.io.*;
import java.util.*;

import fmrisc.ProgramExplorer.Main;

/**
 * 
 * An annotation of the source code.
 *
 */
public class SourceAnnotation
{
  private SourcePosition pos; // the source position of the annotation
  private String text;        // the text of the annotation
  private int[] lines;        // the start positions of lines in the annotation
  
  /***************************************************************************
   * Construct a source annotation.
   * @param pos the position of the annotation in a source.
   * @param text the text of the annotation.
   *************************************************************************/
  public SourceAnnotation(SourcePosition pos, String text)
  {
    this.pos = pos;
    this.text = text;
    parse();
  }
  
  /***************************************************************************
   * Decompose the annotation into lines.
   **************************************************************************/
  private void parse()
  {
    try
    {
      BufferedReader reader = new BufferedReader(new StringReader(text));
      int pos = 0;
      Vector<Integer> lines0 = new Vector<Integer>();
      while (true)
      {
        String line = reader.readLine();
        if (line == null) break;
        lines0.add(pos);
        pos = pos+line.length()+1;
      }
      int n = lines0.size();
      lines = new int[n];
      for (int i=0; i<n; i++)
        lines[i] = lines0.elementAt(i);
    }
    catch(IOException e)
    {
      Main.internalError("Error in parsing annotation: " + e);
    }
  }
  
  /***************************************************************************
   * Get the source position of the annotation.
   * @return the source position.
   **************************************************************************/
  public SourcePosition getPosition()
  {
    return pos;
  }
  
  /***************************************************************************
   * Get the number of lines of the annotation.
   * @return the number of lines.
   **************************************************************************/
  public int getLines()
  {
    return lines.length;
  }
  
  /***************************************************************************
   * Get a line from the annotation.
   * @param i the index of a line in the annotation.
   * @return the content of the annotation.
   **************************************************************************/
  public String getLine(int i)
  {
    int from = lines[i];
    int to = text.indexOf(Source.NL, from);
    if (to == -1) to = text.length();
    return text.substring(from, to);
  }
  
  
  /***************************************************************************
   * Construct position in this source.
   * @param line the line number.
   * @param column the column number.
   * @return the position.
   **************************************************************************/
  public SourcePosition newPosition(int line, int column)
  {
    // the parser generates position (-1,-1) on EOF
    if (line >= 0) return new SourceAnnotationPosition(this, line, column);
    int last = getLines()-1;
    for (int i=last; i>=0; i--)
    {
      String text = getLine(i);
      int n = text.length();
      for (int j=0; j<n; j++)
      {
        if (!Character.isWhitespace(text.charAt(j)))
        {
          // apparently (?) annotation cannot be generated in last line
          if (i == last && i > 0) i = i-1;
          return new SourceAnnotationPosition(this, i, 0);
        }
      }
    }
    return new SourceAnnotationPosition(this, 0, 0);    
  }
}
