// ----------------------------------------------------------------------------
// ExpSubstitution.java
// The expression substitution P[E1/E2]
// $Id: ExpSubstitution.java,v 1.4 2011/09/30 15:47:38 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.ewuni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import java.util.*;

import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * The expression substitution P[E1/E2]
 *
 */
public final class ExpSubstitution extends ASTCloner
{
  
  private final Expression oldExp; // the old expression
  private final Expression newExp; // the new expression
  private final String oldString;  // the string representation of oldExp
  private final Collection<String> vars; // the free variables in old/newExp
  
  /***************************************************************************
   * Create a copy of expression with the denoted substitution performed.
   * @param exp the expression to be copied.
   * @param oldExp the subexpression to be substituted in exp.
   * @param newExp the replacement for oldExp.
   * @return a copy of exp with the corresponding substitution performed
   * (only clones of the substitution expression are integrated into the copy)
   **************************************************************************/
  public static Expression perform(Expression exp,
    Expression oldExp, Expression newExp)
  {
    ExpSubstitution subst = new ExpSubstitution(oldExp, newExp);
    return (Expression)subst.clone0(exp);
  }   
  
  /***************************************************************************
   * Create a substitution (parameters as in "perform").
   **************************************************************************/
  private ExpSubstitution(Expression oldExp, Expression newExp)
  {
    this.oldExp = oldExp;
    this.newExp = newExp;
    this.oldString = oldExp.toString();
    vars = FreeVariables.compute(oldExp);
    FreeVariables.add(newExp, vars);
  }
 
  /****************************************************************************
   * Substitute expression.
   * If substitution is successful, the result is set to a non-null value.
   * @param tree the (potential) expression to be substituted.
   ***************************************************************************/
  protected void setResult(AST tree) 
  { 
    result = null;
    if (!(tree instanceof Expression)) return;
    Expression exp = (Expression)tree;
    if (exp instanceof QuantifiedExpression) 
    {
      QuantifiedExpression exp0 = (QuantifiedExpression)exp;
      TypedIdentifier[] tids = exp0.getVariables();
      for (TypedIdentifier tid : tids)
      {
        if (vars.contains(tid.getName().getString()))
        {
          // prevent name capture
          result = ASTCloner.clone(exp);
          return;
        }
      }
    }
    if (exp.getClass() != oldExp.getClass()) return;
    String string = exp.toString();
    if (!string.equals(oldString)) return;
    result = ASTCloner.clone(newExp);
  }
}
