// ----------------------------------------------------------------------------
// Substitution.java
// The program variable substitution P[U1/VAR x,...]
// $Id: Substitution.java,v 1.8 2011/09/30 15:47:38 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.ewuni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Semantics.*;

import java.util.*;

/**
 * 
 * The program variable substitution P[U1/VAR x,...]
 *
 */
public final class Substitution extends ASTCloner
{
  
  private final Term oldStateTerm; // the substitution for "now" (may be null)
  private final Term newStateTerm; // the substitution for "next" (may be null)
  
  // the action to be performed on a pre/post-state reference:
  // none: leave as it is
  // flip: switch from pre-state to post-state and vice versa
  // term: perform the substitution by the given term
  public static enum Action { none, flip, term }
  
  /****************************************************************************
   * The substitution for a reference to a program variable.
   ***************************************************************************/
  public final static class Value
  {
    // the pre-state variable substitution
    public final Action preAction;
    public final Term preTerm;  
    
    // the post-state variable substitution
    public final Action postAction;
    public final Term postTerm; 
    
    /*************************************************************************
     * Describe the substitution result of a program variable.
     * @param preAction the action on the pre-state value
     * @param preTerm if preAction=term, then the replacement term
     * @param postAction the action on the post-state value
     * @param postTerm if postAction=term, then the replacement term
     ************************************************************************/
    public Value(Action preAction, Term preTerm, 
      Action postAction, Term postTerm)
    {
      this.preAction = preAction;
      this.preTerm = preTerm;
      this.postAction = postAction;
      this.postTerm = postTerm;
    }
  }

  // a variable symbol is mapped to a 
  // pair of pre/post-state reference replacements
  // each element of the pair may be null (then no substitution is performed)
  private final Map<VariableSymbol,Value> vars;
  
  /***************************************************************************
   * Create a copy of expression with the denoted substitutions performed.
   * @param exp the expression to be copied.
   * @param oldStateTerm the substitution for "now" (if null, no substitution)
   * @param newStateTerm the substitution for "next (if null, no substitution)
   * @param vars the substitutions for (references to) program variables
   * @return a copy of exp with the corresponding substitutions performed
   * (only clones of the substitution terms are integrated into the copy)
   **************************************************************************/
  public static Expression perform(Expression exp,
    Term oldStateTerm, Term newStateTerm, Map<VariableSymbol,Value> vars)
  {
    Substitution subst = new Substitution(oldStateTerm, newStateTerm, vars);
    return (Expression)subst.clone0(exp);
  }   
  
  /***************************************************************************
   * Create a substitution (parameters as in "perform").
   **************************************************************************/
  private Substitution(Term oldStateTerm, Term newStateTerm,
    Map<VariableSymbol,Value> vars)
  {
    this.oldStateTerm = oldStateTerm;
    this.newStateTerm = newStateTerm;
    this.vars = vars;
  }
  
  /****************************************************************************
   * Substitute expression.
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(OldState exp) 
  { 
    if (oldStateTerm == null) 
    {
      super.visit(exp);
      return;
    }
    result = ASTCloner.clone(oldStateTerm);
  }
  
  /****************************************************************************
   * Substitute expression.
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(NewState exp) 
  { 
    if (newStateTerm == null) 
    {
      super.visit(exp);
      return;
    }
    result = ASTCloner.clone(newStateTerm);
  }
  
  /****************************************************************************
   * Substitute expression.
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(PostVariable exp) 
  { 
    VariableSymbol symbol = (VariableSymbol)exp.getName().getSymbol();
    Value value = vars.get(symbol);
    if (value == null) 
    {
      super.visit(exp);
      return;
    }
    if (value.postAction == Action.none) 
    {
      super.visit(exp);
      return;
    }
    if (value.postAction == Action.flip) 
    {
      result = PreVariable.construct((Name)clone0(exp.getName()));
      ((Term)result).setType(exp.getType());
    }
    else if (value.postAction == Action.term) 
      result = ASTCloner.clone(value.postTerm);
    else
      Main.internalError("unknown kind of substitution action");
  }
  
  /****************************************************************************
   * Substitute expression.
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(PreVariable exp) 
  { 
    VariableSymbol symbol = (VariableSymbol)exp.getName().getSymbol();
    Value value = vars.get(symbol);
    if (value == null) 
    {
      super.visit(exp);
      return;
    }
    if (value.preAction == Action.none) 
    {
      super.visit(exp);
      return;
    }
    if (value.preAction == Action.flip) 
    {
      result = PostVariable.construct((Name)clone0(exp.getName()));
      ((Term)result).setType(exp.getType());
    }
    else if (value.preAction == Action.term) 
      result = ASTCloner.clone(value.preTerm);
    else
      Main.internalError("unknown kind of substitution action");
  }
}
