// ----------------------------------------------------------------------------
// ProofNavigatorProblem.java
// A proving problem for the RISC ProofNavigator.
// $Id: ProofNavigatorProblem.java,v 1.25 2012/10/16 15:49:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.ewuni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Proving;

import java.io.*;
import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Tasks.Task;

import fmrisc.ProofNavigator.State;
import fmrisc.ProofNavigator.Syntax.*;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Commands.*;
import fmrisc.ProofNavigator.Communication.Store;
import fmrisc.ProofNavigator.Semantics.*;

/**
 * 
 * A proving problem for the RISC ProofNavigator.
 *
 */
public final class ProofNavigatorProblem 
{  
  // the file names
  private final static String GOALNAME = "goal_";
  // private final static String NEGGOALNAME = "neggoal_";
  
  // flat sequence of declarations
  private final Declaration[] decls; 
  
  // goal formula and declaration
  private final Expression goal;

  // stream for printing messages
  private final ErrorStream out;
  
  // the set of symbols referenced by the goal (created by processing the goal)
  private Collection<Symbol> grefSymbols;
  
  // goal environment (created by processing the goal) and old environment
  private Environment goalEnvironment;
  private static Environment oldEnvironment;
  
  // goal symbol (created by attempt to interactive proofs)
  private FormulaSymbol goalSymbol;
  
  // the corresponding goal declaration
  private FormulaDeclaration goalDeclaration;
  
  // the directory to be used for logs
  private File dir;
  
  // true if goal was proved to be true
  private boolean isTrue;
  
  // true if goal was proved to be false
  private boolean isFalse;
  
  // true if automatic proof was already attempted
  private boolean executed;
  
  // the current goal being proved (set if a proof is ongoing)
  private static FormulaSymbol currentGoalSymbol = null;
  
  // the current task being solved (set if a proof is ongoing)
  private static Task currentTask = null;
  
  /****************************************************************************
   * Construct a proving problem from declarations and goal.
   * @param decls the declarations.
   * @param goal the goal to be proved in the context of the declarations.
   * @param out stream for printing messages
   * @param dir the directory to be used for persistent data.
   ***************************************************************************/
  public ProofNavigatorProblem(Declaration[] decls, Expression goal, File dir,
    ErrorStream out)
  {
    this.decls = decls;
    this.goal = goal;
    this.dir = dir;
    this.out = out;
    this.grefSymbols = null;
    this.goalEnvironment = null;
    this.goalSymbol = null;
    this.goalDeclaration = null;
    this.isTrue = false;
    this.isFalse = false;
    this.executed = false;
  }
  
  /***************************************************************************
   * Determine whether there exists a (possibly incomplete or invalidated)
   * proof for this problem in the store denoted by dir.
   * @param dir the directory for the proof.
   * @return true if a proof exists.
   **************************************************************************/
  public static boolean existsProof(File dir)
  {
    File cdir = new File(dir, Directory.CONTEXT);
    State.setContextDir(cdir);
    Store store = State.getStore();
    return store.existsProof(GOALNAME);
  }
  
  /***************************************************************************
   * Determine whether there exists a (possibly incomplete or invalidated)
   * proof for this problem in the store.
   * @return true if a proof exists.
   **************************************************************************/
  public boolean existsProof()
  {
    return existsProof(dir);
  }
  
  /***************************************************************************
   * Determine whether problem is solved.
   * @return true if and only if problem. is solved.
   **************************************************************************/
  public boolean isSolved()
  {
    return isTrue;
  }
  
  /***************************************************************************
   * Determine whether problem is not solvable.
   * @return true if and only if problem cannot be solved.
   **************************************************************************/
  public boolean isUnsolvable()
  {
    return isFalse;
  }
  
  /****************************************************************************
   * Print the problem.
   * @param out the stream on which the task is to be printed.
   ***************************************************************************/
  public void print(PrintWriter out)
  {
    out.println("Declarations:");
    for (Declaration d : decls)
    {
      d.print(out);
      out.println(";");
    }
    out.print("Goal: ");
    goal.print(out);
    out.println("");
  }
  
  
  /****************************************************************************
   * Process the task by type checking all declarations and formulas.
   * Sets up the goal environment as a side effect.
   ***************************************************************************/
  private void process()
  {
    // start new prover context
    State.setNewProver();
    //State.getProver().pushContext();
    
    // prepare environment
    Environment env = State.getEnvironment();
    goalEnvironment = new Environment();
    State.setEnvironment(goalEnvironment);
    
    // establish type-checking context
    // Context context = State.getContext();

    // process declarations
    for (Declaration d : decls)
    {
      // context.open();
      new DeclarationC(d).process(false);
      String error = State.getError();
      if (error != null)
      {
        Main.internalError("error in processing declaration " + d + 
            ": " + error);
      }
      // Expression tcc = context.getFormula();
      // context.commit();
      // if (tcc != null) goalEnvironment.addTCC(tcc);
    }
    
    // process goal
    Checking.resetReferencedSymbols();
    // context.open();
    Type type = Checking.processFormula(goal);
    if (type == null)
    {
      Main.internalError("error in processing goal " + goal); 
    }
    grefSymbols = Checking.getReferencedSymbols();
    // Expression tcc = context.getFormula();
    // context.commit();
    // if (tcc != null) goalEnvironment.addTCC(tcc);
    
    // restore original environment and prover context
    State.setEnvironment(env);
    State.restoreOldProver();
    //State.getProver().popContext();
  }
  
  /***************************************************************************
   * Prove goal automatically
   * @param log the stream to be used for log messages
   * @param goal the goal to be proved.
   * @return true if proof succeeded.
   **************************************************************************/
  private boolean proveAuto(Expression goal, PrintWriter log)
  {
    Prover prover = State.newProver(log, 0, true);
    for (Declaration d : decls)
    {
      if (d instanceof AxiomDeclaration)
      {
        // assert truth of axioms
        AxiomDeclaration d0 = (AxiomDeclaration)d;
        prover.assertAxiom(d0.getFormula());
        continue;
      }
      if (d instanceof FormulaDeclaration)
      {
        // ignore plain formulas
        continue;
      }
      // add plain declaration
      prover.addDeclaration(d);
    }
    
    Answer answer = prover.queryFormula(goal);
    prover.terminate(true);
    return answer instanceof ValidAnswer;
  }
  
  /***************************************************************************
   * Prove goal automatically
   * @return true if proof succeeded.
   **************************************************************************/
  private boolean proveGoalAuto()
  {
    try
    {
      File file = new File(dir, GOALNAME);
      PrintWriter log = new PrintWriter(new FileOutputStream(file), true);
      return proveAuto(goal, log);
    }
    catch(FileNotFoundException file)
    {
      Main.internalError("error in opening file " + file);
    }
    return false;
  }

  /***************************************************************************
   * Prove negated goal automatically
   * @return true if proof succeeded.
   **************************************************************************/
  /* private boolean proveNegatedGoalAuto()
  {
    try
    {
      File file = new File(dir, NEGGOALNAME);
      PrintWriter log = new PrintWriter(new FileOutputStream(file), true);
      return proveAuto(Construct.notFormula(goal), log);
    }
    catch(FileNotFoundException file)
    {
      Main.internalError("error in opening file " + file);
    }
    return false;
  }
  */
  
  /***************************************************************************
   * Attempt to prove goal automatically.
   * @param task the associated task
   **************************************************************************/
  public void proveAuto(Task task)
  {    
    // do not attempt twice
    if (executed) return;

    // sets the goal environment as a side effect
    executed = true;
    process();
    isTrue = proveGoalAuto();
    if (isTrue)
    {
      task.setStatus(Task.Status.DONE, 
          "solved by decision procedure", null);
      return;
    }
    /*// speedup procedure
    isFalse = proveNegatedGoalAuto();
    if (isFalse)
    {
      task.setStatus(Task.Status.FAILED,
          "deemed unsolvable by decision procedure", null);
      return;
    }
    */
  }
  
  /***************************************************************************
   * Prepare interactive proof
   **************************************************************************/
  private void prepareProof()
  {
    // avoid repetition
    if (goalSymbol != null) return;

    // set up temporary context (old context restored before return)
    File cdir = new File(dir, Directory.CONTEXT);
    File odir = State.getContextDir();
    State.setContextDir(cdir);
    Proof oproof = State.getProof();
    State.setProof(null);
    
    // prepare goal environment
    if (goalEnvironment == null) process();
    FormulaDeclIdentifier ident = Construct.formulaDeclIdentifier(GOALNAME);
    goalSymbol = ident.getSymbol();
    goalSymbol.setFormula(goal);
    goalSymbol.setEnvironment(goalEnvironment);  
    goalDeclaration = Construct.formulaDeclaration(ident, goal);

    // write goal declaration to store
    Environment oenv = State.getEnvironment();
    State.setEnvironment(goalEnvironment);
    Store store = State.getStore();
    Vector<Symbol> referenced = 
      DeclarationC.getReferenced(grefSymbols, goalEnvironment);
    boolean okay = store.write(goalDeclaration, referenced);
    State.setEnvironment(oenv);
    if (!okay) 
    {
      State.setContextDir(odir);
      return;
    }
      
    // attempt to read proof
    if (store.existsProof(goalSymbol))
    {
      Environment fullEnvironment = goalEnvironment.cloneGlobal();
      if (goalEnvironment.getFormulaSymbol(ident) != null)
        Main.internalError("goal symbol contained in environment");
      fullEnvironment.putFormulaSymbol(goalSymbol);
      State.setEnvironment(fullEnvironment);
      Proof proof = store.readProof(goalSymbol);
      State.setEnvironment(oenv);
      PrintWriter out = State.getSystemOut();
      if (proof == null) 
        Main.internalError("error in reading proof from directory " + dir);
      goalSymbol.setProof(proof);
      out.print("Proof read (");
      proof.printStatus(out);
      out.println(").");
    }
    
    // restore context
    State.setContextDir(odir);
    State.setProof(oproof);
  }
  
  /***************************************************************************
   * Initiate proof.
   * @param task the associated task
   **************************************************************************/
  private void initProof(Task task)
  {
    // no nested proofs supported
    if (currentGoalSymbol != null)
    {
      out.newMessage("Cannot start interactive proof (close ongoing proof first)");
      return;
    }

    // prepare proof environment
    prepareProof();
    
    // set up proof context
    File cdir = new File(dir, Directory.CONTEXT);
    State.setContextDir(cdir);
    State.setProof(null);
    oldEnvironment = State.getEnvironment();
    State.setEnvironment(goalEnvironment);
    
    // start/resume proof
    currentGoalSymbol = goalSymbol;
    currentTask = task;
  }
  
  /***************************************************************************
   * Attempt pseudo-interactive proof with automatic strategy
   * @param task the associated task
   **************************************************************************/
  public void proveStrategy(Task task)
  {
    initProof(task);
    Prove.prove(goalSymbol);
    Proof proof = State.getProof();
    if (proof == null) return; // proof state is already closed
    new Scatter().process();
    if (proof.isClosed())
    {
      goalSymbol.setProof(proof);
      isTrue = true;
      task.setStatus(Task.Status.DONE, 
          "solved by automatic proof strategy", proof);
      State.getStore().write(proof);
    }
    closeProof();
  }
  
  /***************************************************************************
   * Start/resume manual proof.
   * @param task the associated task
   **************************************************************************/
  public void proveManual(Task task)
  {
    initProof(task);
    fmrisc.ProofNavigator.Main.displayGoal(decls, goalDeclaration);
    Prove.prove(goalSymbol);
  }
  
  /***************************************************************************
   * Called if a proof is completed by the ProofNavigator.
   * @param proof the completed proof.
   **************************************************************************/
  public static void proofDone(Proof proof)
  {
    // may not be "our" proof (user may use ProofNavigator)
    if (currentGoalSymbol == null) return;
    currentGoalSymbol.setProof(proof);
    if (proof.isClosed())
      currentTask.setStatus(Task.Status.DONE, 
          "solved by interactive proof", proof);
    else
      currentTask.setStatus(Task.Status.OPEN,
          "interactive proof started but not completed", proof);
    closeProof();
  }
  
  /***************************************************************************
   * Closes any pending proof.
   **************************************************************************/
  public static void closeProof()
  {
    currentGoalSymbol = null;
    currentTask = null;
    State.setProof(null);
    State.setEnvironment(oldEnvironment);
    oldEnvironment = null;
  }
  
  /****************************************************************************
   * Get current proof, possibly load it from file.
   * @return proof the current proof (may be null).
   ***************************************************************************/
  public Proof getProof()
  {
    prepareProof();
    return goalSymbol.getProof();
  }
}
