// ----------------------------------------------------------------------------
// DirectorySelectionWindow.java
// A window for selecting multiple directories.
// $Id: DirectorySelectionWindow.java,v 1.6 2010/03/03 16:01:28 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.SWT;

import fmrisc.ProgramExplorer.*;

import java.io.*;

import org.eclipse.swt.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.*;

/**
 * 
 * A window for selecting multiple directories.
 *
 */
public final class DirectorySelectionWindow extends Composite
{
  // the GUI of the window
  private TopWindow top;
  
  // the shell of the window
  private Shell shell;
  
  // the directory paths
  private File[] paths;
  
  /****************************************************************************
   * Construct window connected to denoted top Window
   * @param top the top window.
   * @param paths the preselected directory paths.
   * @return a window connected to the GUI.
   ***************************************************************************/
  public static DirectorySelectionWindow construct(TopWindow top, File[] paths)
  {
    Display display = top.getDisplay();
    Shell shell = new Shell(display, 
      SWT.CLOSE | SWT.MIN | SWT.MAX | SWT.RESIZE | SWT.TITLE | SWT.BORDER);
    shell.setText("RISC ProgramExplorer: Select Workspace Directories");
    Image riscImage = new Image(shell.getDisplay(),
        Main.class.getResourceAsStream("/" + MainSWT.riscImageName));
    shell.setImage(riscImage);
    return new DirectorySelectionWindow(top, shell, paths);
  }
  
  /****************************************************************************
   * Construct panel for denoted top window and shell.
   * @param top the top window.
   * @param shell the shell to which the panel is attached.
   * @param paths the preselected directory paths.
   ***************************************************************************/
  public DirectorySelectionWindow(TopWindow top, Shell shell, File[] paths)
  {
    // initialize window
    super(shell, SWT.NONE);
    this.top = top;
    this.shell = shell;
    this.paths = paths;
    
    // resize it
    shell.setSize(500, 300);
    
    // set layout
    GridLayout shellLayout = new GridLayout();
    shell.setLayout(shellLayout);
    shellLayout.numColumns = 1;
    shellLayout.marginWidth = 0;
    shellLayout.marginHeight = 0;
    
    // set own layout
    setLayoutData(new GridData(GridData.FILL_BOTH));
    GridLayout topLayout =  new GridLayout();
    setLayout(topLayout);
    topLayout.numColumns = 1;
    topLayout.verticalSpacing = 10;
    topLayout.marginTop = 3;
    topLayout.marginHeight = 3;
    topLayout.marginWidth = 5;
    
    // create widgets
    createMenu();
    createPanel();
  }
  
  /***************************************************************************
   * Create menu bar.
   **************************************************************************/
  private void createMenu()
  {
    new Menu(shell, SWT.BAR);
    shell.setMenuBar(null);
  }
  
  /***************************************************************************
   * Create panel.
   **************************************************************************/
  private void createPanel()
  {
    Group listGroup = new Group(this, SWT.BORDER);
    listGroup.setLayout(new FillLayout());
    listGroup.setText("Workspace Directories (from PE_CLASSPATH)");
    
    GridData listData = new GridData(GridData.FILL_HORIZONTAL);
    listGroup.setLayoutData(listData);
    listData.heightHint=130;
    
    final List list = new List(listGroup, SWT.MULTI | SWT.BORDER | 
        SWT.H_SCROLL | SWT.V_SCROLL);
    for (File p : paths)
      list.add(p.getAbsolutePath());
    
    Composite actions = new Composite(this, SWT.NONE);
    actions.setLayout(new RowLayout());
    Button add = new Button(actions, SWT.PUSH);
    add.setText("Add Directory");
    Button remove = new Button(actions, SWT.PUSH);
    remove.setText("Remove Directory");
    Button restore = new Button(actions, SWT.PUSH);
    restore.setText("Restore Directories");
    
    GridData actionData = new GridData();
    actions.setLayoutData(actionData);
    actionData.horizontalAlignment = GridData.CENTER;
      
    Group done = new Group(this, SWT.BORDER);
    done.setLayout(new RowLayout());
    Button okay = new Button(done, SWT.PUSH);
    okay.setText("Okay");
    okay.setImage(top.getOkayImage());
    shell.setDefaultButton(okay);
    Button cancel = new Button(done, SWT.PUSH);
    cancel.setText("Cancel");
    cancel.setImage(top.getCancelImage());
    
    GridData doneData = new GridData();
    done.setLayoutData(doneData);
    doneData.horizontalAlignment = GridData.CENTER;
    
    add.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        DirectoryDialog dialog = new DirectoryDialog(shell);
        dialog.setText("Select Directory");
        String result = dialog.open();
        if (result == null) return;
        String[] items = list.getItems();
        for (String item: items)
        {
          if (result.equals(item)) return;
        }
        list.add(result);
      }
    });
    
    remove.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        int[] indices = list.getSelectionIndices();
        list.remove(indices);
      }
    });
    
    restore.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        list.removeAll();
        for (File p : paths)
          list.add(p.getAbsolutePath());
      }
    });
    
    okay.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        String[] names = list.getItems();
        int n = names.length;
        paths = new File[n];
        for (int i=0; i<n; i++)
          paths[i] = new File(names[i]);
        top.redrawFileExplorer(paths);
        shell.close();
      }
    });
    
    cancel.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        paths = null;
        shell.close();
      }
    });
    
  }
}

