// ----------------------------------------------------------------------------
// PropertyConfigurationWindow.java
// A window for configuring the system properties
// $Id: PropertyConfigurationWindow.java,v 1.10 2015/09/17 09:31:05 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.SWT;

import fmrisc.ProgramExplorer.*;

import org.eclipse.swt.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.custom.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.*;

/**
 * 
 * A window for selecting multiple directories.
 *
 */
public final class PropertyConfigurationWindow extends Composite
{
  // the GUI of the window
  private TopWindow top;
  
  // the shell of the window
  private Shell shell;
  
  // the tab folder
  private CTabFolder folder;
  private CTabItem pathsTab;
  private SashForm pathsForm;
  private CTabItem settingsTab;
  private Composite settingsForm;
  
  // the various paths settings
  private String compileCommand;
  private String executeCommand;
  private String workingDirectory;
  private String mainClass;
  private String cvclCommand;
  private String cvc3Command;
  private String cvc4Command;
  
  // the various checker settings
  private String proverVersion;
  private String firstOrder;
  private String toLogical;
  
  /****************************************************************************
   * Construct window connected to denoted top Window
   * @param top the top window.
   * @param properties the initial values of the configurable properties.
   * @return a window connected to the GUI.
   ***************************************************************************/
  public static PropertyConfigurationWindow construct(TopWindow top, 
    String[] properties)
  {
    Display display = top.getDisplay();
    Shell shell = new Shell(display, 
      SWT.CLOSE | SWT.MIN | SWT.MAX | SWT.RESIZE | SWT.TITLE | SWT.BORDER);
    shell.setText("RISC ProgramExplorer: Configure System Properties");
    Image riscImage = new Image(shell.getDisplay(),
        Main.class.getResourceAsStream("/" + MainSWT.riscImageName));
    shell.setImage(riscImage);
    return new PropertyConfigurationWindow(top, shell, properties);
  }
  
  /****************************************************************************
   * Construct window connected to denoted top window.
   * @param top the top window.
   * @param shell the parent shell.
   * @param properties the initial values of the configurable properties.
   ***************************************************************************/
  private PropertyConfigurationWindow(TopWindow top, Shell shell,
    String[] properties)
  {
    // initialize window
    super(shell, SWT.NONE);
    this.top = top;
    this.shell = shell;
    
    // set properties
    if (properties.length != 10) 
      Main.internalError("invalid number of properties");
    this.compileCommand = properties[0];
    this.executeCommand = properties[1];
    this.workingDirectory = properties[2];
    this.mainClass = properties[3];
    this.cvclCommand = properties[4];
    this.cvc3Command = properties[5];
    this.cvc4Command = properties[6];
    this.proverVersion = properties[7];
    this.firstOrder = properties[8];
    this.toLogical = properties[9];
    
    // resize it
    shell.setSize(500, 600);
    
    // set layout
    GridLayout shellLayout = new GridLayout();
    shell.setLayout(shellLayout);
    shellLayout.numColumns = 1;
    shellLayout.marginWidth = 0;
    shellLayout.marginHeight = 0;
    
    // set own layout
    setLayoutData(new GridData(GridData.FILL_BOTH));
    GridLayout topLayout =  new GridLayout();
    setLayout(topLayout);
    topLayout.numColumns = 1;
    topLayout.verticalSpacing = 10;
    topLayout.marginTop = 3;
    topLayout.marginHeight = 3;
    topLayout.marginWidth = 5;
    
    // create widgets
    createMenu();
    createPanel();
  }
  
  /***************************************************************************
   * Create menu bar.
   **************************************************************************/
  private void createMenu()
  {
    new Menu(shell, SWT.BAR);
    shell.setMenuBar(null);
  }
  
  /***************************************************************************
   * Create panel.
   **************************************************************************/
  private void createPanel()
  {    
 // ----------------------------------------------------------------------- 
    // create tab folder
    // -----------------------------------------------------------------------
    
    folder = new CTabFolder(this, 
        SWT.TOP | SWT.LEFT_TO_RIGHT | SWT.MULTI | SWT.BORDER);
    folder.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    folder.setSimple(false);
    folder.setSelectionBackground(folder.getBackground());
    folder.addSelectionListener(      
        new SelectionListener() {
          public void widgetSelected(SelectionEvent e) {
            CTabItem item = (CTabItem)e.item;
            if (item == pathsTab)
            {
              folder.setSelection(pathsTab);
            }
            else if (item == settingsTab)
            {
              folder.setSelection(settingsTab);
            }
            else
              Main.internalError("unknown view");
          }
          public void widgetDefaultSelected(SelectionEvent e) {
            // widgetSelected(e);
          }
        });
    
    pathsTab = new CTabItem(folder, SWT.NULL);
    pathsTab.setText("File Paths");
    // pathsTab.setImage(folderImage);
    pathsForm = new SashForm(folder, 
        SWT.LEFT_TO_RIGHT | SWT.VERTICAL | SWT.SMOOTH);
    pathsTab.setControl(pathsForm);
    
    settingsTab = new CTabItem(folder, SWT.NULL);
    settingsTab.setText("Checker Settings");
    // pathsTab.setImage(folderImage);
    settingsForm = new Composite(folder, SWT.NONE);
    settingsTab.setControl(settingsForm);
    
    GridLayout settingsLayout = new GridLayout();
    settingsForm.setLayout(settingsLayout);
    settingsLayout.numColumns = 1;
    settingsLayout.marginWidth = 0;
    
    folder.setSelection(settingsTab);
    
    // ----------------------------------------------------------------------- 
    // compile property
    // -----------------------------------------------------------------------
    
    Group compileGroup = new Group(pathsForm, SWT.BORDER);
    compileGroup.setText("Compile Command (from PE_JAVAC)");
    GridLayout compileLayout = new GridLayout();
    compileGroup.setLayout(compileLayout);
    compileLayout.numColumns = 2;
    
    GridData compileData = new GridData(GridData.FILL_HORIZONTAL);
    compileGroup.setLayoutData(compileData);
    
    final Text compileText = new Text(compileGroup, SWT.SINGLE | SWT.BORDER | 
        SWT.H_SCROLL | SWT.V_SCROLL);
    compileText.setText(compileCommand);
    compileText.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL 
    | GridData.GRAB_HORIZONTAL));
    
    Button compileRestore = new Button(compileGroup, SWT.PUSH);
    compileRestore.setText("Restore");
    
    compileRestore.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        compileText.setText(compileCommand);
      }
    });
    
    // ----------------------------------------------------------------------- 
    // execute property
    // -----------------------------------------------------------------------
    
    Group executeGroup = new Group(pathsForm, SWT.BORDER);
    executeGroup.setText("Execute Command (from PE_JAVA)");
    GridLayout executeLayout = new GridLayout();
    executeGroup.setLayout(executeLayout);
    executeLayout.numColumns = 2;
    
    GridData executeData = new GridData(GridData.FILL_HORIZONTAL);
    executeGroup.setLayoutData(executeData);
    
    final Text executeText = new Text(executeGroup, SWT.SINGLE | SWT.BORDER | 
        SWT.H_SCROLL | SWT.V_SCROLL);
    executeText.setText(executeCommand);
    executeText.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL 
    | GridData.GRAB_HORIZONTAL));
    
    Button executeRestore = new Button(executeGroup, SWT.PUSH);
    executeRestore.setText("Restore");
    
    executeRestore.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        executeText.setText(executeCommand);
      }
    });
    
    // ----------------------------------------------------------------------- 
    // working directory property
    // -----------------------------------------------------------------------
    
    Group workingGroup = new Group(pathsForm, SWT.BORDER);
    workingGroup.setText("Working Directory (from PE_CWD)");
    GridLayout workingLayout = new GridLayout();
    workingGroup.setLayout(workingLayout);
    workingLayout.numColumns = 2;
    
    GridData workingData = new GridData(GridData.FILL_HORIZONTAL);
    workingGroup.setLayoutData(workingData);
    
    final Text workingText = new Text(workingGroup, SWT.SINGLE | SWT.BORDER | 
        SWT.H_SCROLL | SWT.V_SCROLL);
    workingText.setText(workingDirectory);
    workingText.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL 
    | GridData.GRAB_HORIZONTAL));
    
    Button workingRestore = new Button(workingGroup, SWT.PUSH);
    workingRestore.setText("Restore");
    
    workingRestore.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        workingText.setText(workingDirectory);
      }
    });
    
    // ----------------------------------------------------------------------- 
    // working directory property
    // -----------------------------------------------------------------------
    
    Group mainGroup = new Group(pathsForm, SWT.BORDER);
    mainGroup.setText("Main Class (from PE_MAIN)");
    GridLayout mainLayout = new GridLayout();
    mainGroup.setLayout(mainLayout);
    mainLayout.numColumns = 2;
    
    GridData mainData = new GridData(GridData.FILL_HORIZONTAL);
    mainGroup.setLayoutData(mainData);
    
    final Text mainText = new Text(mainGroup, SWT.SINGLE | SWT.BORDER | 
        SWT.H_SCROLL | SWT.V_SCROLL);
    mainText.setText(mainClass);
    mainText.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL 
    | GridData.GRAB_HORIZONTAL));
    
    Button mainRestore = new Button(mainGroup, SWT.PUSH);
    mainRestore.setText("Restore");
    
    mainRestore.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        mainText.setText(mainClass);
      }
    });
    
    // ----------------------------------------------------------------------- 
    // cvcl property
    // -----------------------------------------------------------------------
    
    Group cvclGroup = new Group(pathsForm, SWT.BORDER);
    cvclGroup.setText("CVCL Command (from PE_CVCL)");
    GridLayout cvclLayout = new GridLayout();
    cvclGroup.setLayout(cvclLayout);
    cvclLayout.numColumns = 2;
    
    GridData cvclData = new GridData(GridData.FILL_HORIZONTAL);
    cvclGroup.setLayoutData(cvclData);
    
    final Text cvclText = new Text(cvclGroup, SWT.SINGLE | SWT.BORDER | 
        SWT.H_SCROLL | SWT.V_SCROLL);
    cvclText.setText(cvclCommand);
    cvclText.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL 
    | GridData.GRAB_HORIZONTAL));
    // cvclText.setEditable(false);
    // cvclText.setBackground(getDisplay().getSystemColor(SWT.COLOR_GRAY));
    
    Button cvclRestore = new Button(cvclGroup, SWT.PUSH);
    cvclRestore.setText("Restore");
    
    cvclRestore.addSelectionListener(new SelectionAdapter() {
       public void widgetSelected(SelectionEvent e) {
         cvclText.setText(cvclCommand);
       }
     });
    
    // ----------------------------------------------------------------------- 
    // cvc3 property
    // -----------------------------------------------------------------------
    
    Group cvc3Group = new Group(pathsForm, SWT.BORDER);
    cvc3Group.setText("CVC3 Command (from PE_CVC3)");
    GridLayout cvc3Layout = new GridLayout();
    cvc3Group.setLayout(cvc3Layout);
    cvc3Layout.numColumns = 2;
    
    GridData cvc3Data = new GridData(GridData.FILL_HORIZONTAL);
    cvc3Group.setLayoutData(cvc3Data);
    
    final Text cvc3Text = new Text(cvc3Group, SWT.SINGLE | SWT.BORDER | 
        SWT.H_SCROLL | SWT.V_SCROLL);
    cvc3Text.setText(cvc3Command);
    cvc3Text.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL 
    | GridData.GRAB_HORIZONTAL));
    // cvc3Text.setEditable(false);
    // cvc3Text.setBackground(getDisplay().getSystemColor(SWT.COLOR_GRAY));
    
    Button cvc3Restore = new Button(cvc3Group, SWT.PUSH);
    cvc3Restore.setText("Restore");
    
    cvc3Restore.addSelectionListener(new SelectionAdapter() {
       public void widgetSelected(SelectionEvent e) {
         cvc3Text.setText(cvc3Command);
       }
     });
    
    // ----------------------------------------------------------------------- 
    // cvc4 property
    // -----------------------------------------------------------------------
    
    Group cvc4Group = new Group(pathsForm, SWT.BORDER);
    cvc4Group.setText("CVC4 Command (from PE_CVC4)");
    GridLayout cvc4Layout = new GridLayout();
    cvc4Group.setLayout(cvc4Layout);
    cvc4Layout.numColumns = 2;
    
    GridData cvc4Data = new GridData(GridData.FILL_HORIZONTAL);
    cvc4Group.setLayoutData(cvc4Data);
    
    final Text cvc4Text = new Text(cvc4Group, SWT.SINGLE | SWT.BORDER | 
        SWT.H_SCROLL | SWT.V_SCROLL);
    cvc4Text.setText(cvc4Command);
    cvc4Text.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL 
    | GridData.GRAB_HORIZONTAL));
    // cvc4Text.setEditable(false);
    // cvc4Text.setBackground(getDisplay().getSystemColor(SWT.COLOR_GRAY));
    
    Button cvc4Restore = new Button(cvc4Group, SWT.PUSH);
    cvc4Restore.setText("Restore");
    
    cvc4Restore.addSelectionListener(new SelectionAdapter() {
       public void widgetSelected(SelectionEvent e) {
         cvc4Text.setText(cvc4Command);
       }
     });
    
    // ----------------------------------------------------------------------- 
    // button group for selecting default checker
    // -----------------------------------------------------------------------
    
    Group checker = new Group(settingsForm, SWT.BORDER);
    checker.setText("Default Validity Checker (from PE_CHECKER)");
    checker.setLayout(new RowLayout());
    final Button cvclButton = new Button(checker, SWT.RADIO);
    cvclButton.setText("CVCL");
    final Button cvc3Button = new Button(checker, SWT.RADIO);
    cvc3Button.setText("CVC3");
    final Button cvc4Button = new Button(checker, SWT.RADIO);
    cvc4Button.setText("CVC4");
    // cvc4Button.setEnabled(false);
    
    GridData checkerData = new GridData();
    checker.setLayoutData(checkerData);
    checkerData.horizontalAlignment = GridData.FILL;
    checkerData.verticalAlignment = GridData.BEGINNING;
    checkerData.grabExcessHorizontalSpace = true;
    
    Group firstOrder = new Group(settingsForm, SWT.BORDER);
    final Button firstOrderTrueButton = new Button(firstOrder, SWT.RADIO);
    final Button firstOrderFalseButton = new Button(firstOrder, SWT.RADIO);
    
    Group logical = new Group(settingsForm, SWT.BORDER);
    final Button logicalTrueButton = new Button(logical, SWT.RADIO);
    final Button logicalFalseButton = new Button(logical, SWT.RADIO);
    
    switch (proverVersion)
    {
    case "CVCL": 
        cvclButton.setSelection(true);
        break;
    case "CVC3": 
        cvc3Button.setSelection(true);
        break;
    case "CVC4": 
        cvc4Button.setSelection(true);
        break;
    default:
        Main.internalError("unknown prover type");
    }
    cvclButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        cvclButton.setSelection(true);
        cvc3Button.setSelection(false);  
        cvc4Button.setSelection(false);
        proverVersion = "CVCL";
        firstOrderTrueButton.setEnabled(true);
        firstOrderFalseButton.setEnabled(true);
        if (PropertyConfigurationWindow.this.firstOrder.equals("TRUE"))
        {
          logicalTrueButton.setEnabled(true);
          logicalFalseButton.setEnabled(true);
        }
        else
        {
          logicalTrueButton.setEnabled(false);
          logicalFalseButton.setEnabled(false);
        }
      }
    });
    cvc3Button.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        cvclButton.setSelection(false);  
        cvc3Button.setSelection(true);
        cvc4Button.setSelection(false);
        proverVersion = "CVC3";
        firstOrderTrueButton.setSelection(true);
        firstOrderTrueButton.setEnabled(true);
        firstOrderFalseButton.setSelection(false);
        firstOrderFalseButton.setEnabled(false);
        logicalTrueButton.setEnabled(true);
        logicalFalseButton.setEnabled(true);
      }
    });
    cvc4Button.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        cvclButton.setSelection(false);  
        cvc3Button.setSelection(false);
        cvc4Button.setSelection(true);
        proverVersion = "CVC4";
        firstOrderTrueButton.setSelection(true);
        firstOrderTrueButton.setEnabled(true);
        firstOrderFalseButton.setSelection(false);
        firstOrderFalseButton.setEnabled(false);
        logicalTrueButton.setEnabled(true);
        logicalFalseButton.setEnabled(true);
      }
    });
    
    // ----------------------------------------------------------------------- 
    // button group for selecting first order
    // -----------------------------------------------------------------------
      
    firstOrder.setText("Logic (from PE_HIGHERORDER, selection only for CVCL)");
    firstOrder.setLayout(new RowLayout());
    firstOrderTrueButton.setText("First Order");
    firstOrderTrueButton.setEnabled(true);
    firstOrderFalseButton.setText("Higher Order");;
    firstOrderFalseButton.setEnabled(proverVersion.equals("CVCL"));
    
    GridData firstOrderData = new GridData();
    firstOrder.setLayoutData(firstOrderData);
    firstOrderData.horizontalAlignment = GridData.FILL;;
    firstOrderData.verticalAlignment = GridData.BEGINNING;
    switch (this.firstOrder)
    {
    case "TRUE": 
        firstOrderTrueButton.setSelection(true);
        break;
    case "FALSE": 
        firstOrderFalseButton.setSelection(true);
        break;
    default:
        Main.internalError("unknown order type");
    }
    firstOrderTrueButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        firstOrderTrueButton.setSelection(true);
        firstOrderFalseButton.setSelection(false);  
        PropertyConfigurationWindow.this.firstOrder = "TRUE";
        logicalTrueButton.setEnabled(true);
        logicalFalseButton.setEnabled(true);
      }
    });
    firstOrderFalseButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        firstOrderTrueButton.setSelection(false);
        firstOrderFalseButton.setSelection(true);  
        PropertyConfigurationWindow.this.firstOrder = "FALSE";
        logicalTrueButton.setEnabled(false);
        logicalFalseButton.setEnabled(false);
      }
    });
    
    // ----------------------------------------------------------------------- 
    // button group for selecting logical type
    // -----------------------------------------------------------------------
  
    logical.setText("Formula Type (from PE_BOOLEANFORMULA)");
    logical.setLayout(new RowLayout());
    logicalTrueButton.setText("LOGICAL (value type is BOOLEAN)");
    boolean first = this.firstOrder.equals("TRUE");
    logicalTrueButton.setEnabled(first);
    logicalFalseButton.setText("BOOLEAN (value type is BIT)");;
    logicalFalseButton.setEnabled(first);
    
    GridData logicalData = new GridData();
    logical.setLayoutData(logicalData);
    logicalData.horizontalAlignment = GridData.FILL;;
    logicalData.verticalAlignment = GridData.BEGINNING;
    switch (this.toLogical)
    {
    case "TRUE": 
        logicalTrueButton.setSelection(true);
        break;
    case "FALSE": 
        logicalFalseButton.setSelection(true);
        break;
    default:
        Main.internalError("unknown order type");
    }
    logicalTrueButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        logicalTrueButton.setSelection(true);
        logicalFalseButton.setSelection(false);  
        PropertyConfigurationWindow.this.toLogical = "TRUE";
      }
    });
    logicalFalseButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        logicalTrueButton.setSelection(false);
        logicalFalseButton.setSelection(true);  
        PropertyConfigurationWindow.this.toLogical = "FALSE";
      }
    });
    
   
    // ----------------------------------------------------------------------- 
    // button group for confirming action
    // -----------------------------------------------------------------------
    
    Group done = new Group(this, SWT.BORDER);
    done.setLayout(new RowLayout());
    Button okay = new Button(done, SWT.PUSH);
    okay.setText("Okay");
    okay.setImage(top.getOkayImage());
    shell.setDefaultButton(okay);
    Button cancel = new Button(done, SWT.PUSH);
    cancel.setText("Cancel");
    cancel.setImage(top.getCancelImage());
    
    GridData doneData = new GridData();
    done.setLayoutData(doneData);
    doneData.horizontalAlignment = GridData.CENTER;
    
    okay.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        Main.setProperties(
            compileText.getText(), executeText.getText(),
            workingText.getText(), mainText.getText(),
            cvclText.getText(), 
            cvc3Text.getText(),
            cvc4Text.getText(),
            proverVersion,
            proverVersion.equals("CVCL") ?
              PropertyConfigurationWindow.this.firstOrder : "TRUE",
            toLogical);
        shell.close();
        top.refresh();
      }
    });
    
    cancel.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        shell.close();
      }
    });
    
  }
}


