// ----------------------------------------------------------------------------
// FreeVariables.java
// Compute the free variables of an expression.
// $Id: FreeVariables.java,v 1.12 2012/12/16 18:09:40 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import java.util.*;

/**
 * 
 * Compute the free variables of an expression (revised version).
 *
 */
public final class FreeVariables extends ASTVisitorBase
{
  /***************************************************************************
   * Compute free variables of expression.
   * @param exp an expression.
   * @return a collection of the names of the free variables of the expression
   **************************************************************************/
  public static Collection<String> compute(Expression exp)
  {
    FreeVariables fv = new FreeVariables();
    exp.accept(fv);
    return fv.getVariables();
  }
  
  /***************************************************************************
   * Compute free variables of expression.
   * @param exp an expression.
   * @param result a collection to which the names of the free variables 
   *        of the expression are added.
   **************************************************************************/
  public static void add(Expression exp, Collection<String> result)
  {
    FreeVariables fv = new FreeVariables(result);
    exp.accept(fv);
  }
  
  // the collection holding the free variable symbols computed up to now
  private Collection<String> free = null;
  
  /***************************************************************************
   * Create visitor with empty result collection.
   **************************************************************************/
  private FreeVariables()
  {
    this(new SetClass<String>());
  }
  
  /****************************************************************************
   * Create visitor with denoted result collection.
   * @param result the collection to which the free variables are added.
   ***************************************************************************/
  private FreeVariables(Collection<String> result)
  {
    free = result;
  }
  
  /****************************************************************************
   * Get free variables computed by the expression visitor.
   * @return a collection of the names of the free variables.
   ***************************************************************************/
  private Collection<String> getVariables()
  {
    return free;
  }
  
  /****************************************************************************
   * Process reference.
   * @param ref a reference.
   ***************************************************************************/
  public void visit(Reference ref) 
  { 
    // strings from identifiers are unique by construction
    Identifier ident = ref.getName().idents[0];
    if (!(ident.getSymbol() instanceof ValueSymbol)) return;
    String var = ident.getString();
    if (!free.contains(var)) { free.add(var); }
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(QuantifiedExpression exp)
  {    
    // collect free variables of expression body
    Collection<String> f = compute(exp.getBody());
    
    // remove bound variables
    TypedIdentifier[] variables = exp.getVariables();
    for (TypedIdentifier var : variables)
      f.remove(var.getName().getString());
    
    // add remaining variables to collection
    free.addAll(f);
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(LetExpression exp)
  { 
    // the definitions in the expression
    Definition[] definitions = exp.getDefinitions();
    
    // collect free variables of expression body and definition values
    // remove defined values correspondingly
    Collection<String> f = compute(exp.getBody());
    int n = definitions.length;
    for (int i=n-1; i>=0; i--)
    {
      Definition d = definitions[i];
      f.remove(d.getName().getString());
      add(d.getValue(), f);
    }
    
    // add remaining variables to collection
    free.addAll(f);
  }
}