// ----------------------------------------------------------------------------
// LogicEnvironment.java
// An environment giving meaning to identifiers in the logic context.
// $Id: LogicEnvironment.java,v 1.34 2011/07/05 09:26:18 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;

import fmrisc.ProgramExplorer.Main;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Program.Import;
import fmrisc.ProgramExplorer.Syntax.Program.PackageImport;
import fmrisc.ProgramExplorer.Syntax.Program.ClassImport;

/**
 * 
 * An environment giving meaning to identifiers in the logic context.
 *
 */
public final class LogicEnvironment
{
  // the top-level package for looking up global names
  public PackageSymbol top;
  
  // the stream for printing error messages
  private ErrorStream error;

  // the actually used theories in topological dependence order
  private Vector<TheorySymbol> theories;
  
  // the value environment (nested)
  private Stack<SymbolTable<ValueSymbol>> vstack = 
    new Stack<SymbolTable<ValueSymbol>>();        // stack of value tables
  private SymbolTable<ValueSymbol> vtable = null; // current top of vstack;
  
  // the type and formula environment (not nested)
  private SymbolTable<TypeSymbol> ttable = null;
  private SymbolTable<FormulaSymbol> ftable = null;
  
  // the declarations (order preserved)
  private Stack<Vector<Declaration>> dstack =
    new Stack<Vector<Declaration>>();         // stack of declarations
  private Vector<Declaration> decls =         // current top of dstack
    new Vector<Declaration>(); 
  
  /****************************************************************************
   * Create a new environment at top-level.
   * @param top the top-level package for looking up global names
   * @param error the stream for printing error messages.
   ***************************************************************************/
  public LogicEnvironment(PackageSymbol top, ErrorStream error)
  {
    // initialize state
    this.top = top;
    this.error = error;
    ftable = new SymbolTable<FormulaSymbol>(error); 
    ttable = new SymbolTable<TypeSymbol>(error);
    
    // enter scope
    enter();
  }
  
  /****************************************************************************
   * Create a new environment from theory symbol.
   * @param tsymbol the symbol from which to create the environment
   * @param error the stream for printing error messages.
   ***************************************************************************/
  public LogicEnvironment(TheorySymbol tsymbol, ErrorStream error)
  {
    this.top = tsymbol.getPackage().getTopPackage();
    this.error = error;
    ftable = tsymbol.getFormulas();
    ttable = tsymbol.getTypes();
    
    // enter scope from theory environment
    vtable = tsymbol.getValues();
    vstack.push(vtable);
    decls = new Vector<Declaration>();
    Declaration[] ds = tsymbol.getDeclaration().getDeclarations();
    for (Declaration d : ds)
      decls.add(d);
    dstack.push(decls);
  }
  
  /***************************************************************************
   * Create a new and empty environment.
   * @param error the error stream to write to.
   **************************************************************************/
  public LogicEnvironment(ErrorStream error)
  {
    // we are not allowed to lookup theories
    this.top = null;
    this.error = error;
    ftable = new SymbolTable<FormulaSymbol>(error); 
    ttable = new SymbolTable<TypeSymbol>(error);
    
    // enter scope
    enter();
  }
  
  /***************************************************************************
   * Set in this environment the theories that may be referenced
   * (ensuring that they are indeed read from file).
   * @param psymbol the package in which this environment is processed.
   * @param baseTheory the always referable base theory (may be null)
   * @param classTheory the theory of the current class (may be null)
   * @param imports the import statements to be processed.
   * @param tnames the theory names that may be referenced.
   * @return the vector of theory symbols corresponding to tnames
   *         in topological dependence order (null, if an error occurred).
   **************************************************************************/
  public Vector<TheorySymbol> setTheories(PackageSymbol psymbol,
      TheorySymbol baseTheory, TheorySymbol classTheory,
      Import[] imports, Name[] tnames)
  {    
    // 1. determine the theories visible in current package
    SymbolTable<TheorySymbol> visible = new SymbolTable<TheorySymbol>(error);

    // 1a. process imports to link names to symbols
    int n = imports.length;
    for (int i=0; i<n; i++)
    {
      Import imp = imports[i];
      Name iname = imp.getName();
      Symbol symbol = null;
      if (imp instanceof ClassImport)
        symbol = lookupTheory(iname);
      else
        symbol = top.getPackage(iname);
      if (symbol == null)
      {
        error.explainError(imp.toString());
        return null;
      }
    }
    
    // 1b. theories declared in current package
    visible.add(psymbol.getTheories(), false, false);
        
    // 1c. theories in package imports, skip those that are already visible
    for (Import imp : imports)
    {
      if (!(imp instanceof PackageImport)) continue;
      PackageSymbol isymbol = (PackageSymbol)imp.getName().getSymbol();
      visible.add(isymbol.getTheories(), false, true);
    }
    
    // 1d. theories in class imports, override those that are already visible
    SymbolTable<TheorySymbol> isymbols = new SymbolTable<TheorySymbol>(error);
    // base theory first, must not be overridden
    if (baseTheory != null) isymbols.put(baseTheory); 
    // class theory next, must not be overridden
    if (classTheory != null) isymbols.put(classTheory); 
    for (Import imp : imports)
    {
      if (!(imp instanceof ClassImport)) continue;
      Name iname = imp.getName();
      Identifier base = iname.getBase();
      TheorySymbol isymbol = (TheorySymbol)base.getSymbol();
      boolean okay = isymbols.put(isymbol);
      if (!okay)
      {
        TheorySymbol isymbol0 = isymbols.get(base);
        error.newError(base, "import of theory " + iname + 
          " conflicts with already imported theory " + isymbol0.getName());
        return null;
      }
    } 
    visible.add(isymbols, true, false);
    
    // 2. set those theories that may be actually used
    //    and ensure that they are read from file
    // base theory may be used without naming it
    theories = new Vector<TheorySymbol>();
    if (baseTheory != null) theories.add(baseTheory); 
    if (classTheory != null) theories.add(classTheory); 
    for (Name name : tnames)
    {
      int m = name.getLength();
      TheorySymbol usymbol;
      if (m == 1)
        usymbol = visible.get(name.getBase());
      else
        usymbol = lookupTheory(name);  
      if (usymbol == null)
      {
        error.newError(name.getBase(), 
            "there is no theory " + name + " visible");
        return null;
      }
      theories.add(usymbol);
    }
    
    // 3. make sure theories are read
    for (TheorySymbol usymbol : theories)
    {
      if (!usymbol.isRead()) 
      {
        boolean okay = usymbol.read(error);
        if (!okay) 
        {
          error.explainError2("error in theory " + usymbol.getLongName());
          return null;
        }
      }
    }
    return theories;
  }
  
  /***************************************************************************
   * Get theories that may be referenced in environment in dependence order.
   * @return the collection of theories.
   **************************************************************************/
  public Vector<TheorySymbol> getTheories()
  {
    return theories;
  }
  
  /***************************************************************************
   * Get theory denoted by name.
   * @param name the name of a theory.
   * @return the corresponding theory symbol (null, if none).
   **************************************************************************/
  private TheorySymbol lookupTheory(Name name)
  {
    int n = name.getLength();
    PackageSymbol psymbol;
    if (n == 1)
      psymbol = top;
    else
    {
      Name prefix = name.getPrefix();
      psymbol = top.getPackage(prefix);
      if (psymbol == null) return null;
    }
    Identifier ident = name.getBase();
    TheorySymbol tsymbol = psymbol.getTheories().get(ident);
    if (tsymbol == null)
    {
      error.newError(ident, "there is no theory named " + name);
      return null;
    }
    return tsymbol;
  }
  
  /***************************************************************************
   * Enter a new scope.
   * All changes done to the environment after enter() will be undone
   * after exit().
   **************************************************************************/
  public void enter()
  {
    vtable = new SymbolTable<ValueSymbol>(error);
    vstack.push(vtable);
    decls = new Vector<Declaration>();
    dstack.push(decls);
  }
  
  /***************************************************************************
   * Exit a scope.
   * All changes done after the after the last enter() are undone
   **************************************************************************/
  public void exit()
  {
    vstack.pop();
    vtable = vstack.peek();
    dstack.pop();
    decls = dstack.peek();
  }
  
  /****************************************************************************
   * Attempts to put type into type table.
   * @param name the name of the type
   * @param type the type associated to the name
   * @param decl the declaration of the type
   * @return the constructed type symbol (null, if type table already
   *         contained a symbol with the denoted name)
   ***************************************************************************/
  public TypeSymbol putType(Identifier name, Type type, TypeDeclaration decl)
  {
    TypeSymbol symbol = ttable.get(name);
    if (symbol != null) return null;
    symbol = TypeSymbol.construct(name, type, decl);
    ttable.put(symbol);
    decls.add(decl);
    return symbol;
  }
  
  /****************************************************************************
   * Returns type symbol associated to name.
   * @param name the name of the type.
   * @return the type symbol associated to the name (may be null).
   ***************************************************************************/
  public TypeSymbol getTypeSymbol(Identifier name)
  {
    return ttable.get(name);
  }
  
  /***************************************************************************
   * Get all type symbols.
   * @return collection of all type symbols declared in current scope.
   **************************************************************************/
  public Collection<TypeSymbol> getTypeSymbols()
  {
    return ttable.getSymbols();
  }
  
  /****************************************************************************
   * Attempts to put value into value table
   * @param name the name of the value
   * @param type the type associated to the value
   * @param decl the declaration of the value
   * @return the constructed symbol (null, if value table already contained
   *         a symbol with the denoted name)
   ***************************************************************************/
  public ValueSymbol putValue(Identifier name, Type type, Declaration decl)
  {
    ValueSymbol symbol = vtable.get(name);
    if (symbol != null) return null;
    symbol = ValueSymbol.construct(name, type, decl);
    vtable.put(symbol);
    decls.add(decl);
    return symbol;
  }
  
  /****************************************************************************
   * Returns value symbol associated to name.
   * @param name the name of the value.
   * @return the value symbol associated to the name (may be null).
   ***************************************************************************/
  public ValueSymbol getValueSymbol(Identifier name)
  {
    int i = vstack.size()-1;
    SymbolTable<ValueSymbol> table = vtable;
    while (true)
    {
      ValueSymbol symbol = table.get(name);
      if (symbol != null) return symbol;
      i = i-1;
      if (i < 0) return null;
      table = vstack.elementAt(i);
    }
  }
  
  /***************************************************************************
   * Get all value symbols.
   * @return collection of all value symbols declared in current scope.
   **************************************************************************/
  public Collection<ValueSymbol> getValueSymbols()
  {
    return vtable.getSymbols();
  }
  
  /****************************************************************************
   * Attempts to put formula into value table
   * @param name the name of the formula
   * @param formula the formula itself
   * @param decl the formula definition.
   * @param theory the theory in which the formula occurs (may be null)
   * @return the constructed symbol (null, if formula table already contained
   *         a symbol with the denoted name)
   ***************************************************************************/
  public FormulaSymbol putFormula(Identifier name, 
    Formula formula, FormulaDefinition decl, TheorySymbol theory)
  {
    FormulaSymbol symbol = ftable.get(name);
    if (symbol != null) return null;
    if (decl instanceof AxiomDefinition)
      symbol = AxiomSymbol.construct(name, formula, (AxiomDefinition)decl, theory);
    else
      symbol = FormulaSymbol.construct(name, formula, decl, theory);
    ftable.put(symbol);
    decls.add(decl);
    return symbol;
  }
  
  /****************************************************************************
   * Returns formula symbol associated to name.
   * @param name the name of the formula
   * @return the formula symbol associated to the name (may be null).
   ***************************************************************************/
  public FormulaSymbol getFormulaSymbol(Identifier name)
  {
    return ftable.get(name);
  }
  
  /***************************************************************************
   * Get all formula symbols.
   * @return collection of all formula symbols declared in current scope.
   **************************************************************************/
  public Collection<FormulaSymbol> getFormulaSymbols()
  {
    return ftable.getSymbols();
  }
  
  /***************************************************************************
   * Get sequence of ordered declarations sequences where declarations in a
   * later sequence supersede the previous declarations.
   * @return the sequence of declaration sequences.
   **************************************************************************/
  public Declaration[][] getDeclarations()
  {
    int n = dstack.size();
    Declaration[][] d = new Declaration[n][];
    for (int i=0; i<n; i++)
    {
      Vector<Declaration> ds = dstack.get(i);
      int n0 = ds.size();
      Declaration[] d0 = new Declaration[n0];
      for (int j=0; j<n0; j++)
        d0[j] = ds.elementAt(j);
      d[i] = d0;
    }
    return d;
  }
  
  /****************************************************************************
   * Get theory denoted by name in current environment, also checks
   * if theory may be actually referenced.
   * @param name the name of the theory.
   * @return the theory symbolr respectively null, if an error occurred.
   ***************************************************************************/
  private TheorySymbol getTheory(Name name)
  {
    TheorySymbol tsymbol = getTheoryCore(name);
    if (tsymbol == null) return null;
    if (!theories.contains(tsymbol))
    {
      error.newError(name.getBase(), "theory " + name + 
          " is referenced but not listed in uses clause");
      return null;
    }
    if (!tsymbol.isRead())
    {
      boolean okay = tsymbol.read(error);
      if (!okay) return null;
    }
    return tsymbol;
  }
  
  /****************************************************************************
   * Get theory denoted by identifier in current environment.
   * @param ident the name of the theory.
   * @return the theory symbol respectively null, if the theory was not found.
   ***************************************************************************/
  private TheorySymbol getTheory(Identifier ident)
  {
    if (theories == null) Main.internalError("no theory: " + ident + " " + ident.getSymbol());
    String key = ident.getString();
    for (TheorySymbol tsymbol : theories)
    {
      if (tsymbol.getName().getString() == key) return tsymbol;
    }
    return null;
  }
  
  /****************************************************************************
   * Get theory denoted by name in current environment.
   * @param name the name of the theory.
   * @return the theory symbol respectively null, if an error occurred.
   ***************************************************************************/
  private TheorySymbol getTheoryCore(Name name)
  {
    Identifier[] idents = name.idents;
    int n = idents.length;
    if (n == 0) Main.internalError("empty name encountered");
    Identifier ident = idents[0];
    
    // name is plain identifier
    if (n == 1) {
      TheorySymbol symbol = getTheory(ident);
      if (symbol != null) return symbol;
      error.newError(ident, "theory " + ident + " not listed in uses clause");
      return null;
    }
    
    // name prefix must denote a package
    Name prefix = name.getPrefix();
    Identifier ident0 = name.getBase();
    PackageSymbol psymbol = top.getPackage(prefix);
    if (psymbol == null) return null;
    
    // lookup theory in package
    TheorySymbol tsymbol = psymbol.getTheories().get(ident0);
    if (tsymbol == null) 
    {
      error.newError(ident0,
          "in package " + prefix + " there is no theory " + ident0);
      return null;
    }
    return tsymbol;
  }
  
  /****************************************************************************
   * Get value denoted by reference in current environment
   * @param ref the reference
   * @return the value symbol respectively null, if an error occurred.
   **************************************************************************/
  public ValueSymbol getValueSymbol(Reference ref)
  {
    // the general form of a value reference is
    // [(Package.)*Theory.](Value.*)Value
    // pos always denotes position of next identifier to be processed
    Name name = ref.getName();
    int n = name.getLength();
    int pos = 0; 
    
    // look for first value symbol
    Identifier first = name.idents[pos];
    ValueSymbol symbol = getValueSymbol(first);
    if (symbol != null)
    {
      pos = 1;
    }
    else
    {
      TheorySymbol tsymbol = getTheory(first);
      if (tsymbol != null)
      {
        pos = 1;
      }
      else
      {
        PackageSymbol psymbol = top.getPackage(Name.construct(first), false);
        if (psymbol == null)
        {
          error.newError("there is no logic value named " + first);
          return null;
        }
        pos = 1;
        while (pos < n-2) // at least "Theory.Value" must follow
        {
          Identifier ident = name.idents[pos];
          PackageSymbol psymbol0 = psymbol.getPackage(Name.construct(ident), false);
          if (psymbol0 == null) break;
          psymbol = psymbol0;
          pos = pos+1;
        }
        SymbolTable<TheorySymbol> ttable = psymbol.getTheories();
        tsymbol = ttable.get(name.idents[pos]);
        if (tsymbol == null)
        {
          error.newError("there is no theory named " + name.idents[pos]);
          return null;
        }
        pos = pos+1;
      }
      SymbolTable<ValueSymbol> vtable = tsymbol.getValues();
      Identifier ident = name.idents[pos];
      symbol = vtable.get(ident);
      if (symbol == null)
      {
        error.newError("there is no value named " + ident);
        return null;
      }
      pos = pos+1;
    }
    
    // first value symbol encountered, process record selectors
    while (pos < n)
    {
      Identifier ident = name.idents[pos];
      Type type = symbol.getType();
      if (!(type instanceof RecordType))
      {
        error.newError("value " + symbol.getName() + " is not a record");
        return null;
      }
      RecordType type0 = (RecordType)type;
      TypedIdentifier[] tids = type0.getBase();
      ValueSymbol symbol0 = null;
      for (TypedIdentifier tid : tids)
      {
        Identifier ident0 = tid.getName();
        if (ident.getString() == ident0.getString()) 
        {
          symbol0 = (ValueSymbol)ident0.getSymbol();
          ident.setSymbol(symbol0);
          break;
        }
      }
      if (symbol0 == null)
      {
        error.newError("there is no value named " + ident);
        return null;
      }
      symbol = symbol0;
      pos = pos+1;
    }
    return symbol;
  }
  
  /****************************************************************************
   * Get type denoted by reference in current environment
   * @param ref the reference
   * @return the type symbol respectively null, if an error occurred.
   **************************************************************************/
  public TypeSymbol getTypeSymbol(Reference ref)
  {
    Name name = ref.getName();
    Identifier ident = name.getBase();
    int n = name.getLength();
    TypeSymbol symbol;
    if (n == 1)
      symbol = getTypeSymbol(ident);
    else
    {
      Name prefix = name.getPrefix();
      TheorySymbol theory = getTheory(prefix);
      if (theory == null) return null;
      symbol = theory.getTypes().get(ident);
    }
    if (symbol == null)
    {
      error.newError(ident, "there is no type named " + ref);
      return null;
    }
    return symbol;
  }
}
