// ----------------------------------------------------------------------------
// NewVariables.java
// Create new variables that are not used variables in an expression.
// $Id: NewVariables.java,v 1.4 2011/05/18 14:05:54 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is bound software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;
import fmrisc.ProgramExplorer.*;

/**
 * 
 * Create new variables that are not used variables in an expression.
 *
 */
public final class NewVariables
{  
  /***************************************************************************
   * Compute unique and unused variable names.
   * @param vars a sequence of variables that are candidates for the result.
   * @param suffix a suffix to be added to variable name.
   * @param used a collection of already used (and thus forbidden) names
   *        (may be null, then denotes the empty set).
   * @return an array that is as long as vars and holds at each position
   *         either the corresponding string of vars or a string that
   *         is derived from that string. It is ensured that no string in 
   *         the result occurs in "used" or twice in the result.
   *************************************************************************/
  public static String[] newVariables(VariableSymbol[] vars, String suffix,
    Collection<String> used)
  {
    Collection<String> used2 = new SetClass<String>(used);
    int n = vars.length;
    String[] result = new String[n];
    for (int i=0; i<n; i++)
    {
      String var = vars[i].getName().getString()+suffix;
      String var2 = newVariable(var, used2);
      result[i] = var2;
      used2.add(var2);
    }
    return result;
  }

  /***************************************************************************
   * Compute a new variable name derived from var that does not occur in used.
   * @param var the name from which the result is derived.
   * @param used a set of forbidden name.
   * @return a name that is derived from var and does not occur in used.
   **************************************************************************/
  public static String newVariable(String var, Collection<String> used)
  {
    if (!(used.contains(var))) return var;
    int pos = var.lastIndexOf("_");
    if (pos != -1) var = var.substring(0, pos);
    if (!(used.contains(var))) return var;
    var = var + "_";
    int ext = 0;
    while (true)
    {
      String result = var + ext;
      if (!used.contains(result)) return result;
      ext = ext+1;
    }     
  }
}
