// ----------------------------------------------------------------------------
// ParamSymbol.java
// A symbol denoting a parameterized entity.
// $Id: ParamSymbol.java,v 1.30 2011/08/02 21:03:14 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Tasks.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Judgements.*;

/**
 * 
 * A symbol denoting a parameterized entity.
 *
 */
public abstract class ParamSymbol 
extends GlobalSymbolBase implements GlobalSymbol
{
  private final ClassSymbol csymbol;     // the class holding the value
  private final TypeExpression[] atypes; // the argument types
  
  // the set of parameters and the corresponding (ordered array)
  private SymbolTable<ParameterSymbol> params; 
  private final ParameterSymbol[] parray;
  
  // the various task folders of this symbol
  private TaskFolder tasks = null;         // the tasks of this symbol
  private TaskFolder tccs = null;          // the type checking conditions
  private TaskFolder loops = null;         // the loop conditions
  private TaskFolder preconditions = null; // the method preconditions
  private TaskFolder optional = null;      // the optional conditions
  private TaskFolder obligations = null;   // the other proof obligations
  
  // the method specification (may be null)
  private MethodSpec spec = null;
  
  // the body statement judgement 
  private StatJudgement bjudge = null;
  
  // the precondition annotation
  private Map<Statement, Formula> preMap = null;
  
  // the judgement processor
  private Satisfies satisfies = null;
  
  // the set of methods in the call set of this method
  private Set<ParamSymbol> call = null;
  
  // the set of methods in the recursion set of this method
  private Set<ParamSymbol> recursion = null;
  
  /***************************************************************************
   * Construct value symbol from declaration.
   * Attention: may fail on parameters with duplicate names; in this case, 
   * getParams() returns null.
   * @param csymbol the symbol of the class holding the declaration.
   * @param name the name of the symbol.
   * @param params the parameters.
   * @param atypes the types of its arguments.
   * @param error the stream to be used for printing errors.
   **************************************************************************/
  protected ParamSymbol(ClassSymbol csymbol, Identifier name,
    Parameter[] params, TypeExpression[] atypes, ErrorStream error)
  {
    super(csymbol, name);
    this.csymbol = csymbol;
    this.atypes = atypes;
    this.params = new SymbolTable<ParameterSymbol>(error);
    int n = params.length;
    parray = new ParameterSymbol[n];
    for (int i=0; i<n; i++)
    {
      ParameterSymbol param = 
        ParameterSymbol.construct(this, params[i], atypes[i]);
      boolean okay = this.params.put(param);
      if (!okay) 
      {
        this.params = null;
        return;
      }
      parray[i] = param;
    }
    this.tasks = new TaskFolder(this, "method " + name, csymbol.getTasks(), false);
    this.obligations = this.tasks; // new TaskFolder("correctness conditions", this.tasks);
    this.preconditions = new TaskFolder(null, "preconditions", this.tasks);
    this.loops = new TaskFolder(null, "loops", this.tasks);
    this.tccs = new TaskFolder(null, "type checking conditions", this.tasks);
    this.optional = new TaskFolder(null, "specification validation (optional)", this.tasks);
  }
  
  /***************************************************************************
   * Get the method's parameter set.
   * @return the set of parameters.
   **************************************************************************/
  public SymbolTable<ParameterSymbol> getParams()
  {
    return params;
  }
  
  /****************************************************************************
   * Get the class to which the entity belongs.
   * @return the class of the symbol
   ***************************************************************************/
  public ClassSymbol getClassSymbol()
  {
    return csymbol;
  }
  
  /****************************************************************************
   * Get the argument types of the entity
   * @return the argument types.
   ***************************************************************************/
  public TypeExpression[] getArgumentTypes()
  {
    return atypes;
  }
  
  /****************************************************************************
   * Get array of parameters of this object.
   * @return the array of parameters.
   ***************************************************************************/
  public ParameterSymbol[] getParameterArray()
  {
    return parray;
  }
  
  /***************************************************************************
   * Get declaration associated to symbol (may be null).
   * @return the declaration.
   **************************************************************************/
  public abstract ParamDeclaration getDeclaration();
  
  /***************************************************************************
   * Get the type checking conditions.
   * @return the folder holding the type checking conditions.
   **************************************************************************/
  public TaskFolder getTCCs()
  {
    return tccs;
  }
  
  /***************************************************************************
   * Get the loops.
   * @return the folder holding the loop conditions.
   **************************************************************************/
  public TaskFolder getLoops()
  {
    return loops;
  }
  
  /***************************************************************************
   * Get the preconditions.
   * @return the folder holding the preconditions.
   **************************************************************************/
  public TaskFolder getPreconditions()
  {
    return preconditions;
  }
  
  /***************************************************************************
   * Get the optional conditions.
   * @return the folder holding the optional conditions.
   **************************************************************************/
  public TaskFolder getOptional()
  {
    return optional;
  }
  
  /***************************************************************************
   * Get the other proof obligations.
   * @return the folder holding the other proof obligations.
   **************************************************************************/
  public TaskFolder getObligations()
  {
    return obligations;
  }
  
  /****************************************************************************
   * Reset symbol
   ***************************************************************************/
  public void reset()
  {
    tccs.clear();
    obligations.clear();
    loops.clear();
    preconditions.clear();
    optional.clear();
  }
  
  /****************************************************************************
   * Set method specification.
   * @param spec the specification
   ***************************************************************************/
  public void setSpecification(MethodSpec spec)
  {
    this.spec = spec;
  }
  
  /****************************************************************************
   * Get method specification.
   * @return the specification
   ***************************************************************************/
  public MethodSpec getSpecification()
  {
    return spec;
  }
  
  /****************************************************************************
   * Set body statement judgement.
   * @param bjudge the body statement judgement.
   ***************************************************************************/
  public void setBodyJudgement(StatJudgement bjudge)
  {
    this.bjudge = bjudge;
  }
  
  /****************************************************************************
   * Get body statement judgement.
   * @return the body statement judgement.
   ***************************************************************************/
  public StatJudgement getBodyJudgement()
  {
    return bjudge;
  }
  
  /****************************************************************************
   * Set mapping of body statements to prestate knowledge.
   * @param preMap the mapping (may be null).
   ***************************************************************************/
  public void setPreMap(Map<Statement, Formula> preMap)
  {
    this.preMap = preMap;
  }
  
  /****************************************************************************
   * Get mapping of body statements to prestate knowledge.
   * @return the mapping (may be null).
   ***************************************************************************/
  public Map<Statement, Formula> getPreMap()
  {
    return preMap;
  }
  
  /***************************************************************************
   * Set judgement processor for this method.
   * @param satisfies the judgement processor.
   **************************************************************************/
  public void setSatisfies(Satisfies satisfies)
  {
    this.satisfies = satisfies;
  }
  
  /***************************************************************************
   * Get judgement processor for this method.
   * @return the judgement processor (may be null)
   **************************************************************************/
  public Satisfies getSatisfies()
  {
    return satisfies;
  }
  
  /**************************************************************************
   * Set the set of methods directly called by this method.
   * @param call the call set.
   *************************************************************************/
  public void setCall(Set<ParamSymbol> call)
  {
    this.call = call;
  }
  
  /**************************************************************************
   * Get the set of methods called by this method.
   * @return recursion the call set.
   *************************************************************************/
  public Set<ParamSymbol> getCall()
  {
    return call;
  }
  
  /**************************************************************************
   * Set the set of methods (directly or indirectly) recursively called
   * by this method.
   * @param recursion the recursion set.
   *************************************************************************/
  public void setRecursion(Set<ParamSymbol> recursion)
  {
    this.recursion = recursion;
  }
  
  /**************************************************************************
   * Get the set of methods (directly or indirectly) recursively called
   * by this method.
   * @return recursion the recursion set.
   *************************************************************************/
  public Set<ParamSymbol> getRecursion()
  {
    return recursion;
  }
}