// ----------------------------------------------------------------------------
// SymbolTableBase.java
// A symbol table.
// $Id: SymbolTable.java,v 1.15 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.InOut.*;

/**
 * 
 * A symbol table.
 *
 */
public final class SymbolTable<S extends Symbol> 
{
  // our symbol collection
  private final MapClass<String,S> table = new MapClass<String,S>();
  
  // the stream to print messages on
  private ErrorStream error;
  
  /****************************************************************************
   * Construct symbol table with denoted error stream.
   * @param error the error stream.
   ***************************************************************************/
  public SymbolTable(ErrorStream error)
  {
    this.error = error;
  }
  
  /****************************************************************************
   * Return collection of symbols in table.
   * @return the collection of all symbols.
   ***************************************************************************/
  public Collection<S> getSymbols()
  {
    return table.values();
  }
  
  /****************************************************************************
   * Puts symbol into table to be looked up for the name of the symbol.
   * @param value the symbol.
   * @return true, if symbol could be added; 
   *         false, if symbol with same name was already in table
   *                (in this case the symbol is not added)
   ***************************************************************************/
  public boolean put(S value)
  {
    Identifier key = value.getName();
    if (get(key) != null) 
    {
      error.newError(key, "symbol " + key + 
          " is already defined in current environment");
      return false;
    }
    table.put(key.getString(), value);
    return true;
  }
  
  /****************************************************************************
   * Returns symbol associated to key and links key with symbol.
   * @param key the lookup key.
   * @return the associated symbol (null, if there is none).
   ***************************************************************************/
  public S get(Identifier key)
  {
    S symbol = table.get(key.getString());
    if (symbol != null) key.setSymbol(symbol);
    return symbol;
  }
  
  /****************************************************************************
   * Checks whether table has an entry for key.
   * @param key the lookup key.
   * @return true iff there is an entry with this key.
   ***************************************************************************/
  public boolean has(Identifier key)
  {
    return (table.get(key.getString()) != null);
  }
  
  /****************************************************************************
   * Add symbols from another table to this table.
   * @param stable the table providing new symbols for this table.
   * @param override if true, new declarations override old ones.
   * @param skip if true (and there is no overriding), conflicting symbols
   *        are skipped without reporting errors.
   * @return true if no conflicts with errors occurred.
   **************************************************************************/
  public boolean add(SymbolTable<S> stable, boolean override, boolean skip)
  {
    Collection<S> values = stable.table.values();
    for (S s : values)
    {
      Identifier key = s.getName();
      if (override) table.remove(key.getString());
      else if (skip && has(key)) continue;
      boolean okay = put(s);
      if (!okay) return false;
    }
    return true;
  }
}