// ----------------------------------------------------------------------------
// TheorySymbol.java
// A symbol denoting a theory.
// $Id: TheorySymbol.java,v 1.24 2011/07/05 09:26:19 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Tasks.*;

import java.io.*;
import java.util.*;

/**
 * 
 * A symbol denoting a theory.
 *
 */
public final class TheorySymbol extends GlobalSymbolBase 
{
  // the package holding the theory
  private final PackageSymbol psymbol; 
  
  // the reader for reading the source file (may be null)
  private TheoryReader reader;
  
  // the class from which to derive the theory declaration (may be null)
  private ClassSymbol csymbol;
  
  // the declaration of the theory after reading/deriving (null before)
  private TheoryDeclaration decl;
  
  // indicate whether theory has errors
  private boolean errors = false;
  
  // the symbols of the theory
  private SymbolTable<TypeSymbol> types = null;
  private SymbolTable<ValueSymbol> values = null;
  private SymbolTable<FormulaSymbol> formulas = null;
    
  // the various task folders of this theory
  private TaskFolder tasks = null;     // the tasks of this theory
  private TaskFolder tccs = null;      // the type checking conditions
  private TaskFolder tformulas = null; // the formulas to be proved
  
  // the symbols of the units (classes/theories) referenced by this theory
  private Collection<Symbol> symbols = null;
  
  // the class for which this theory is a local theory (may be null)
  private ClassSymbol hostClass;
  
  /***************************************************************************
   * Construct theory symbol from reader.
   * @param psymbol the symbol of the package holding the theory.
   * @param name the name of the theory.
   * @param reader the reader for getting the declaration.
   **************************************************************************/
  public TheorySymbol(PackageSymbol psymbol, Identifier name, TheoryReader reader)
  {
    super(psymbol, name);
    this.psymbol = psymbol;
    this.reader = reader;
    
    resetTasks();
    
    // process declaration when it is read from source
  }

  /***************************************************************************
   * Reset set of tasks related to class.
   **************************************************************************/
  private void resetTasks()
  {
    if (tasks != null)
    {
      TaskFolder parent = tasks.getParent();
      parent.remove(tasks);
      tasks.setParent(null);
      tasks.setStatus(Task.Status.REVOKED, null, null);
    }
    tasks = new TaskFolder(this, "theory " + getName(), psymbol.getTasks());
    tccs = new TaskFolder(null, "type checking conditions", this.tasks);
    this.tformulas = new TaskFolder(null, "formulas to be proved", this.tasks);
  }
  
  /***************************************************************************
   * Construct theory symbol for class theory for later processing
   * (no declaration is provided yet).
   * @param csymbol the symbol of the class.
   **************************************************************************/
  public TheorySymbol(ClassSymbol csymbol)
  {
    // theory is named like class
    super(csymbol.getPackage(), 
        Identifier.construct(csymbol.getName().getString()));
    this.psymbol = csymbol.getPackage();
    this.csymbol = csymbol;
    csymbol.setClassTheory(this); 
    
    this.tasks = new TaskFolder(this, "theory " + csymbol.getName(), 
      psymbol.getTasks());
    this.tccs = new TaskFolder(null, "type checking conditions", this.tasks);
    this.tformulas = new TaskFolder(null, "formulas to be proved", this.tasks);
    
    // process declaration when it is derived from class
  }
  
  /***************************************************************************
   * Construct theory symbol from declaration.
   * @param psymbol the symbol of the package holding the theory.
   * @param name the name of the theory.
   * @param decl the theory declaration.
   * @param error the error stream to print messages to
   **************************************************************************/
  public TheorySymbol(PackageSymbol psymbol, Identifier name, 
    TheoryDeclaration decl, ErrorStream error)
  {
    super(psymbol, name);
    this.psymbol = psymbol;
    this.decl = decl;

    this.tasks = new TaskFolder(this, "theory " + name, psymbol.getTasks());
    this.tccs = new TaskFolder(null, "type checking conditions", this.tasks);
    this.tformulas = new TaskFolder(null, "formulas to be proved", this.tasks);
    
    // process declaration now
    processDeclaration(error);
  }
  
  /***************************************************************************
   * Construct local theory symbol from declaration.
   * @param csymbol the symbol of the class holding the theory.
   * @param name the name of the theory.
   * @param decl the theory declaration.
   * @param error the error stream to print messages to
   **************************************************************************/
  public TheorySymbol(ClassSymbol csymbol, Identifier name, 
    TheoryDeclaration decl, ErrorStream error)
  {
    super(csymbol, name);
    this.psymbol = csymbol.getPackage();
    this.decl = decl;
    this.hostClass = csymbol;
    
    this.tasks = new TaskFolder(this, "theory " + name, csymbol.getTasks());
    this.tccs = new TaskFolder(null, "type checking conditions", this.tasks);
    this.tformulas = new TaskFolder(null, "formulas to be proved", this.tasks);
    
    // process declaration now
    processDeclaration(error);
  }
  
  /****************************************************************************
   * Get package of theory.
   * @return the theory's package
   ***************************************************************************/
  public PackageSymbol getPackage()
  {
    return psymbol;
  }
  
  /****************************************************************************
   * Get source of theory (respectively of associated class).
   * @return the theory's source.
   ***************************************************************************/
  public Source getSource()
  {
    if (reader == null) 
    {
      if (csymbol == null) return null;
      return csymbol.getSource();
    }
    return reader.getSource();
  }
  
  /****************************************************************************
   * Get type table of theory.
   * @return the theory's type table.
   ***************************************************************************/
  public TypeTable getTypeTable()
  {
    return psymbol.getTypeTable();
  }
  
  /****************************************************************************
   * Returns true if theory has been read, i.e. its values are defined.
   * @return true if theory has been read.
   ***************************************************************************/
  public boolean isRead()
  {
    return decl != null;
  }
  
  /****************************************************************************
   * Returns true if theory is derived from class.
   * @return true if theory is derived from class.
   ***************************************************************************/
  public boolean isFromClass()
  {
    return csymbol != null;
  }
  
  /****************************************************************************
   * Returns class from which theory is derived (if any).
   * @return the class symbol (may be null).
   ***************************************************************************/
  public ClassSymbol getClassSymbol()
  {
    return csymbol;
  }
  
  // prevent recursion
  private int constructing = 0;
  
  /****************************************************************************
   * Update declaration from reader or derive it from class symbol.
   * @param error the error stream on which to print any messages.
   * @return true if everything went fine, false if an error occurred.
   ***************************************************************************/
  public boolean read(ErrorStream error)
  {
    symbols = null;
    if (csymbol != null)
    {
      if (constructing > 1)
      {
        // recursive type dependency encountered
        error.newError("type " + getLongName() + 
            " depends (directly or indirectly) on itself");
        errors = true;
        return false;
      }
      TypeTranslator translator = csymbol.getPackage().getTypeTranslator();
      errors = false;
      constructing++;
      decl = translator.constructClassTheory(csymbol);
      constructing--;
      if (decl == null) return false;
    }
    else
    {
      if (reader == null) return false;
      if (decl != null && !reader.isModified() && !hasErrors()) return true;
      errors = false;
      decl = reader.read();
    }
    if (decl == null) 
    {
      errors = true;
      return false;
    }
    return processDeclaration(error);
  }
 
  /***************************************************************************
   * Process theory declaration
   * @param error the error stream on which to print any messages.
   * @param return true if everything went fine, false, if an error occurred.
   **************************************************************************/
  private boolean processDeclaration(ErrorStream error)
  {
    init(error);
    boolean okay = LogicChecking.process(this, error);
    // immediately compute dependencies (may be partial)
    symbols = ReferencedUnits.compute(decl);
    symbols.remove(this); // avoid circular dependency
    // do not forget referenced class symbol
    if (csymbol != null) symbols.add(csymbol);
    if (!okay)
    {
      errors = true;
      reset();
      return false;
    } 
    return true;
  }
  
  /***************************************************************************
   * Set and process new theory declaration
   * @param decl the new theory declaration (updating any previous one)
   * @param error the error stream on which to print any messages.
   **************************************************************************/
  public void setDeclaration(TheoryDeclaration decl, ErrorStream error)
  {
    this.decl = decl;
    processDeclaration(error);
  }
  
  /***************************************************************************
   * Show whether class has errors.
   * @return true if class has errors
   **************************************************************************/
  public boolean hasErrors()
  {
    return errors;
  }
  
  /***************************************************************************
   * Set error status.
   * @param status true iff errors, false otherwise.
   **************************************************************************/
  public void setErrors(boolean status)
  {
    errors = status;
  }
  
  /***************************************************************************
   * Initialize symbol tables.
   * @param error the error stream to print messages to.
   **************************************************************************/
  private void init(ErrorStream error)
  {
    types = new SymbolTable<TypeSymbol>(error);
    values = new SymbolTable<ValueSymbol>(error);
    formulas = new SymbolTable<FormulaSymbol>(error);
  }
  
  /***************************************************************************
   * Reset symbol.
   **************************************************************************/
  public void reset()
  {
    decl = null;
    types = null;
    values = null;
    formulas = null;
    symbols = null;
    resetTasks();
  }
  
  /***************************************************************************
   * Get the declaration of the theory.
   * @return the declaration (may be null)
   **************************************************************************/
  public TheoryDeclaration getDeclaration()
  {
    return decl;
  }

  /****************************************************************************
   * Get set of types of this theory.
   * @return the set of types.
   ***************************************************************************/
  public SymbolTable<TypeSymbol> getTypes()
  {
    return types;
  }
  
  /****************************************************************************
   * Get set of values of this theory.
   * @return the set of types.
   ***************************************************************************/
  public SymbolTable<ValueSymbol> getValues()
  {
    return values;
  }
  
  /****************************************************************************
   * Get set of formulas of this class.
   * @return the set of types.
   ***************************************************************************/
  public SymbolTable<FormulaSymbol> getFormulas()
  {
    return formulas;
  }
  
  /****************************************************************************
   * Get type table of class.
   * @return the class's type table.
   ***************************************************************************/
  public LogicTypeTable getLogicTypeTable()
  {
    return psymbol.getLogicTypeTable();
  }
  
  /***************************************************************************
   * Print symbol on output stream.
   * @param out the output stream.
   **************************************************************************/
  public void print(PrintWriter out)
  {
    out.print("theory ");
    getLongName().print(out);
    Source source = getSource();
    if (source == null) return;
    out.print(" (file ");
    out.print(source.getFile().getAbsolutePath());
    out.print(")");
  }
  
  /***************************************************************************
   * Get the type checking conditions.
   * @return the folder holding the type checking conditions.
   **************************************************************************/
  public TaskFolder getTCCs()
  {
    return tccs;
  }
  
  /***************************************************************************
   * Get the formula tasks.
   * @return the folder holding the formulas to be proved.
   **************************************************************************/
  public TaskFolder getFormulaTasks()
  {
    return tformulas;
  }
  
  /***************************************************************************
   * Get symbols of units referenced by this unit.
   * @return the collection of symbols.
   **************************************************************************/
  public Collection<Symbol> getReferencedUnits()
  {
    return symbols;
  }
  
  /***************************************************************************
   * Add symbol to collection of symbols (if it is not there).
   * @param symbols the symbol collection to which to add.
   * @param symbol the symbol to be added.
   **************************************************************************/
  public static void add(Collection<TheorySymbol> symbols, TheorySymbol symbol)
  {
    if (symbols.contains(symbol)) return;
    symbols.add(symbol);
  }
      
  /***************************************************************************
   * Add symbols to collection of symbols (if it is not there).
   * @param symbols the symbol collection to which to add.
   * @param symbols0 the symbols to be added.
   **************************************************************************/
  public static void 
  add(Collection<TheorySymbol> symbols, Collection<TheorySymbol> symbols0)
  {
    for (TheorySymbol symbol : symbols0)
      add(symbols, symbol);
  } 
  
  /***************************************************************************
   * Get the class for which this theory is local.
   * @return the class (null, if the theory is not local).
   **************************************************************************/
  public ClassSymbol getHostClass()
  {
    return hostClass;
  }
}

