// ----------------------------------------------------------------------------
// TypeTable.java
// Make type expressions unique such that they can be compared by reference.
// $Id: TypeTable.java,v 1.12 2012/10/15 15:16:48 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;
import java.util.*;

/**
 * 
 * Make type expressions unique such that they can be compared by reference.
 *
 */
public class TypeTable
{
  private Vector<TypeExpression>   utype;  // table of unique types
  private Vector<TypeExpression[]> utypes; // table of unique type sequences
  
  // the unique primitive types
  public final TypeExpression booleanType;
  public final TypeExpression charType;
  public final TypeExpression intType;
  public final TypeExpression voidType;
  public final TypeExpression nullType;
  public final TypeExpression exceptionType;
  
  // string type can be replaced by class
  private ClassSymbol stringClass = null;
  private TypeExpression stringType = null;
  
  /****************************************************************************
   * Construct a table for making types unique.
   ***************************************************************************/
  public TypeTable()
  {
    utype = new Vector<TypeExpression>();
    utypes = new Vector<TypeExpression[]>();
    booleanType = BooleanType.construct();
    charType = CharType.construct();
    intType = IntType.construct();
    stringType = StringType.construct();
    voidType = VoidType.construct();
    nullType = NullType.construct();
    exceptionType = ExceptionType.construct();
    utype.add(booleanType);
    utype.add(charType);
    utype.add(intType);
    utype.add(stringType);
    utype.add(voidType);
    utype.add(nullType); 
    utype.add(exceptionType);
  }
  
  /****************************************************************************
   * Get string type.
   * @return the string type.
   ***************************************************************************/
  public TypeExpression getStringType()
  {
    return stringType;
  }
  
  /****************************************************************************
   * Get string class.
   * @return the string class (may be null, if none).
   ***************************************************************************/
  public ClassSymbol getStringClass()
  {
    return stringClass;
  }
  
  /****************************************************************************
   * Set string type to class.
   * @param symbol the symbol denoting the class (null, if none).
   ***************************************************************************/
  public void setStringType(ClassSymbol symbol)
  {
    stringClass = symbol;
    if (symbol == null)
      stringType = StringType.construct();
    else
      stringType = process(NamedType.construct(symbol.getLongName()));
  }
  
  /***************************************************************************
   * Return unique version of a type.
   * @param type a type.
   * @return the unique version of the type
   *         (i.e. equality can be checked by pointer comparison)
   *         or null if an error occurred.
   **************************************************************************/
  public TypeExpression process(TypeExpression type)
  {
    // extract information and make base of array type unique
    TypeExpression type0 = type;
    TypeExpression b = null;
    ClassSymbol s = null;
    if (type instanceof NamedType)
    {
      s = ((NamedType)type).getSymbol();
      if (s == null)
        Main.internalError("named type " + type + " has no symbol");
    }
    else if (type instanceof ArrayType) 
    {
      b = process(((ArrayType)type).base);
      if (b == null) return null;
      type0 = ArrayType.construct(b);
    }
    else if (type instanceof NewType) 
    {
      b = process(((NewType)type).base);
      if (b == null) return null;
      type0 = NewType.construct(b);
    }
    
    // replace string type
    if (type instanceof StringType)
    {
      // annotate type with string class (may be null)
      StringType stype = (StringType)type;
      stype.setStringClass(stringClass);
      return stringType;
    }
    
    // search for type
    Class<? extends TypeExpression> c = type0.getClass();
    for (TypeExpression t: utype)
    {
      if (c != t.getClass()) continue;
      if (t instanceof PrimitiveType) return t;
      if (t instanceof NamedType)
      {
        NamedType t0 = (NamedType)t;
        ClassSymbol s0 = t0.getSymbol();
        if (s == s0) return t;
        continue;
      }
      if (t instanceof ArrayType)
      {
        TypeExpression b0 = ((ArrayType)t).base;
        if (b == b0) return t;
        continue;
      }
      if (t instanceof NewType)
      {
        TypeExpression b0 = ((NewType)t).base;
        if (b == b0) return t;
        continue;
      }
    }
    if (type0 instanceof NamedType)
    {
      NamedType t0 = (NamedType)type0;
      ClassSymbol s0 = t0.getSymbol();
      type0 = s0.getType();
    }
    utype.add(type0);
    return type0;
  }
  
  /***************************************************************************
   * Return unique version of a sequence of types.
   * @param types a sequence of types.
   * @return the unique version of these types
   *         (i.e. equality can be checked by pointer comparison)
   *         or null if an error occurred.
   **************************************************************************/
  public TypeExpression[] process(TypeExpression[] types)
  {
    // make type sequence unique
    int n = types.length;
    TypeExpression[] types0 = new TypeExpression[n];
    for (int i=0; i<n; i++)
    {
      types0[i] = process(types[i]);
      if (types0[i] == null) return null;
    }
    
    // search for type sequence
    search:
      for (TypeExpression[] ts: utypes)
      {
        if (n != ts.length) continue;
        for (int i=0; i<n; i++)
        {
          if (types0[i] != ts[i]) continue search;
        }
        return ts;
      }
    utypes.add(types0);
    return types0;
  }
}
