// ----------------------------------------------------------------------------
// TypeTranslator.java
// Translate program types to logical types.
// $Id: TypeTranslator.java,v 1.29 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Syntax.Program.Import;
import fmrisc.ProgramExplorer.Syntax.Program.NamedType;
import fmrisc.ProgramExplorer.Syntax.Program.ObjectVariableDeclaration;
import fmrisc.ProgramExplorer.Syntax.Program.TypeExpression;
import fmrisc.ProgramExplorer.Judgements.*;

/**
 * 
 * Translate program types to logical types.
 *
 */
public class TypeTranslator
{
  private ErrorStream error;                    // the error stream
  private TheorySymbol baseTheory;              // the constructed theory
  
  // special strings 
  public final static String baseTheoryName = "Base"; // the name of the base theory
  
  // names of symbols declared in base theory
  
  // basic types and values
  public final static String minIntName = "MIN_INT";  // lower bound of machine integers
  public final static String maxIntName = "MAX_INT";  // upper bound of machine integers
  public final static String intName = "int";         // type of machine integers
  public final static String natName = "nat";         // type of unsigned machine integers
  public final static String charName = "char";       // type of machine characters
  
  // the corresponding symbols
  private TypeSymbol intSymbol;
  private TypeSymbol natSymbol;
  private TypeSymbol charSymbol;
  
  // machine arithmetic
  public final static String intPlusName =  "intPlus";   // addition 
  public final static String intNegName =   "intNeg";     // negation
  public final static String intMinusName = "intMinus"; // subtraction
  public final static String intTimesName = "intTimes"; // multiplication
  public final static String intDivName =   "intDiv";   // division
  public final static String intDivName0 =  "intDiv0";  // division without overflow
  public final static String intRemName =   "intRem";   // remainder
  
  // corresponding auxiliary functions
  public final static String intPlusOverflowName =  "intPlusOverflow";   // addition overflow
  public final static String intNegOverflowName =   "intNegOverflow";    // negation overflow
  public final static String intMinusOverflowName = "intMinusOverflow";  // subtraction overflow
  public final static String intTimesOverflowName = "intTimesOverflow";  // multiplication overflow
  public final static String intTruncName = "intTrunc";                  // integer truncation
  public final static String intDivOverflowName = "intDivOverflow";      // division overflow
  
  // null values
  public final static String zeroName = "zero";                     // the int 0
  public final static String nullCharName = "nullChar";             // null character
  public final static String nullStringName = "nullString";        // null string
  public final static String newStringName = "newString";          // new literal string
  
  // arrays
  public final static String typeIntArrayName = "IntArray";         // type of array
  public final static String typeCharArrayName = "CharArray";       // type of array
  public final static String typeBooleanArrayName = "BooleanArray"; // type of array
  public final static String typeStringArrayName = "StringArray";   // type of array
  public final static String nullIntArrayName = "nullIntArray";         // null array
  public final static String nullCharArrayName = "nullCharArray";       // null array
  public final static String nullBooleanArrayName = "nullBooleanArray"; // null array
  public final static String nullStringArrayName = "nullStringArray";   // null array
  public final static String newIntArrayName = "newIntArray";         // new array
  public final static String newCharArrayName = "newCharArray";       // new array
  public final static String newBooleanArrayName = "newBooleanArray"; // new array
  public final static String newStringArrayName = "newStringArray";   // new array
  public final static String newIntArrayAxiomName = "newIntArrayAxiom";         // new array axiom
  public final static String newCharArrayAxiomName = "newCharArrayAxiom";       // new array axiom
  public final static String newBooleanArrayAxiomName = "newBooleanArrayAxiom"; // new array axiom
  public final static String newStringArrayAxiomName = "newStringArrayAxiom";   // new array axiom
  
  // special constant names in every class theory
  public final static String typeArrayName = "Array";
  public final static String nullName = "null";
  public final static String newObjectName = "newObject";
  public final static String nullArrayName = "nullArray";
  public final static String newArrayName = "newArray";
  public final static String newArrayAxiomName = "newArrayAxiom";
  
  // array fields
  public final static String arrayLength = "length";
  public final static String arrayValue = "value";
  
  // null field in classes and arrays 
  // (since "null" is keyword, cannot conflict with name of object field)
  public final static String isNullName = "null";
  
  // new field in objects
  // (since "new" is keyword, cannot conflict with name of object field)
  public final static String newName = "new";
  
  // the current string type and the theory containing the string functions
  private Type stringType;
  private TheorySymbol stringTheory;
  
  /****************************************************************************
   * Construct type translator for given root package.
   * Still need to be initialized.
   ***************************************************************************/
  public TypeTranslator()
  {
  }
  
  /****************************************************************************
   * Get the current string type.
   * @return the string type.
   ***************************************************************************/
  public Type getStringType()
  {
    if (stringType == null) Main.internalError("no string type");
    return stringType;
  }
  
  /****************************************************************************
   * Get the current string theory
   * @return the string theory.
   ***************************************************************************/
  public TheorySymbol getStringTheory()
  {
    if (stringTheory == null) Main.internalError("no string theory");
    return stringTheory;
  }
  
  /****************************************************************************
   * Initialize type translator (which creates a base theory).
   * @param top the root package to be used for initialization.
   * @param error the error stream.
   ***************************************************************************/
  public void init (PackageSymbol top, ErrorStream error)
  {
    this.error = error;
    
    // set string type to default values (theory is set at the end)
    stringType = LogicTypeTable.stringType;

    // the bounds of machine integers
    // MIN_INT: INT = -2147483648;
    // MAX_INT: INT =  2147483647;
    ValueDefinition minIntDef =
      ValueDeclarationInitialized.construct(Identifier.construct(minIntName),
          IntType.construct(),
          NegationTerm.construct(IntLiteral.construct("2147483648")));
    ValueDefinition maxIntDef =
      ValueDeclarationInitialized.construct(Identifier.construct(maxIntName),
          IntType.construct(),
          IntLiteral.construct("2147483647"));      
      
    // introduce axioms that may be used without expanding definitions
    // minIntAxiom: AXIOM MIN_INT < 0;
    // maxIntAxiom: AXIOM MAX_INT > 0;
    AxiomDefinition minIntAxiom =
      AxiomDefinition.construct(Identifier.construct("minIntAxiom"),
          LessFormula.construct(
              Reference.construct(Identifier.construct(minIntName)),
              IntLiteral.construct("0")));
    AxiomDefinition maxIntAxiom =
      AxiomDefinition.construct(Identifier.construct("maxIntAxiom"),
          GreaterFormula.construct(
              Reference.construct(Identifier.construct(maxIntName)),
              IntLiteral.construct("0")));
    
    // the type of machine integers
    // int: TYPE = [MIN_INT..MAX_INT] 
    Type intType = SubrangeType.construct(
           Reference.construct(Identifier.construct(minIntName)),
           Reference.construct(Identifier.construct(maxIntName))); 
    TypeDefinition intDef = 
      TypeDefinition.construct(Identifier.construct(intName), intType);
    Type natType = SubrangeType.construct(
        IntLiteral.construct("0"),
        Reference.construct(Identifier.construct(maxIntName))); 
    TypeDefinition natDef = 
     TypeDefinition.construct(Identifier.construct(natName), natType);
 
    // the types of machine characters 
    // char: TYPE;
    TypeDeclaration charDecl = 
      TypeDeclaration.construct(Identifier.construct(charName));
    
    // arithmetic operations: addition
    // intPlus(x,y) = IF MIN_INT <= x+y AND x+y <= MAX_INT 
    //                  THEN x+y ELSE intPlusOverflow(x,y) ENDIF
    ValueDeclaration intPlusOverflowDecl =
      ValueDeclarationClass.construct(Identifier.construct(intPlusOverflowName),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName)),
                Reference.construct(Identifier.construct(intName))
              },
              Reference.construct(Identifier.construct(intName))));
    ValueDeclaration intPlusDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(intPlusName),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName)),
                Reference.construct(Identifier.construct(intName))
              },
              IntType.construct()),
          LambdaTerm.construct(
              new TypedIdentifier[] {
                  TypedIdentifier.construct(
                      Identifier.construct("x"), 
                      Reference.construct(Identifier.construct(intName))),
                  TypedIdentifier.construct(
                      Identifier.construct("y"), 
                      Reference.construct(Identifier.construct(intName)))},
              IfThenElseTerm.construct(
                  AndFormula.construct(
                      LessEqualFormula.construct(
                          Reference.construct(Identifier.construct(minIntName)),
                          PlusTerm.construct(
                              Reference.construct(Identifier.construct("x")), 
                              Reference.construct(Identifier.construct("y")))),
                      LessEqualFormula.construct(
                          PlusTerm.construct(
                              Reference.construct(Identifier.construct("x")), 
                              Reference.construct(Identifier.construct("y"))),
                          Reference.construct(Identifier.construct(maxIntName)))),
                  PlusTerm.construct(
                      Reference.construct(Identifier.construct("x")), 
                      Reference.construct(Identifier.construct("y"))),
                  FunctionApplication.construct(
                      Reference.construct(Identifier.construct(intPlusOverflowName)),
                      new Term[] {
                        Reference.construct(Identifier.construct("x")), 
                        Reference.construct(Identifier.construct("y"))}))));

    // arithmetic operations: negation
    // intNeg(x) = IF MIN_INT <= -x AND -x <= MAX_INT 
    //                  THEN -x ELSE intNegOverflow(x) ENDIF
    ValueDeclaration intNegOverflowDecl =
      ValueDeclarationClass.construct(Identifier.construct(intNegOverflowName),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName))
              },
              Reference.construct(Identifier.construct(intName))));
    ValueDeclaration intNegDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(intNegName),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName))
              },
              IntType.construct()),
          LambdaTerm.construct(
              new TypedIdentifier[] {
                  TypedIdentifier.construct(
                      Identifier.construct("x"), 
                      Reference.construct(Identifier.construct(intName)))},
              IfThenElseTerm.construct(
                  AndFormula.construct(
                      LessEqualFormula.construct(
                          Reference.construct(Identifier.construct(minIntName)),
                          NegationTerm.construct(
                              Reference.construct(Identifier.construct("x")))),
                      LessEqualFormula.construct(
                          NegationTerm.construct(
                              Reference.construct(Identifier.construct("x"))),
                          Reference.construct(Identifier.construct(maxIntName)))),
                  NegationTerm.construct(
                      Reference.construct(Identifier.construct("x"))),
                  FunctionApplication.construct(
                      Reference.construct(Identifier.construct(intNegOverflowName)),
                      new Term[] {
                        Reference.construct(Identifier.construct("x"))}))));
    
    // arithmetic operations: subtraction
    // intMinus(x,y) = IF MIN_INT <= x-y AND x-y <= MAX_INT 
    //                  THEN x-y ELSE intMinusOverflow(x,y) ENDIF
    ValueDeclaration intMinusOverflowDecl =
      ValueDeclarationClass.construct(Identifier.construct(intMinusOverflowName),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName)),
                Reference.construct(Identifier.construct(intName))
              },
              Reference.construct(Identifier.construct(intName))));
    ValueDeclaration intMinusDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(intMinusName),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName)),
                Reference.construct(Identifier.construct(intName))
              },
              IntType.construct()),
          LambdaTerm.construct(
              new TypedIdentifier[] {
                  TypedIdentifier.construct(
                      Identifier.construct("x"), 
                      Reference.construct(Identifier.construct(intName))),
                  TypedIdentifier.construct(
                      Identifier.construct("y"), 
                      Reference.construct(Identifier.construct(intName)))},
              IfThenElseTerm.construct(
                  AndFormula.construct(
                      LessEqualFormula.construct(
                          Reference.construct(Identifier.construct(minIntName)),
                          MinusTerm.construct(
                              Reference.construct(Identifier.construct("x")), 
                              Reference.construct(Identifier.construct("y")))),
                      LessEqualFormula.construct(
                          MinusTerm.construct(
                              Reference.construct(Identifier.construct("x")), 
                              Reference.construct(Identifier.construct("y"))),
                          Reference.construct(Identifier.construct(maxIntName)))),
                  MinusTerm.construct(
                      Reference.construct(Identifier.construct("x")), 
                      Reference.construct(Identifier.construct("y"))),
                  FunctionApplication.construct(
                      Reference.construct(Identifier.construct(intMinusOverflowName)),
                      new Term[] {
                        Reference.construct(Identifier.construct("x")), 
                        Reference.construct(Identifier.construct("y"))}))));
    
    // arithmetic operations: multiplication
    // intTimes(x,y) = IF MIN_INT <= x*y AND x*y <= MAX_INT 
    //                   THEN x*y ELSE intTimesOverflow(x,y) ENDIF
    ValueDeclaration intTimesOverflowDecl =
      ValueDeclarationClass.construct(Identifier.construct(intTimesOverflowName),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName)),
                Reference.construct(Identifier.construct(intName))
              },
              Reference.construct(Identifier.construct(intName))));
    ValueDeclaration intTimesDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(intTimesName),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName)),
                Reference.construct(Identifier.construct(intName))
              },
              IntType.construct()),
          LambdaTerm.construct(
              new TypedIdentifier[] {
                  TypedIdentifier.construct(
                      Identifier.construct("x"), 
                      Reference.construct(Identifier.construct(intName))),
                  TypedIdentifier.construct(
                      Identifier.construct("y"), 
                      Reference.construct(Identifier.construct(intName)))},
              IfThenElseTerm.construct(
                  AndFormula.construct(
                      LessEqualFormula.construct(
                          Reference.construct(Identifier.construct(minIntName)),
                          TimesTerm.construct(
                              Reference.construct(Identifier.construct("x")), 
                              Reference.construct(Identifier.construct("y")))),
                      LessEqualFormula.construct(
                          TimesTerm.construct(
                              Reference.construct(Identifier.construct("x")), 
                              Reference.construct(Identifier.construct("y"))),
                          Reference.construct(Identifier.construct(maxIntName)))),
                  TimesTerm.construct(
                      Reference.construct(Identifier.construct("x")), 
                      Reference.construct(Identifier.construct("y"))),
                  FunctionApplication.construct(
                      Reference.construct(Identifier.construct(intTimesOverflowName)),
                      new Term[] {
                        Reference.construct(Identifier.construct("x")), 
                        Reference.construct(Identifier.construct("y"))}))));
    
    // arithmetic operations: division
    // intTruncAxiom: AXIOM FORALL(x: REAL):
    //   IF x >= 0 THEN x-1 < intTrunc(x) AND intTrunc(x) <= x
    //             ELSE x <= intTrunc(x) AND intTrunc(x) < x+1 ENDIF
    // intDiv(x,y) = 
    //    IF y /= 0 AND MIN_INT <= intTrunc(x/y) AND intTrunc(x/y) <= MAX_INT 
    //      THEN intTrunc(x/y) ELSE intDivOverflow(x,y) ENDIF
    // intDiv0(x,y) = 
    //    IF y /= 0 THEN intTrunc(x/y) ELSE intDivOverflow(x,y) ENDIF
    ValueDeclaration intTruncDecl =
      ValueDeclarationClass.construct(Identifier.construct(intTruncName),
          FunctionType.construct(
              new Type[] { RealType.construct() }, IntType.construct()));
    AxiomDefinition intTruncAxiom =
      AxiomDefinition.construct(Identifier.construct("intTruncAxiom"),
          ForallFormula.construct(
              new TypedIdentifier[] {
                TypedIdentifier.construct(Identifier.construct("x"), RealType.construct())},
              IfThenElseFormula.construct(
                  GreaterEqualFormula.construct(
                      Reference.construct(Identifier.construct("x")),
                      IntLiteral.construct("0")),
                  AndFormula.construct(
                      LessFormula.construct(
                          MinusTerm.construct(
                              Reference.construct(Identifier.construct("x")),
                              IntLiteral.construct("1")),
                          FunctionApplication.construct(
                              Reference.construct(Identifier.construct(intTruncName)),
                              new Term[] {
                                Reference.construct(Identifier.construct("x"))})),
                      LessEqualFormula.construct(
                          FunctionApplication.construct(
                              Reference.construct(Identifier.construct(intTruncName)),
                              new Term[] {
                                Reference.construct(Identifier.construct("x"))}),
                          Reference.construct(Identifier.construct("x")))),
                  AndFormula.construct(
                      LessEqualFormula.construct(
                          Reference.construct(Identifier.construct("x")),
                          FunctionApplication.construct(
                              Reference.construct(Identifier.construct(intTruncName)),
                                  new Term[] {
                                      Reference.construct(Identifier.construct("x"))})),
                      LessFormula.construct(
                          FunctionApplication.construct(
                              Reference.construct(Identifier.construct(intTruncName)),
                                  new Term[] {
                                      Reference.construct(Identifier.construct("x"))}),
                          PlusTerm.construct(
                              Reference.construct(Identifier.construct("x")),
                              IntLiteral.construct("1")))))));           
    ValueDeclaration intDivOverflowDecl =
      ValueDeclarationClass.construct(Identifier.construct(intDivOverflowName),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName)),
                Reference.construct(Identifier.construct(intName))
              },
              Reference.construct(Identifier.construct(intName))));
    ValueDeclaration intDivDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(intDivName),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName)),
                Reference.construct(Identifier.construct(intName))
              },
              IntType.construct()),
          LambdaTerm.construct(
              new TypedIdentifier[] {
                  TypedIdentifier.construct(
                      Identifier.construct("x"), 
                      Reference.construct(Identifier.construct(intName))),
                  TypedIdentifier.construct(
                      Identifier.construct("y"), 
                      Reference.construct(Identifier.construct(intName)))},
              IfThenElseTerm.construct(
                  AndFormula.construct(
                      NotEqualsFormula.construct(
                          Reference.construct(Identifier.construct("y")),
                          IntLiteral.construct("0")),
                          AndFormula.construct(
                              LessEqualFormula.construct(
                                  Reference.construct(Identifier.construct(minIntName)),
                                  FunctionApplication.construct(
                                      Reference.construct(Identifier.construct(intTruncName)),
                                      new Term[] {
                                        DividesTerm.construct(
                                            Reference.construct(Identifier.construct("x")), 
                                            Reference.construct(Identifier.construct("y")))})),
                                            LessEqualFormula.construct(
                                                FunctionApplication.construct(
                                                    Reference.construct(Identifier.construct(intTruncName)),
                                                    new Term[] {
                                                      DividesTerm.construct(
                                                          Reference.construct(Identifier.construct("x")), 
                                                          Reference.construct(Identifier.construct("y")))}),
                                                          Reference.construct(Identifier.construct(maxIntName))))),
                  FunctionApplication.construct(
                     Reference.construct(Identifier.construct(intTruncName)),
                     new Term[] {
                       DividesTerm.construct(
                         Reference.construct(Identifier.construct("x")), 
                         Reference.construct(Identifier.construct("y")))}),
                  FunctionApplication.construct(
                      Reference.construct(Identifier.construct(intDivOverflowName)),
                      new Term[] {
                        Reference.construct(Identifier.construct("x")), 
                        Reference.construct(Identifier.construct("y"))}))));
    ValueDeclaration intDivDecl0 =
      ValueDeclarationInitialized.construct(Identifier.construct(intDivName0),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName)),
                Reference.construct(Identifier.construct(intName))
              },
              IntType.construct()),
          LambdaTerm.construct(
              new TypedIdentifier[] {
                  TypedIdentifier.construct(
                      Identifier.construct("x"), 
                      Reference.construct(Identifier.construct(intName))),
                  TypedIdentifier.construct(
                      Identifier.construct("y"), 
                      Reference.construct(Identifier.construct(intName)))},
              IfThenElseTerm.construct(
                      NotEqualsFormula.construct(
                          Reference.construct(Identifier.construct("y")),
                          IntLiteral.construct("0")),
                  FunctionApplication.construct(
                     Reference.construct(Identifier.construct(intTruncName)),
                     new Term[] {
                       DividesTerm.construct(
                         Reference.construct(Identifier.construct("x")), 
                         Reference.construct(Identifier.construct("y")))}),
                  FunctionApplication.construct(
                      Reference.construct(Identifier.construct(intDivOverflowName)),
                      new Term[] {
                        Reference.construct(Identifier.construct("x")), 
                        Reference.construct(Identifier.construct("y"))}))));
    
    // arithmetic operations: remainder
    // intRem(x,y) = x-intDiv(x,y)*y
    ValueDeclaration intRemDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(intRemName),
          FunctionType.construct(
              new Type[] {
                Reference.construct(Identifier.construct(intName)),
                Reference.construct(Identifier.construct(intName))
              },
              IntType.construct()),
          LambdaTerm.construct(
              new TypedIdentifier[] {
                  TypedIdentifier.construct(
                      Identifier.construct("x"), 
                      Reference.construct(Identifier.construct(intName))),
                  TypedIdentifier.construct(
                      Identifier.construct("y"), 
                      Reference.construct(Identifier.construct(intName)))},
              MinusTerm.construct(
                  Reference.construct(Identifier.construct("x")),
                  TimesTerm.construct(
                      FunctionApplication.construct(
                          Reference.construct(Identifier.construct(intDivName)),
                          new Term[] {
                              Reference.construct(Identifier.construct("x")), 
                              Reference.construct(Identifier.construct("y"))}),
                      Reference.construct(Identifier.construct("y"))))));
    
    // the null values
    ValueDeclaration zeroDecl =
      ValueDeclarationInitialized.construct(
          Identifier.construct(zeroName),
          Reference.construct(Identifier.construct(intName)), 
          IntLiteral.construct("0"));
    ValueDeclaration nullcharDecl =
      ValueDeclarationClass.construct(Identifier.construct(nullCharName),
          Reference.construct(Identifier.construct(charName)));
    ValueDeclaration nullstringDecl =
      ValueDeclarationClass.construct(Identifier.construct(nullStringName),
          stringType);
    ValueDeclaration literalstringDecl =
      ValueDeclarationClass.construct(Identifier.construct(newStringName),
          FunctionType.construct(
              new Type[] { IntType.construct() },
              stringType));
    
    // array constants/functions for machine types 
    TypeDeclaration tintDecl =
      TypeDefinition.construct(Identifier.construct(typeIntArrayName),
        arrayType(
            Reference.construct(Identifier.construct(intName)), true, null));
    ValueDeclaration nintDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(nullIntArrayName),
          Reference.construct(Identifier.construct(typeIntArrayName)),
          RecordTerm.construct(
              new ValuedIdentifier[] {
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayValue),
                    ArrayTerm.construct(
                        new TypedIdentifier[] {
                            TypedIdentifier.construct(
                                Identifier.construct("y"),
                                Reference.construct(Identifier.construct(intName))) },
                        Reference.construct(Identifier.construct(zeroName)))),
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayLength),
                    Reference.construct(Identifier.construct(zeroName))),   
                  ValuedIdentifier.construct(
                    Identifier.construct(isNullName),
                    booleanConstant(true)),
              }));
    ValueDeclaration newintDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(newIntArrayName),
          FunctionType.construct(
              new Type[] { Reference.construct(Identifier.construct(natName)) },
          Reference.construct(Identifier.construct(typeIntArrayName))),
          LambdaTerm.construct(
              new TypedIdentifier[] {
                TypedIdentifier.construct(
                    Identifier.construct("x"),
                    Reference.construct(Identifier.construct(natName))) },
          RecordTerm.construct(
              new ValuedIdentifier[] {
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayValue),
                    ArrayTerm.construct(
                        new TypedIdentifier[] {
                            TypedIdentifier.construct(
                                Identifier.construct("y"),
                                Reference.construct(Identifier.construct(intName))) },
                        Reference.construct(Identifier.construct(zeroName)))),
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayLength),
                    Reference.construct(Identifier.construct("x"))),   
                  ValuedIdentifier.construct(
                    Identifier.construct(isNullName),
                    booleanConstant(false)),
              })));
    /*
    AxiomDefinition newintAxiom =
      AxiomDefinition.construct(Identifier.construct(newIntArrayAxiomName),
          ForallFormula.construct(
              new TypedIdentifier[] {
                  TypedIdentifier.construct(
                      Identifier.construct("x"),
                      Reference.construct(Identifier.construct(natName))) },
                      LetFormula.construct(
                          new ValueDefinition[] {
                              ValueDefinitionClass.construct(
                                  Identifier.construct("y"),
                                  FunctionApplication.construct(
                                      Reference.construct(Identifier.construct(newIntArrayName)),
                                      new Term[] {
                                        Reference.construct(Identifier.construct("x")) })) },
                                        AndFormula.construct(
                                            NotFormula.construct(
                                                Reference.construct(Name.construct(
                                                  new Identifier[] { Identifier.construct("y"),
                                                                     Identifier.construct(isNullName) }))),
                                                AndFormula.construct(
                                                    EqualsFormula.construct(
                                                        SelectionTerm.construct(
                                                            Reference.construct(Identifier.construct("y")),
                                                            SelectorIdentifier.construct(Identifier.construct(arrayLength))),
                                                            Reference.construct(Identifier.construct("x"))),
                                                            ForallFormula.construct(
                                                                new TypedIdentifier[] {
                                                                    TypedIdentifier.construct(
                                                                        Identifier.construct("i"),
                                                                        Reference.construct(Identifier.construct(natName))) },
                                                                        ImpliesFormula.construct(
                                                                            LessFormula.construct(
                                                                                Reference.construct(Identifier.construct("i")),
                                                                                Reference.construct(Identifier.construct("x"))),
                                                                                EqualsFormula.construct(
                                                                                    SelectionTerm.construct(
                                                                                        SelectionTerm.construct(
                                                                                            Reference.construct(Identifier.construct("y")),
                                                                                            SelectorIdentifier.construct(Identifier.construct(arrayValue))),
                                                                                            SelectorIndex.construct(Reference.construct(Identifier.construct("i")))),
                                                                                            IntLiteral.construct("0")))))))));
                                                                                            */
    TypeDeclaration tcharDecl =
      TypeDefinition.construct(Identifier.construct(typeCharArrayName),
        arrayType(
            Reference.construct(Identifier.construct(charName)), true, null));
    ValueDeclaration ncharDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(nullCharArrayName),
          Reference.construct(Identifier.construct(typeCharArrayName)),
          RecordTerm.construct(
              new ValuedIdentifier[] {
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayValue),
                    ArrayTerm.construct(
                        new TypedIdentifier[] {
                            TypedIdentifier.construct(
                                Identifier.construct("y"),
                                Reference.construct(Identifier.construct(intName))) },
                        Reference.construct(Identifier.construct(nullCharName)))),
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayLength),
                    Reference.construct(Identifier.construct(zeroName))),   
                  ValuedIdentifier.construct(
                    Identifier.construct(isNullName),
                    booleanConstant(true))
              }));
    ValueDeclaration newcharDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(newCharArrayName),
          FunctionType.construct(
              new Type[] { Reference.construct(Identifier.construct(natName)) },
          Reference.construct(Identifier.construct(typeCharArrayName))),
          LambdaTerm.construct(
              new TypedIdentifier[] {
                TypedIdentifier.construct(
                    Identifier.construct("x"),
                    Reference.construct(Identifier.construct(natName))) },
          RecordTerm.construct(
              new ValuedIdentifier[] {
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayValue),
                    ArrayTerm.construct(
                        new TypedIdentifier[] {
                            TypedIdentifier.construct(
                                Identifier.construct("y"),
                                Reference.construct(Identifier.construct(intName))) },
                        Reference.construct(Identifier.construct(nullCharName)))),
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayLength),
                    Reference.construct(Identifier.construct("x"))),   
                  ValuedIdentifier.construct(
                    Identifier.construct(isNullName),
                    booleanConstant(false)),
              })));
    TypeDeclaration tbooleanDecl =
      TypeDefinition.construct(Identifier.construct(typeBooleanArrayName),
        arrayType(booleanType(), true, null));
    ValueDeclaration nbooleanDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(nullBooleanArrayName),
          Reference.construct(Identifier.construct(typeBooleanArrayName)),
          RecordTerm.construct(
              new ValuedIdentifier[] {
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayValue),
                    ArrayTerm.construct(
                        new TypedIdentifier[] {
                            TypedIdentifier.construct(
                                Identifier.construct("y"),
                                Reference.construct(Identifier.construct(intName))) },
                                booleanConstant(false))),
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayLength),
                    Reference.construct(Identifier.construct(zeroName))),   
                  ValuedIdentifier.construct(
                    Identifier.construct(isNullName),
                    booleanConstant(true))
              }));
    ValueDeclaration newbooleanDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(newBooleanArrayName),
          FunctionType.construct(
              new Type[] { Reference.construct(Identifier.construct(natName)) },
          Reference.construct(Identifier.construct(typeBooleanArrayName))),
          LambdaTerm.construct(
              new TypedIdentifier[] {
                TypedIdentifier.construct(
                    Identifier.construct("x"),
                    Reference.construct(Identifier.construct(natName))) },
          RecordTerm.construct(
              new ValuedIdentifier[] {
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayValue),
                    ArrayTerm.construct(
                        new TypedIdentifier[] {
                            TypedIdentifier.construct(
                                Identifier.construct("y"),
                                Reference.construct(Identifier.construct(intName))) },
                        booleanConstant(false))),
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayLength),
                    Reference.construct(Identifier.construct("x"))),   
                  ValuedIdentifier.construct(
                    Identifier.construct(isNullName),
                    booleanConstant(false)),
              })));
    /*
    AxiomDefinition newbooleanAxiom =
      AxiomDefinition.construct(Identifier.construct(newBooleanArrayAxiomName),
          ForallFormula.construct(
              new TypedIdentifier[] {
                  TypedIdentifier.construct(
                      Identifier.construct("x"),
                      Reference.construct(Identifier.construct(natName))) },
                      LetFormula.construct(
                          new ValueDefinition[] {
                              ValueDefinitionClass.construct(
                                  Identifier.construct("y"),
                                  FunctionApplication.construct(
                                      Reference.construct(Identifier.construct(newBooleanArrayName)),
                                      new Term[] {
                                        Reference.construct(Identifier.construct("x")) })) },
                                        AndFormula.construct(
                                            NotFormula.construct(
                                                Reference.construct(Name.construct(
                                                  new Identifier[] { Identifier.construct("y"),
                                                                     Identifier.construct(isNullName) }))),
                                                AndFormula.construct(
                                                    EqualsFormula.construct(
                                                        SelectionTerm.construct(
                                                            Reference.construct(Identifier.construct("y")),
                                                            SelectorIdentifier.construct(Identifier.construct(arrayLength))),
                                                            Reference.construct(Identifier.construct("x"))),
                                                            ForallFormula.construct(
                                                                new TypedIdentifier[] {
                                                                    TypedIdentifier.construct(
                                                                        Identifier.construct("i"),
                                                                        Reference.construct(Identifier.construct(natName))) },
                                                                        ImpliesFormula.construct(
                                                                            LessFormula.construct(
                                                                                Reference.construct(Identifier.construct("i")),
                                                                                Reference.construct(Identifier.construct("x"))),
                                                                                EqualsFormula.construct(
                                                                                    SelectionTerm.construct(
                                                                                        SelectionTerm.construct(
                                                                                            Reference.construct(Identifier.construct("y")),
                                                                                            SelectorIdentifier.construct(Identifier.construct(arrayValue))),
                                                                                            SelectorIndex.construct(Reference.construct(Identifier.construct("i")))),
                                                                                            False.construct()))))))));   
                                                                                            */                  
    TypeDeclaration tstringDecl =
      TypeDefinition.construct(Identifier.construct(typeStringArrayName),
        arrayType(stringType, true, null));
    ValueDeclaration nstringDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(nullStringArrayName),
          Reference.construct(Identifier.construct(typeStringArrayName)),
          RecordTerm.construct(
              new ValuedIdentifier[] {
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayValue),
                    ArrayTerm.construct(
                        new TypedIdentifier[] {
                            TypedIdentifier.construct(
                                Identifier.construct("y"),
                                Reference.construct(Identifier.construct(intName))) },
                                Reference.construct(Identifier.construct(nullStringName)))),
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayLength),
                    Reference.construct(Identifier.construct(zeroName))),   
                  ValuedIdentifier.construct(
                    Identifier.construct(isNullName),
                    booleanConstant(true)),
              }));
    ValueDeclaration newstringDecl =
      ValueDeclarationInitialized.construct(Identifier.construct(newStringArrayName),
          FunctionType.construct(
              new Type[] { Reference.construct(Identifier.construct(natName)) },
          Reference.construct(Identifier.construct(typeStringArrayName))),
          LambdaTerm.construct(
              new TypedIdentifier[] {
                TypedIdentifier.construct(
                    Identifier.construct("x"),
                    Reference.construct(Identifier.construct(natName))) },
          RecordTerm.construct(
              new ValuedIdentifier[] {
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayValue),
                    ArrayTerm.construct(
                        new TypedIdentifier[] {
                            TypedIdentifier.construct(
                                Identifier.construct("y"),
                                Reference.construct(Identifier.construct(intName))) },
                        Reference.construct(Identifier.construct(nullStringName)))),
                  ValuedIdentifier.construct(
                    Identifier.construct(arrayLength),
                    Reference.construct(Identifier.construct("x"))),   
                  ValuedIdentifier.construct(
                    Identifier.construct(isNullName),
                    booleanConstant(false)),
              })));
    /*
    AxiomDefinition newstringAxiom =
      AxiomDefinition.construct(Identifier.construct(newStringArrayAxiomName),
          ForallFormula.construct(
              new TypedIdentifier[] {
                  TypedIdentifier.construct(
                      Identifier.construct("x"),
                      Reference.construct(Identifier.construct(natName))) },
                      LetFormula.construct(
                          new ValueDefinition[] {
                              ValueDefinitionClass.construct(
                                  Identifier.construct("y"),
                                  FunctionApplication.construct(
                                      Reference.construct(Identifier.construct(newStringArrayName)),
                                      new Term[] {
                                        Reference.construct(Identifier.construct("x")) })) },
                                        AndFormula.construct(
                                            NotFormula.construct(
                                                Reference.construct(Name.construct(
                                                  new Identifier[] { Identifier.construct("y"),
                                                                     Identifier.construct(isNullName) }))),
                                                AndFormula.construct(
                                                    EqualsFormula.construct(
                                                        SelectionTerm.construct(
                                                            Reference.construct(Identifier.construct("y")),
                                                            SelectorIdentifier.construct(Identifier.construct(arrayLength))),
                                                            Reference.construct(Identifier.construct("x"))),
                                                            ForallFormula.construct(
                                                                new TypedIdentifier[] {
                                                                    TypedIdentifier.construct(
                                                                        Identifier.construct("i"),
                                                                        Reference.construct(Identifier.construct(natName))) },
                                                                        ImpliesFormula.construct(
                                                                            LessFormula.construct(
                                                                                Reference.construct(Identifier.construct("i")),
                                                                                Reference.construct(Identifier.construct("x"))),
                                                                                EqualsFormula.construct(
                                                                                    SelectionTerm.construct(
                                                                                        SelectionTerm.construct(
                                                                                            Reference.construct(Identifier.construct("y")),
                                                                                            SelectorIdentifier.construct(Identifier.construct(arrayValue))),
                                                                                            SelectorIndex.construct(Reference.construct(Identifier.construct("i")))),
                                                                                            Reference.construct(Identifier.construct(nullStringName))))))))));   
                                                                                            */                  
    // the base theory
    Identifier name = Identifier.construct(baseTheoryName);
    TheoryDeclaration decl = TheoryDeclaration.construct(null, 
        new Import[] { },
        name,
        new Name[] { },
        ExpressionLogic.baseFunctions ? 
        new Declaration[] { 
          minIntDef, maxIntDef, minIntAxiom, maxIntAxiom, 
          intDef, natDef, charDecl, 
          intPlusOverflowDecl, intPlusDecl,
          intNegOverflowDecl, intNegDecl,
          intMinusOverflowDecl, intMinusDecl, 
          intTimesOverflowDecl, intTimesDecl,
          intTruncDecl, intTruncAxiom, 
          intDivOverflowDecl, intDivDecl, intDivDecl0,
          intRemDecl,
          zeroDecl, nullcharDecl, nullstringDecl, literalstringDecl,
          tintDecl, nintDecl, newintDecl, /*newintAxiom,*/
          tcharDecl, ncharDecl, newcharDecl,
          tbooleanDecl, nbooleanDecl, newbooleanDecl, /*newbooleanAxiom,*/
          tstringDecl, nstringDecl, newstringDecl /*,newstringAxiom*/
        } :
          new Declaration[] { 
            minIntDef, maxIntDef, minIntAxiom, maxIntAxiom, 
            intDef, natDef, charDecl,
            intTruncDecl, intTruncAxiom, 
            intDivOverflowDecl, intDivDecl0,
            zeroDecl, nullcharDecl, nullstringDecl, literalstringDecl,
            tintDecl, nintDecl, newintDecl, /*newintAxiom,*/
            tcharDecl, ncharDecl, newcharDecl,
            tbooleanDecl, nbooleanDecl, newbooleanDecl, /*newbooleanAxiom,*/
            tstringDecl, nstringDecl, newstringDecl /*,newstringAxiom*/
          }
    );

     // set base theory symbol and process its declaration
     baseTheory = new TheorySymbol(top, name, decl, error);
     if (!baseTheory.isRead()) Main.internalError("error in base theory");
     boolean okay = top.getTheories().put(baseTheory);
     if (!okay)
     {
       error.newError("cannot create theory " + baseTheory.getLongName() +
           " (top package already contains theory with this name)");
     }
     
     // the default string theory
     stringTheory = baseTheory;
         
     // set symbols
     SymbolTable<TypeSymbol> types = baseTheory.getTypes();
     intSymbol = types.get(Identifier.construct(intName));
     natSymbol = types.get(Identifier.construct(natName));
     charSymbol = types.get(Identifier.construct(charName));
  }
  
  /****************************************************************************
   * Get array type for denoted base type
   * @param base type
   * @param inBase true if type is referenced within baseTheory
   * @param theories if not null and inBase is true, collection 
   *                 to which base theory is to be added.
   * @return the array type with the denoted base type.
   ***************************************************************************/
  private Type arrayType(Type base, boolean inBase, Collection<TheorySymbol> theories)
  {
    // references to type "int" and "nat"
    Reference intRef;
    Reference natRef;
    if (inBase) 
    {
      intRef = Reference.construct(Identifier.construct(intName));
      natRef = Reference.construct(Identifier.construct(natName));
    }
    else
    {
      Name baseName = baseTheory.getLongName();
      if (theories != null) TheorySymbol.add(theories, baseTheory);
      Name longIntName = clonePrefix(baseName, intSymbol.getName());
      Name longNatName = clonePrefix(baseName, natSymbol.getName());
      intRef = Reference.construct(longIntName);
      natRef = Reference.construct(longNatName);
    }
    
    Identifier valueIdent = Identifier.construct(arrayValue);
    TypedIdentifier valueTid = 
      TypedIdentifier.construct(valueIdent, ArrayType.construct(intRef, base));
    if (!inBase)
    {
      ValueSymbol valueSymbol = ValueSymbol.construct(valueIdent, 
          ArrayType.construct(intSymbol.getType(), base), valueTid);
      valueIdent.setSymbol(valueSymbol);
    }
    Identifier lengthIdent = Identifier.construct(arrayLength);
    TypedIdentifier lengthTid = TypedIdentifier.construct(lengthIdent, natRef);
    if (!inBase)
    {
      ValueSymbol lengthSymbol = ValueSymbol.construct(lengthIdent, 
          natSymbol.getType(), lengthTid);
      lengthIdent.setSymbol(lengthSymbol);
    }
    Identifier nullIdent = Identifier.construct(isNullName);
    TypedIdentifier nullTid = TypedIdentifier.construct(nullIdent, 
        booleanType());
    if (!inBase)
    {
      ValueSymbol nullSymbol = ValueSymbol.construct(nullIdent, 
          booleanType(), nullTid);
      nullIdent.setSymbol(nullSymbol);
    }
    return RecordType.construct(new TypedIdentifier[] 
             {valueTid, lengthTid, nullTid});
  }
  
  /****************************************************************************
   * Get the base theory symbol.
   * @return the symbol of the base theory.
   ***************************************************************************/
  public TheorySymbol getBaseTheory()
  {
    return baseTheory;
  }
  
  /***************************************************************************
   * Return name whose prefix is a clone of the denoted prefix that
   * preserves the identifiers of the prefix.
   * @param base the prefix of the name (to be cloned).
   * @param ident the suffix of the name (not cloned).
   * @return a name consisting of a clone of the prefix and a clone of the suffix
   *         such that the symbols of the prefix are preserved.
   **************************************************************************/
  private Name clonePrefix(Name prefix, Identifier ident)
  {
    int n = prefix.getLength();
    Identifier[] idents = new Identifier[n+1];
    for (int i=0; i<n; i++) 
    {
      Identifier ident0 = prefix.idents[i];
      idents[i] = Identifier.construct(ident0.getString());
      idents[i].setSymbol(ident0.getSymbol());
    }
    idents[n] = ident;
    return Name.construct(idents);
  }
  
  /****************************************************************************
   * Translate a program's type expression to a logical type.
   * @param exp a type expression that does not contain the "void" type
   * @return the logical type (null, if something went wrong)
   ***************************************************************************/
  public Type logicType(TypeExpression exp)
  {
    Vector<TheorySymbol> theories = new Vector<TheorySymbol>();
    return logicType(exp, theories);
  }
  
  /****************************************************************************
   * Translate a program's type expression to a logical type and collect
   * the referenced theories.
   * @param exp a type expression that does not contain the "void" type
   * @param theories a set of theories to which any theory referenced in 
   *        the translated type is to be added (may be null)
   * @return the logical type (null, if something went wrong)
   ***************************************************************************/
  private Type logicType(TypeExpression exp, Vector<TheorySymbol> theories)
  {
    if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.BooleanType)
    {
      if (Main.firstOrder())
        return LogicTypeTable.bitType;
      else
        return LogicTypeTable.booleanType;
    }
    if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.CharType)
    {
      Name base = baseTheory.getLongName();
      Name name = clonePrefix(base, charSymbol.getName());
      if (theories != null) TheorySymbol.add(theories, baseTheory);
      return Reference.construct(name);
    }
    if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.ExceptionType)
    {
      // an exception is just interpreted as the string it contains
      return stringType;
    }
    if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.IntType)
    {
      Name base = baseTheory.getLongName();
      Name name = clonePrefix(base, intSymbol.getName());
      if (theories != null) TheorySymbol.add(theories, baseTheory);
      return Reference.construct(name);
    }
    if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.NullType)
    {
      Main.internalError("unexpected null type encountered");
    }
    if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.StringType)
      return stringType;
    if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.VoidType)
    {
      Main.internalError("unexpected void type encountered");
    }
    if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.ArrayType)
    {
      fmrisc.ProgramExplorer.Syntax.Program.ArrayType exp0 =
        (fmrisc.ProgramExplorer.Syntax.Program.ArrayType)exp;
      Type btype = logicType(exp0.base, theories);
      if (btype == null) return null;
      return arrayType(btype, false, theories);
    }
    if (exp instanceof NamedType)
    {
      Name name = ((NamedType)exp).name;
      ClassSymbol csymbol = (ClassSymbol)name.getSymbol();
      TheorySymbol tsymbol = csymbol.getClassTheory(error);
      if (tsymbol == null) return null;
      if (theories != null) TheorySymbol.add(theories, tsymbol);
      Name base = tsymbol.getLongName();
      Identifier tident = Identifier.construct(tsymbol.getName().getString());
      SymbolTable<TypeSymbol> types = tsymbol.getTypes();
      if (types == null) Main.internalError("empty class theory: " + base);
      TypeSymbol type = types.get(tident);
      if (type == null) Main.internalError("class theory type not found");
      Name tname = clonePrefix(base, tident);
      return Reference.construct(tname);
    }
    if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.NewType)
    {
      Main.internalError("unexpected new type encountered");
    }
    Main.internalError("unknown program type expression: " + exp);
    return null;
  }
  
  /****************************************************************************
   * Construct a theory declaration for the denoted class.
   * @param csymbol the symbol of the class
   * @return the declaration (null, on error)
   ***************************************************************************/
  public TheoryDeclaration constructClassTheory(ClassSymbol csymbol)
  {
    // make sure class declaration is available
    if (!csymbol.isRead())
    {
      boolean okay = csymbol.read(error);
      if (!okay) return null;
    }
    
    // all theories referenced by the theory
    Vector<TheorySymbol> tset = new Vector<TheorySymbol>();
    
    // construct record type for the object variables
    Declaration[] decls = csymbol.getDeclaration().decls; 
    Vector<TypedIdentifier> tidents = new Vector<TypedIdentifier>();
    for (Declaration d : decls)
    {
      if (!(d instanceof ObjectVariableDeclaration)) continue;
      ObjectVariableDeclaration d0 = (ObjectVariableDeclaration)d;
      Identifier ident = d0.getName();
      // extract type from symbol (may have been translated from declaration)
      ObjectVariableSymbol symbol = (ObjectVariableSymbol)ident.getSymbol();
      // there may not be a symbol due to mutual recursion of classes
      if (symbol == null) return null;
      Type type = logicType(symbol.getType(), tset);
      if (type == null) return null;
      Identifier ident0 = Identifier.construct(ident.getString());
      tidents.add(TypedIdentifier.construct(ident0, type));
    }
    Identifier nullIdent = Identifier.construct(isNullName);
    tidents.add(TypedIdentifier.construct(nullIdent, booleanType()));
    Identifier newIdent = Identifier.construct(newName);
    tidents.add(TypedIdentifier.construct(newIdent, IntType.construct()));
    TypedIdentifier[] tidents0 = new TypedIdentifier[tidents.size()];
    tidents.toArray(tidents0);
    RecordType type = RecordType.construct(tidents0);
    
    // construct theory with record type and corresponding functions/constants
    Name cname = csymbol.getLongName();
    Name pname = null;
    if (cname.idents.length > 1) pname = cname.getPrefix();
    Identifier cident = cname.getBase();
    Identifier tident = Identifier.construct(cident.getString());
    Identifier ident = Identifier.construct(cident.getString());
    Identifier nident = Identifier.construct(nullName);
    Identifier noident = Identifier.construct(newObjectName);
    Identifier taident = Identifier.construct(typeArrayName);
    Identifier naident = Identifier.construct(nullArrayName);
    Identifier newident = Identifier.construct(newArrayName);
    Identifier natIdent = 
      baseTheory.getTypes().get(Identifier.construct(natName)).getName();
    Name natName = Name.construct(baseTheory.getLongName(), natIdent);
    Identifier intIdent = 
        baseTheory.getTypes().get(Identifier.construct(intName)).getName();
    Name intName = Name.construct(baseTheory.getLongName(), intIdent);
    Identifier zeroIdent = 
        baseTheory.getValues().get(Identifier.construct(zeroName)).getName();
    Name zeroName = Name.construct(baseTheory.getLongName(), zeroIdent);
    TypeDefinition tdef = TypeDefinition.construct(ident, type);
    ValueDeclaration ndecl = ValueDeclarationClass.construct(nident, 
        Reference.construct(Identifier.construct(ident.getString())));
    ValueDeclaration nodecl =
      ValueDeclarationInitialized.construct(noident,
          FunctionType.construct(
              new Type[] { IntType.construct() },
              Reference.construct(Identifier.construct(ident.getString()))),
              LambdaTerm.construct(
                  new TypedIdentifier[] {
                      TypedIdentifier.construct(
                          Identifier.construct("x"),
                          IntType.construct())
                  },
                  UpdateTerm.construct(
                      UpdateTerm.construct(
                          Reference.construct(Identifier.construct(nident.getString())),
                          new Selector[] {
                            SelectorIdentifier.construct(
                                Identifier.construct(isNullName)) },
                                booleanConstant(false)
                      ),
                      new Selector[] {
                          SelectorIdentifier.construct(
                              Identifier.construct(newName)) },
                              Reference.construct(Identifier.construct("x"))
              )));
    TypeDeclaration tadecl =
      TypeDefinition.construct(taident,
         arrayType(
           Reference.construct(Identifier.construct(ident.getString())), false, null));
    ValueDeclaration nadecl = ValueDeclarationClass.construct(naident, 
          Reference.construct(Identifier.construct(typeArrayName)));
    ValueDeclaration newdecl = ValueDeclarationClass.construct(newident, 
        FunctionType.construct(
            new Type[] { Reference.construct(natName) },
            Reference.construct(Identifier.construct(typeArrayName))));
    AxiomDefinition newaxiom =
      AxiomDefinition.construct(Identifier.construct(newArrayAxiomName),
          ForallFormula.construct(
              new TypedIdentifier[] {
                  TypedIdentifier.construct(
                      Identifier.construct("x"),
                      Reference.construct(natName)) },
                      LetFormula.construct(
                          new ValueDefinition[] {
                              ValueDefinitionClass.construct(
                                  Identifier.construct("y"),
                                  FunctionApplication.construct(
                                      Reference.construct(Identifier.construct(newArrayName)),
                                      new Term[] {
                                        Reference.construct(Identifier.construct("x")) })) },
                                        AndFormula.construct(
                                            Main.firstOrder() ?
                                                EqualsFormula.construct(
                                                    Reference.construct(Name.construct(
                                                      new Identifier[] { Identifier.construct("y"),
                                                                         Identifier.construct(isNullName) })),
                                                booleanConstant(false)) :
                                            NotFormula.construct(
                                                Reference.construct(Name.construct(
                                                  new Identifier[] { Identifier.construct("y"),
                                                                     Identifier.construct(isNullName) })))
                                                                     ,
                                                AndFormula.construct(
                                                    EqualsFormula.construct(
                                                        SelectionTerm.construct(
                                                            Reference.construct(Identifier.construct("y")),
                                                            SelectorIdentifier.construct(Identifier.construct(arrayLength))),
                                                            Reference.construct(Identifier.construct("x"))),
                                                            ForallFormula.construct(
                                                                new TypedIdentifier[] {
                                                                    TypedIdentifier.construct(
                                                                        Identifier.construct("i"),
                                                                        Reference.construct(natName)) },
                                                                        ImpliesFormula.construct(
                                                                            LessFormula.construct(
                                                                                Reference.construct(Identifier.construct("i")),
                                                                                Reference.construct(Identifier.construct("x"))),
                                                                                EqualsFormula.construct(
                                                                                    SelectionTerm.construct(
                                                                                        SelectionTerm.construct(
                                                                                            Reference.construct(Identifier.construct("y")),
                                                                                            SelectorIdentifier.construct(Identifier.construct(arrayValue))),
                                                                                            SelectorIndex.construct(Reference.construct(Identifier.construct("i")))),
                                                                                            Reference.construct(Identifier.construct(nullName))))))))));                     

    if (cname.toString().equals(ClassSymbol.stringClass))
    {
      // set string type
      stringType = type;
      
      // add string declarations
      ValueDeclaration nullstringDecl =
          ValueDeclarationClass.construct(Identifier.construct(nullStringName),
              stringType);
      ValueDeclaration literalstringDecl =
          ValueDeclarationClass.construct(Identifier.construct(newStringName),
              FunctionType.construct(
                  new Type[] { IntType.construct() },
                  stringType));
      TypeDeclaration tstringDecl =
          TypeDefinition.construct(Identifier.construct(typeStringArrayName),
              arrayType(stringType, false, null));
      ValueDeclaration nstringDecl =
          ValueDeclarationInitialized.construct(Identifier.construct(nullStringArrayName),
              Reference.construct(Identifier.construct(typeStringArrayName)),
              RecordTerm.construct(
                  new ValuedIdentifier[] {
                      ValuedIdentifier.construct(
                          Identifier.construct(arrayValue),
                          ArrayTerm.construct(
                              new TypedIdentifier[] {
                                  TypedIdentifier.construct(
                                      Identifier.construct("y"),
                                      Reference.construct(intName)) },
                                      Reference.construct(Identifier.construct(nullStringName)))),
                                      ValuedIdentifier.construct(
                                          Identifier.construct(arrayLength),
                                          Reference.construct(zeroName)),   
                                          ValuedIdentifier.construct(
                                              Identifier.construct(isNullName),
                                              booleanConstant(true)),
                  }));
      ValueDeclaration newstringDecl =
          ValueDeclarationInitialized.construct(Identifier.construct(newStringArrayName),
              FunctionType.construct(
                  new Type[] { Reference.construct(natName) },
                  Reference.construct(Identifier.construct(typeStringArrayName))),
                  LambdaTerm.construct(
                      new TypedIdentifier[] {
                          TypedIdentifier.construct(
                              Identifier.construct("x"),
                              Reference.construct(natName)) },
                              RecordTerm.construct(
                                  new ValuedIdentifier[] {
                                      ValuedIdentifier.construct(
                                          Identifier.construct(arrayValue),
                                          ArrayTerm.construct(
                                              new TypedIdentifier[] {
                                                  TypedIdentifier.construct(
                                                      Identifier.construct("y"),
                                                      Reference.construct(intName)) },
                                                      Reference.construct(Identifier.construct(nullStringName)))),
                                                      ValuedIdentifier.construct(
                                                          Identifier.construct(arrayLength),
                                                          Reference.construct(Identifier.construct("x"))),   
                                                          ValuedIdentifier.construct(
                                                              Identifier.construct(isNullName),
                                                              booleanConstant(false)),
                                  })));
      Set<Name> names = NameUtils.getNames(tset);
      Name[] tarray = new Name[names.size()];
      names.toArray(tarray);
      return TheoryDeclaration.construct(pname, 
          new Import[] { }, tident, tarray, 
          new Declaration[] { tdef, ndecl, nodecl, tadecl, nadecl, 
          newdecl, newaxiom, 
          nullstringDecl, literalstringDecl, tstringDecl, 
          nstringDecl, newstringDecl});
    }
    else
    {
      Set<Name> names = NameUtils.getNames(tset);
      Name[] tarray = new Name[names.size()];
      names.toArray(tarray);
      return TheoryDeclaration.construct(pname, 
          new Import[] { }, tident, tarray, 
          new Declaration[] { tdef, ndecl, nodecl, tadecl, nadecl, 
          newdecl, newaxiom });
    }
  }
  
  /*********************************************************************
   * Construct boolean type to be used for values.
   * @return the boolean type.
   ********************************************************************/
  private Type booleanType()
  {
    if (Main.firstOrder())
      return BitType.construct();
    else
      return BooleanType.construct();
  }
  
  /*********************************************************************
   * Construct constant denoting boolean constant.
   * @param true iff the constant shall denote true, false otherwise.
   * @return the constant
   ********************************************************************/
  private Constant booleanConstant(boolean value)
  {
    if (value)
    {
      if (Main.firstOrder())
        return BitTrue.construct();
      else
        return True.construct();
    }
    else
    {
      if (Main.firstOrder())
        return BitFalse.construct();
      else
        return False.construct();
    }  
  }
}
