// ----------------------------------------------------------------------------
// LoopSpec.java
// A loop specification.
// $Id: LoopSpec.java,v 1.9 2011/09/30 15:47:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Spec;

import org.antlr.runtime.*;
import java.util.*;

import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Parser.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * A loop specification.
 *
 */
public final class LoopSpec extends SpecificationBase
{
  private Formula loopInvariant;
  private Term[] terminationTerm;
  
  public final static String invariantKeyword ="invariant";
  public final static String terminationKeyword = "decreases";
  
  /****************************************************************************
   * Construct loop specification from denoted text.
   * @param text the text of the specification.
   * @param error the stream for printing error messages
   * @param pos the source position of the specification
   * @return the loop specification (null, if an error ocurred in parsing)
   ***************************************************************************/
  public static LoopSpec construct(String text, 
    ErrorStream error, SourcePosition pos)
  {
    LoopSpec spec = new LoopSpec(text);
    try
    {
      
      CharStream input = new ANTLRStringStream(text);
      SpecLangLexer lexer = new SpecLangLexer(input, error, pos);
      CommonTokenStream tokens = new CommonTokenStream(lexer);
      SpecLangParser parser = new SpecLangParser(tokens, error, pos, text);
      parser.loopspec(spec);  
    }
    catch (RecognitionException e)
    {
      error.newError("ill-formed specification (" + e.getMessage() + ")");
      return null;
    }
    return spec;
  }
  
  /***************************************************************************
   * Construct loop specification from text (other fields are not set).
   * @param text the specification text.
   **************************************************************************/
  private LoopSpec(String text)
  {
    super(text);
  }

  /***************************************************************************
   * Get loop invariant.
   * @return the loop invariant (null, if none).
   **************************************************************************/
  public Formula getLoopInvariant() { return loopInvariant; }
  
  /***************************************************************************
   * Set loop invariant.
   * @param inv the loop invariant (may be null).
   **************************************************************************/
  public void setLoopInvariant(Formula inv)
  {
    loopInvariant = inv;
  }
  
  /***************************************************************************
   * Get termination term vector.
   * @return the loop termination term vector (null, if none).
   **************************************************************************/
  public Term[] getTerminationTerm() 
  { 
    return terminationTerm; 
  }
  
  /***************************************************************************
   * Set termination term vector.
   * @param term the loop termination term vector (may be null).
   **************************************************************************/
  public void setTerminationTerm(Vector<Term> term)
  {
    int n = term.size();
    terminationTerm = new Term[n];
    term.toArray(terminationTerm);
    for (int i=0; i<n; i++)
      terminationTerm[i] = (Term)PreStateCloner.clone(terminationTerm[i]);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    if (loopInvariant != null) loopInvariant.accept(visitor);
    if (terminationTerm != null) accept(terminationTerm, visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    if (loopInvariant != null) loopInvariant.accept(visitor);
    if (terminationTerm != null) accept(terminationTerm, visitor);
  }
}

