// ----------------------------------------------------------------------------
// StatementSpec.java
// A statement specification.
// $Id: StatementSpec.java,v 1.5 2011/09/30 15:47:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Spec;

import org.antlr.runtime.*;

import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Parser.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * A statement specification.
 *
 */
public final class StatementSpec extends SpecificationBase
{
  private Formula assertion;
  
  public final static String assertionKeyword ="assert";
  
  /****************************************************************************
   * Construct method specification from denoted text.
   * @param text the text of the specification.
   * @param error the stream for printing error messages
   * @param pos the source position of the specification
   * @return the statement specification (null, if an error ocurred in parsing)
   ***************************************************************************/
  public static StatementSpec construct(String text, 
    ErrorStream error, SourcePosition pos)
  {
    StatementSpec spec = new StatementSpec(text);
    try
    {
      
      CharStream input = new ANTLRStringStream(text);
      SpecLangLexer lexer = new SpecLangLexer(input, error, pos);
      CommonTokenStream tokens = new CommonTokenStream(lexer);
      SpecLangParser parser = new SpecLangParser(tokens, error, pos, text);
      parser.statementspec(spec);  
    }
    catch (RecognitionException e)
    {
      error.newError("ill-formed specification (" + e.getMessage() + ")");
      return null;
    }
    return spec;
  }
  
  /***************************************************************************
   * Construct statement specification from text (other fields are not set).
   * @param text the specification text.
   **************************************************************************/
  private StatementSpec(String text)
  {
    super(text);
  }

  /***************************************************************************
   * Get pre-state assertion.
   * @return the assertion (null, if none).
   **************************************************************************/
  public Formula getAssertion() { return assertion; }
  
  /***************************************************************************
   * Set pre-state assertion
   * @param ass the assertion (may be null).
   **************************************************************************/
  public void setAssertion(Formula ass)
  {
    assertion = (Formula)PreStateCloner.clone(ass);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    if (assertion != null) assertion.accept(visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    if (assertion != null) assertion.accept(visitor);
  }
}

