// ----------------------------------------------------------------------------
// Theory.java
// The declaration of a logical theory.
// $Id: TheoryDeclaration.java,v 1.9 2017/07/04 08:04:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Spec;

import java.util.Vector;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.CompilationUnitBase;
import fmrisc.ProgramExplorer.Syntax.Program.Import;

/**
 * 
 * The declaration of a logical theory.
 *
 */
public final class TheoryDeclaration extends CompilationUnitBase
{  
  private final Name[] theories;
  private final Declaration[] decls;
  
  /****************************************************************************
   * Construct a theory.
   * @param pname package name (may be null)
   * @param imports import statements
   * @param name theory identifier
   * @param theories the names of the theories on which this theory depends
   *        (may be null)
   * @param decls sequence of declarations
   * @return the declaration.
   ***************************************************************************/
  public static TheoryDeclaration construct(Name pname, Vector<Import> imports,
    Identifier name, Vector<Name> theories, Vector<Declaration> decls)
  {
    if (theories == null) theories = new Vector<Name>();
    return construct(pname,
        imports.toArray(new Import[imports.size()]), 
        name,  
        theories.toArray(new Name[theories.size()]),
        decls.toArray(new Declaration[decls.size()]));
  }
  
  /****************************************************************************
   * Construct a theory declaration
   * @param pname package name (may be null)
   * @param imports import statements
   * @param name theory identifier
   * @param theories the names of the theories on which this theory depends
   * @param decls sequence of declarations
   * @return the declaration.
   ***************************************************************************/
  public static TheoryDeclaration construct(Name pname, Import[] imports,
    Identifier name, Name[] theories, Declaration[] decls)
  {
    return new TheoryDeclaration(pname, imports, name, theories, decls);
  }
  
  /****************************************************************************
   * Construct a theory declaration
   * @param pname package name
   * @param imports import statements
   * @param name theory identifier
   * @param theories the names of the theories on which this theory depends
   * @param decls sequence of declarations
   ***************************************************************************/
  private TheoryDeclaration(Name pname, Import[] imports,
    Identifier name, Name[] theories, Declaration[] decls)
  {
    super(pname, imports, name);
    this.theories = theories;
    this.decls = decls;
  }
  
  /***************************************************************************
   * Get declarations introduced in unit.
   * @return list of declarations used in unit.
   **************************************************************************/
  public Declaration[] getDeclarations() { return decls; }
  
  /***************************************************************************
   * Get theories on which unit depends.
   * @return theories on which unit depends.
   **************************************************************************/
  public Name[] getTheories() { return theories; }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    accept(theories, visitor);
    accept(decls, visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    accept(theories, visitor);
    accept(decls, visitor);
  }
}

