// ----------------------------------------------------------------------------
// UnitSpec.java
// A unit specification.
// $Id: UnitSpec.java,v 1.8 2011/09/30 15:47:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Spec;

import org.antlr.runtime.*;
import java.util.Vector;

import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Parser.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.Import;

/**
 * 
 * A unit specification.
 *
 */
public final class UnitSpec extends SpecificationBase
{  
  private Import[] imports = new Import[]{ };
  private Name[] theories = new Name[]{ };
  private Declaration[] declarations = new Declaration[]{ };
  
  /****************************************************************************
   * Construct unit specification from denoted text.
   * @param text the text of the specification.
   * @param error the stream for printing error messages
   * @param pos the source position of the specification
   * @return the unit specification (null, if an error occurred in parsing)
   ***************************************************************************/
  public static UnitSpec construct(String text, 
    ErrorStream error, SourcePosition pos)
  {
    UnitSpec spec = new UnitSpec(text);
    spec.setPosition(pos);
    try
    {
      
      CharStream input = new ANTLRStringStream(text);
      SpecLangLexer lexer = new SpecLangLexer(input, error, pos);
      CommonTokenStream tokens = new CommonTokenStream(lexer);
      SpecLangParser parser = new SpecLangParser(tokens, error, pos, text);
      parser.unitspec(spec);  
    }
    catch (RecognitionException e)
    {
      error.newError("ill-formed specification (" + e.getMessage() + ")");
      return null;
    }
    return spec;
  }
  
  /***************************************************************************
   * Construct unit specification from text (other fields are not set).
   * @param text the specification text.
   **************************************************************************/
  private UnitSpec(String text)
  {
    super(text);
  }
  
  /***************************************************************************
   * Get imports used by unit.
   * @return list of imports used by unit.
   **************************************************************************/
  public Import[] getImports() { return imports; }
  
  /***************************************************************************
   * Set imports used by unit.
   * @param ts imports (may be null)
   **************************************************************************/
  public void setImports(Vector<Import> ts)
  {
    if (ts == null) return;
    imports = ts.toArray(new Import[ts.size()]);
  }
  
  /***************************************************************************
   * Get theories referenced by unit.
   * @return list of referenced theories
   **************************************************************************/
  public Name[] getTheories() { return theories; }
  
  /***************************************************************************
   * Set theories referenced by unit.
   * @param ts list of theories (may be null)
   **************************************************************************/
  public void setTheories(Vector<Name> ts)
  {
    if (ts == null) return;
    theories = ts.toArray(new Name[ts.size()]);
  }
  
  /***************************************************************************
   * Get declarations introduced in unit.
   * @return list of declarations used in unit
   **************************************************************************/
  public Declaration[] getDeclarations() { return declarations; }
  
  /***************************************************************************
   * Set declarations introduced by unit.
   * @param ds list of declarations (may be null)
   **************************************************************************/
  public void setDeclarations(Vector<Declaration> ds)
  {
    if (ds == null) return;
    declarations = ds.toArray(new Declaration[ds.size()]);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    accept(imports, visitor);
    accept(theories, visitor);
    accept(declarations, visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    accept(imports, visitor);
    accept(theories, visitor);
    accept(declarations, visitor);
  }
}
