// ----------------------------------------------------------------------------
// StringUtil.java
// String utilities.
// $Id: StringUtil.java,v 1.2 2009/10/21 16:10:03 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax;

/**
 * 
 * String utilities.
 *
 */
public class StringUtil
{
  /***************************************************************************
   * Convert character to 4 digit hexadecimal number.
   * @param c the character
   * @return the four digit hexadecimal number code of c.
   ***************************************************************************/
  private static String toHex(char c)
  {
    String s = Integer.toHexString(c);
    while (s.length() < 4) s = "0" + s;
    return s;
  }
  
  /**************************************************************************
   * Convert string of hexadecimal digits to numeric value.
   * @param s a string of hexadecimal digits (and possibly other characters).
   * @return the hex value represented by the digit string (-1 if a character
   *         was encountered that does not represent a hexadecimal digit).
   ************************************************************************/
  private static int fromHex(String s)
  {
    try
    {
      int i = Integer.parseInt(s, 16);
      if (i < 0) return -1;
      return i;
    }
    catch(NumberFormatException i)
    {
      return -1;
    }
  }
  
  /***************************************************************************
   * Create string literal representation of c.
   * @param c a character
   * @return the representation of s as a string literal in Java (without quotes)
   ***************************************************************************/
  public static String printable(char c)
  {
    if (32 <= c && c <= 126 || 160 <= c && c <= 255) 
      return Character.toString(c);
    switch (c)
    {
      case '\n': return "\\n"; 
      case '\t': return "\\t"; 
      case '\b': return "\\b"; 
      case '\f': return "\\f"; 
      case '\r': return "\\r"; 
      case '\\': return "\\\\"; 
      case '\'': return "\\\'"; 
      case '\"': return "\\\""; 
      default: return "\\u" + toHex(c);
    }
  }
  
  /***************************************************************************
   * Create string literal representation of s.
   * @param s a string
   * @return the representation of s as a string literal in Java (without quotes)
   ***************************************************************************/
  public static String printable(String s)
  {
   int n = s.length();
   StringBuffer b = new StringBuffer(2*n);
   for (int i=0; i<n; i++)
   {
     char c = s.charAt(i);
     b.append(printable(c));
   }
   return b.toString();
  }
  
  /****************************************************************************
   * Get value of string literal.
   * @param s the content of a string literal (potentially with escape codes).
   * @return the value of the literal (null, if an error occurred).
   ***************************************************************************/
  public static String literalString(String s)
  {
    int n = s.length();
    StringBuffer b = new StringBuffer(n);
    int i = 0;
    while (i < n)
    {
      char c = s.charAt(i);
      i = i+1;
      if (c != '\\')
      {
        b.append(c);
        continue;
      }
      if (i == n) return null;
      i = escape(s, i, b);
      if (i == -1) return null;
    }
    return b.toString();
  }
  
  /****************************************************************************
   * Get value of string literal as a single character.
   * @param s the content of a string literal (potentially with escape codes).
   * @return the value of the literal as a character code (-1, if an error)
   ***************************************************************************/
  public static int literalChar(String s)
  {
    String t = literalString(s);
    if (t == null || t.length() != 1) return -1;
    return s.charAt(0);
  }
  
  /****************************************************************************
   * Parse escape code.
   * @param s a string.
   * @param p the position in the string after the escape character.
   * @param b the buffer to append the parsend value to.
   * @return the position after the escape sign (-1 if an error occurred).
   **************************************************************************/
  private static int escape(String s, int p, StringBuffer b)
  {
    char c = s.charAt(p);
    int i = p;
    switch (c)
    {
      case 'n':
      {
        b.append('\n');
        i = i+1;
        break;
      }
      case 't':
      {
        b.append('\t');
        i = i+1;
        break;
      }
      case 'b':
      {
        b.append('\b');
        i = i+1;
        break;
      }
      case 'f':
      {
        b.append('\f');
        i = i+1;
        break;
      }
      case 'r':
      {
        b.append('\r');
        break;
      }
      case '\\':
      {
        b.append('\\');
        i = i+1;
        break;
      }
      case '\'':
      {
        b.append('\'');
        i = i+1;
        break;
      }
      case '\"':
      {
        b.append('\"');
        i = i+1;
        break;
      }
      case 'u':
      {
        i = i+1;
        if (i+4 > s.length()) return -1;
        int d = fromHex(s.substring(i, i+4));
        i = i+4;
        if (d < 0 || d > Character.MAX_VALUE) return -1;
        b.append((char)d);
        break;
      }
    }
    return i;
  }
}
