// ----------------------------------------------------------------------------
// ClassTasks.java
// The delayed generation of class-related tasks.
// $Id: ClassTasks.java,v 1.4 2011/09/30 15:15:57 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.io.File;
import java.io.PrintWriter;
import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Judgements.*;
import fmrisc.ProgramExplorer.Proving.StateProblem;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;

/**
 * 
 * The delayed generation of method-related tasks.
 *
 */
public final class ClassTasks
{
  /**
   * 
   * The info necessary for delayed task creation.
   *
   */
  private final static class Info
  {
    public final ClassSymbol csymbol;
    public final Vector<TheorySymbol> theories;
    public final Declaration[][] decls;
    public final ErrorStream out;
    
    public Info(ClassSymbol csymbol,
      Vector<TheorySymbol> theories, Declaration[][] decls,
      ErrorStream out)
    {
      this.csymbol = csymbol;
      this.theories = theories;
      this.decls = decls;
      this.out = out;
    }
  }
  
  // the info registered for delayed task creation
  private static Vector<Info> infos = new Vector<Info>();
  
  /***************************************************************************
   * Register the construction of tasks for a given class.
   * @param csymbol the symbol of a class.
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   **************************************************************************/
  public static void register(ClassSymbol csymbol,
    Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream out)
  {
    infos.add(new Info(csymbol, theories, decls, out));
  }
  
  /****************************************************************************
   * Determine whether there are tasks.
   * @return true iff there are tasks.
   ***************************************************************************/
  public static boolean exist()
  {
    return !infos.isEmpty();
  }
  
  /****************************************************************************
   * Clear the registered tasks.
   ***************************************************************************/
  public static void clearTasks()
  {
    infos = new Vector<Info>();
  }
  
  /***************************************************************************
   * Generate those tasks for which the information has been previously registered.
   * @param out the stream on which to print progress information.
   ***************************************************************************/
  public static void generateTasks(PrintWriter out)
  {
    for (Info info : infos)
    {
      generate(info.csymbol,
          info.theories, info.decls, info.out);
      out.print(".");
      out.flush();
    }
    clearTasks();
  }
  
  /***************************************************************************
   * Construct the tasks for a given class.
   * @param csymbol the symbol of a class.
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param error the stream to which to write error messages.
   **************************************************************************/
  private static void generate(ClassSymbol csymbol,
    Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream error)
  {        
    // the type checker ensures that initializations are in dependence order
    ExpressionLogic logic = new ExpressionLogic(theories.get(0));
    Formula oformula = null;
    Formula cformula = null;
    Declaration[] decls0 = csymbol.getDeclaration().decls;
    for (Declaration d : decls0)
    {
      if (!(d instanceof GlobalVariableDeclaration)) continue;
      GlobalVariableDeclaration d0 = (GlobalVariableDeclaration)d;
      GlobalVariableSymbolBase vsymbol = 
          (GlobalVariableSymbolBase)d0.name.getSymbol();
      Vector<Formula> pre = new Vector<Formula>();
      ValueExpression value = d0.value;
      if (value == null) 
      {
        TypeExpression type = vsymbol.getType();
        if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.ArrayType ||
            type instanceof fmrisc.ProgramExplorer.Syntax.Program.NamedType)
        {
          NullLiteral literal = NullLiteral.construct();
          literal.setType(type);
          value = literal;
        }
        else if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.BooleanType)
          value = fmrisc.ProgramExplorer.Syntax.Program.BooleanLiteral.construct(false);
        else if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.IntType)
          value = fmrisc.ProgramExplorer.Syntax.Program.IntLiteral.construct("0");
        else
          continue;
      }
      Term term = logic.deriveTerm(value, pre);
      Formula p = d instanceof ObjectVariableDeclaration ? oformula : cformula;
      if (pre.size() != 0)
      {
        Formula goal = Satisfies.impliesFormula(p, Satisfies.andFormula(pre));
        csymbol.getTasks().add(
            new InitializationTask(vsymbol, goal, 
                csymbol.getTasks().getDirectory(), theories, decls, error));
      }
      Formula f =
          EqualsFormula.construct(
              Satisfies.preVariable(NameUtils.uniqueName(vsymbol)),
              term);
      if (d instanceof ObjectVariableDeclaration)
        oformula = Satisfies.andFormula(oformula, f);
      else 
        cformula = Satisfies.andFormula(cformula, f);
    }
    csymbol.setInitFormula(oformula);
  }
  
  /**
   * 
   * The task of proving that an initialization precondition is satisfied.
   *
   */
  private static final class InitializationTask extends StateTaskBase
  {
    private final GlobalVariableSymbolBase osymbol;
    private final Formula goal;
    private final Vector<TheorySymbol> theories;
    private final Declaration[][] decls;
    private final ErrorStream out;

    /***************************************************************************
     * Construct an initialization task for a given object variable.
     * @param method the symbol of a method with a specification.
     * @param goal the goal to be proved.
     * @param parent the parent of the new directory for this task
     * @param stateVars the names of the modifiable variables (null, if none)
     * @param stateExcs the names of the throwable exceptions (null, if none)
     * @param theories the theories on which the following depend in dependence order.
     * @param decls a sequence of declaration lists.
     * @param out the stream to which to write error messages.
     **************************************************************************/
    public InitializationTask(GlobalVariableSymbolBase osymbol, Formula goal,
      File parent, Vector<TheorySymbol> theories, Declaration[][] decls,
      ErrorStream out)
    {
      super("[" + osymbol.getClassSymbol().getLongName() + "." + 
          osymbol.getName() + "] " +
          (osymbol instanceof ObjectVariableSymbol ? "object" : "class") +
          " variable initialization precondition", false,
          osymbol.getDeclaration().getPosition(), parent, 
          new ManualStrategy());
      this.osymbol = osymbol;
      this.goal = goal;
      this.theories = theories;
      this.decls = decls;
      this.out = out;
    }
    
    /*************************************************************************
     * Print linear representation of task.
     * @param out the stream to write to.
     ************************************************************************/
    public void print(PrintWriter out)
    {
      super.print(out);
      out.println("Type: verify initialization precondition");
    }
    
    /*************************************************************************
     * Translate task into a proving problem in state logic.
     * @return the translation result
     ************************************************************************/
    public StateProblem stateProblem()
    {
      ClassSymbol csymbol = osymbol.getClassSymbol();
      FlattenDeclarations.Result result = 
        FlattenDeclarations.flatten(theories, decls, goal, csymbol.getLocalTheory());
      Type stringType = csymbol.getPackage().getTypeTranslator().getStringType();
      return new StateProblem(result.decls, result.goal, result.map,
          null, stringType, new SetClass<VariableSymbol>(), 
          new SetClass<ClassSymbol>(), out);    
    }
  }
}
