// ----------------------------------------------------------------------------
// FormulaTask.java
// The task of proving a user-defined formula.
// $Id: FormulaTask.java,v 1.2 2010/11/08 17:07:44 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.io.*;
import java.util.Vector;

import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Proving.*;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * 
 * The task of proving a user-defined formula.
 *
 */
public class FormulaTask extends ClassicalTaskBase
{
  private FormulaDefinition formula;
  private TheorySymbol theory;
  private ErrorStream out;
  
  /***************************************************************************
   * Construct a task.
   * @param formula the definition of the formula to be proved.
   * @param theory the theory in which the definition occurs.
   * @param parent the parent directory.
   * @param out stream for printing errors
   **************************************************************************/
  public FormulaTask(FormulaDefinition formula, TheorySymbol theory, 
    File parent, ErrorStream out)
  {
    super("Formula " + formula.getSymbol().getLongName(), 
        false, formula.getPosition(), parent, new ManualStrategy());
    this.formula = formula;
    this.theory = theory;
  }
  
  /****************************************************************************
   * Translate task into a classical proving problem
   * @return the translation.
   ***************************************************************************/
  public ClassicalProblem classicalProblem()
  {
    // determine predecessor declarations of formula
    TheoryDeclaration tdecl = theory.getDeclaration();
    Declaration[] decls = tdecl.getDeclarations();
    Vector<Declaration> decls0 = new Vector<Declaration>();
    for (Declaration decl : decls)
    {
      if (decl == formula) break;
      decls0.add(decl);
    }
    int n = decls0.size();
    Declaration[] decls1 = new Declaration[n];
    decls0.toArray(decls1);
    Declaration[][] declarations = new Declaration[][] { decls1 };
    Vector<TheorySymbol> theories = new Vector<TheorySymbol>();
    TheorySymbol baseTheory = theory.getPackage().getTopPackage().getTypeTranslator().getBaseTheory();
    if (baseTheory != null) TheorySymbol.add(theories, baseTheory);
    TheorySymbol.add(theories, NameUtils.getTheorySymbols(tdecl.getTheories()));
    FlattenDeclarations.Result result = 
      FlattenDeclarations.flatten(theories, declarations, formula.getValue(), theory);
    
    // construct classical problem
    return new ClassicalProblem(result.decls, result.goal, out);
  }
}
