// ----------------------------------------------------------------------------
// ProofNavigatorTask.java
// The base of a task suitable for solving with the RISC ProofNavigator.
// $Id: ProofNavigatorTaskBase.java,v 1.7 2011/04/13 16:32:28 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.ewuni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.io.*;

import fmrisc.ProgramExplorer.Main;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Proving.*;
import fmrisc.ProofNavigator.State;
import fmrisc.ProofNavigator.Proving.Proof;

/**
 * 
 * The base of a task suitable for solving with the RISC ProofNavigator.
 *
 */
public abstract class ProofNavigatorTaskBase extends TaskBase 
  implements ProofNavigatorTask
{
  /****************************************************************************
   * Construct task with denoted name and status NEW
   * @param name the name of the task
   * @param optional is this task optional?
   * @param position a related source position (may be null)
   * @param parent the parent of the directory created for writing 
   *        persistent information related to this task.
   ***************************************************************************/
  protected ProofNavigatorTaskBase(String name, boolean optional,
    SourcePosition position, File parent, Strategy strategy)
  {
    super(name, optional, position, parent, strategy);
  }

  /***************************************************************************
   * Compute status information considering previous proof attempts.
   **************************************************************************/
  protected void computeStatus()
  {
    ProofNavigatorProblem problem = proofNavigatorProblem();
    Proof proof = problem.getProof();
    if (proof == null) return;
    Task.Status status0 = 
      proof.isClosed() && proof.isAbsolute() && proof.isTrusted() 
      ? Task.Status.ALMOST : Task.Status.OPEN; 
    String status = proof.getStatus();
    setStatus(status0, status, proof);
  }
 
  /***************************************************************************
   * Set parent folder
   * @param parent the parent folder (may be null)
   **************************************************************************/
  public void setParent(TaskFolder parent)
  {
    computeStatus();
    super.setParent(parent);
  }
  
  /****************************************************************************
   * Reset the task to its initial (new) state.
   ***************************************************************************/
  public void reset()
  {
    if (State.getProof() != null)
    {
      ErrorStream error = Main.getError();
      error.newMessage("cannot reset task: close ongoing proof first.");
      return;
    }
    File dir = getDirectory();
    Directory.free(dir, true);
    setStatus(Status.NEW, null, null);
  }
}
