// ----------------------------------------------------------------------------
// TerminationTask.java
// The task of proving the termination of a statement.
// $Id: TerminationTask.java,v 1.10 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2011-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.util.*;
import java.io.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Judgements.Simplification;
import fmrisc.ProgramExplorer.Proving.StateProblem;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Program.Statement;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * 
 * The task of proving the termination of a statement.
 *
 */
public final class TerminationTask extends StateTaskBase
{
  private ParamSymbol method;
  private Formula goal;
  private Statement stat;
  private String tag;
  private StateType stateType;
  private Set<VariableSymbol> variables;
  private Set<ClassSymbol> exceptions;
  private Vector<TheorySymbol> theories;
  private Declaration[][] decls;
  private ErrorStream out;
  
  /***************************************************************************
   * Construct a termination verification task.
   * @param kind a description of the kind of the task.
   * @param method the method in which the task arises.
   * @param goal the goal to be proved (constructed outside)
   * @param stat the statement whose termination is proved (may be null)
   * @param tag the tag to use for the task if statement is null
   * @param parent the parent file
   * @param stateType the type of the current state (null, if none)
   * @param variables the names of the variables modified in the loop body.
   * @param exceptions the names of the exceptions thrown in the loop body.
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   **************************************************************************/
  public TerminationTask(String kind,
    ParamSymbol method, Formula goal, Statement stat, 
    String tag, File parent, StateType stateType, 
    Set<VariableSymbol> variables, Set<ClassSymbol> exceptions,
    Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream out)
  {
    super(taskName(kind, method, stat, tag), false, 
        stat == null ? null :
          stat.getPosition(), parent, new AutoStrategy());
    this.method = method;
    this.goal = Simplification.perform(goal);
    this.stat = stat;
    this.tag = tag;
    this.stateType = stateType;
    this.variables = variables;
    this.exceptions = exceptions;
    this.theories = theories;
    this.decls = decls;
    this.out = out;
  }
  
  /****************************************************************************
   * Compute from invariant formula a specific task name.
   * @param kind a description of the kind of the task.
   * @param method the method for the task.
   * @param stat the statement whose termination is to be proved (may be null)
   * @param tag the tag to be used if the statement is null.
   * @return a name that is probably different for different statements.
   ***************************************************************************/
  private static String taskName(String kind,
    ParamSymbol method, Statement stat, String tag)
  {
    if (stat == null) 
      return "[" + method.getLongName() + "] " + kind;
    String gname = stat.toStringCore();
    String key = Util.hash(gname, 3);
    return "[" + method.getLongName() + ":" + key + "] " + kind;
  }
  
  /****************************************************************************
   * Determine whether new task is to be resumed.
   * @return true if new task is to be resumed.
   ***************************************************************************/
  public boolean resumeOnNew()
  {
    return true;
  }
  
  /****************************************************************************
   * Print linear representation of task.
   * @param out the stream to write to.
   ***************************************************************************/
  public void print(PrintWriter out)
  {
    super.print(out);
    if (stat != null)
    {
      out.println("Type: verify termination");
      out.println("Statement: " + shortDescription(stat.toString()));
    }
    else
    {
      out.println("Type: verify termination(" + tag + ")");
    }
  }
  
  /****************************************************************************
   * Translate task into a proving problem in state logic.
   * @return the translation result
   ***************************************************************************/
  public StateProblem stateProblem()
  {
    ClassSymbol csymbol = method.getClassSymbol();
    FlattenDeclarations.Result result = 
      FlattenDeclarations.flatten(theories, decls, goal, csymbol.getLocalTheory());
    Type stringType = csymbol.getPackage().getTypeTranslator().getStringType();
    return new StateProblem(result.decls, result.goal, result.map,
        stateType, stringType, variables, exceptions, out); 
  }
}
