// ----------------------------------------------------------------------------
// Flip.java
// The "flip" command.
// $Id: Flip.java,v 1.12 2010/03/03 14:01:14 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.PrintWriter;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import fmrisc.ProofNavigator.State;
import fmrisc.ProofNavigator.Proving.*;

/**
 * 
 * Command "flip": flip a formula from goal position to assumption position 
 * or vice versa. 
 * 
 */
public final class Flip extends ProofCommandBase
{
  public final static String Name = "flip";
  private String label; // the label of the formula to be flipped
  
  /***************************************************************************
   * create a "flip" command
   * @param label the label of the formula to be flipped
   **************************************************************************/
  public Flip(String label)
  {
    super(Name);
    this.label = label;
  }
  
  /****************************************************************************
   * process command "flip label": flip formula denoted by label from goal
   * position to target position or vice versa.
   * @param state the proof state to which the command is applied
   * @return null or array of single new proof state.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    ProofState state0 = new ProofState(state, state.getEnvironment());
    Formula formula = state.getAssumption(label);
    Formula[] assumptions0 = null;
    Formula[] goals0 = null;
    
    if (formula != null)
    {
      assumptions0 = state0.remove(state.getAssumptions(), formula);
      Formula formula0 = new Formula(state0, 
          ProofUtil.negate(formula.getExpression()),
          !formula.isGoal());
      goals0 = state0.add(state.getGoals(), formula0);
    }
    else
    {
      formula = state.getGoal(label);
      if (formula != null)
      {
        goals0 = state0.remove(state.getGoals(), formula);
        Formula formula0 = new Formula(state0, 
            ProofUtil.negate(formula.getExpression()),
            !formula.isGoal());
        assumptions0 = state0.add(state.getAssumptions(), formula0);
      }
      else
      {
        error("no formula with label " + label + " in current state");
        return null;
      }
    }
    
    state0.setAssumptions(assumptions0);
    state0.setGoals(goals0);
    return new ProofState[] { state0 };
  }

  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    node.appendChild(newTextNode(label));
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    if (children.getLength() != 1) 
    {
      State.setError("child expected: " + node.toString());
      return null;
    }
    String label = toString(children.item(0));
    if (label == null) return null;
    return new Flip(label);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    out.print(" ");
    out.print(label);
  }
}
