// ----------------------------------------------------------------------------
// Scatter.java
// The "scatter" command.
// $Id: Scatter.java,v 1.5 2013/10/31 14:09:54 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.*;
import java.util.*;
import org.w3c.dom.*;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Proving.*;

/**
 * 
 * command "scatter": scatter proof state by repeated decomposition
 * and goalsplitting.
 *
 */
public final class Scatter extends ProofCommandBase
{
  public final static String Name = "scatter";
  private static ProofCommand decompose = new Decompose();
  private static ProofCommand split = new Split(null);
  
  /***************************************************************************
   * create a "scatter" command
   **************************************************************************/
  public Scatter()
  {
    super(Name);
  }
  
  /****************************************************************************
   * process command "scatter": 
   * @param state the proof state to which the command is applied
   * @return null or array of single new proof state.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    Vector<ProofState> open = new Vector<ProofState>();   // queue of states to be processed
    Vector<ProofState> closed = new Vector<ProofState>(); // collection of results
    ProofState original = state;                          // original state
    
    open.addElement(state);
    while (!open.isEmpty())
    {
      ProofState state0 = (ProofState)open.remove(0);
      
      // decompose will not simplify its original state, thus we have to do it
      // but also we do not simplify the original state (visible to user)
      if (state0 != original) state0.simplify(false);
      
      // decompose may create a single successor state
      ProofState[] state1 = decompose.process(state0);
      if (state1 != null) state0 = state1[0];
      
      // split may create multiple successor states
      ProofState[] state2 = split.process(state0);
     
      // state could not be split
      if (state2 == null)
      {
        // add state unless it is the original one
        if (state0 != state) closed.add(state0);
        continue;
      }
      
      // add results of state split
      int n = state2.length;
      for (int i=0; i<n; i++)
        open.addElement(state2[i]);
    }
    
    // split may trigger error
    State.resetError();
    
    // if no state was generated, rule is not applicable
    int n = closed.size();
    if (n == 0) return null;
    
    // construct children states
    ProofState[] children = new ProofState[n];
    for (int i=0; i< n; i++)
    {
      ProofState child = (ProofState)closed.elementAt(i);
      child.setParent(state);
      children[i] = child;
    }
    return children;
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    return new Scatter();
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
  }
}
