// ----------------------------------------------------------------------------
// OMDoc.java
// Conversion of AST declarations from/to OMDoc declarations.
// $Id: OMDoc.java,v 1.11 2012/10/16 15:49:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Communication;

import java.io.*;
import org.w3c.dom.*;

import fmrisc.External.*;
import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;

/**
 * 
 * Conversion of AST declarations from/to OMDoc declarations.
 *
 */
public final class OMDoc
{
  private Document document;     // the document in which to create new nodes
  private NewOMDOMWriter writer; // writer to convert OpenMath to DOM
    
  /***************************************************************************
   * Create OMDoc converter for denoted document.
   * @param document document in which to create new nodes.
   *************************************************************************/
  public OMDoc(Document document)
  {
    this.document = document;
    this.writer = new NewOMDOMWriter(document, 
        State.OpenMathURI, State.OpenMathPrefix);
  }
  
  /***************************************************************************
   * Create new OMDoc node with denoted name.
   * @param name the name of the node.
   * @return a node with the denoted name in the current document.
   **************************************************************************/
  private Element createNode(String name)
  {
    return document.createElementNS(State.OMDocURI,
        State.OMDocPrefix + ":" + name);
  }
  
  /***************************************************************************
   * Create new OpenMath node with denoted name.
   * @param name the name of the node.
   * @return a node with the denoted name in the current document.
   **************************************************************************/
  private Element createOpenMathNode(String name)
  {
    return document.createElementNS(State.OpenMathURI,
        State.OpenMathPrefix + ":" + name);
  }
  
  /***************************************************************************
   * Convert AST object to DOM node.
   * @param ast the AST object.
   * @return the corresponding DOM node (holding the OpenMath representation)
   **************************************************************************/
  private Node astToNode(AST ast)
  {
    try
    {
      return writer.writeNode(OpenMath.toOpenMath(ast));
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
  }
  
  /***************************************************************************
   * Convert declaration to OMDoc representation in DOM format.
   * @param decl a declaration
   * @return DOM object holding the OMDoc representation.
   **************************************************************************/
  public Element toNode(Declaration decl)
  {
    Element parent = createNode("omgroup");
    parent.setAttribute("xmlns:" + State.NamespacePrefix, State.NamespaceURI);
    parent.setAttribute("xmlns:" + State.OpenMathPrefix, State.OpenMathURI);
    parent.setAttribute("xmlns:" + State.OMDocPrefix, State.OMDocURI);
    if (decl instanceof FormulaDeclaration)
    {
      FormulaDeclaration fdecl = (FormulaDeclaration)decl;
      Element node = createNode("assertion");
      node.setAttributeNS(State.XMLURI, State.XMLPrefix + ":id", 
        fdecl.getIdentifier().getName());
      if (decl instanceof AxiomDeclaration)
        node.setAttribute("type", "axiom");
      else
        node.setAttribute("type", "formula");
      Element node0 =  createOpenMathNode("FMP");
      node0.appendChild(astToNode(fdecl.getFormula()));
      node.appendChild(node0);
      parent.appendChild(node);
    }
    else if (decl instanceof TypeDeclaration)
    {
      TypeDeclaration tdecl = (TypeDeclaration)decl;
      Element node = createNode("symbol");
      String name =  tdecl.getIdentifier().getName();
      node.setAttribute("name", name);
      node.setAttribute("kind", "type");
      parent.appendChild(node);
      Type type = tdecl.getType();
      if (type != null)
      {
        node = createNode("definition");
        node.setAttribute("name", name + "_def");
        node.setAttribute("for", "#" + name);
        node.setAttribute("type", "simple");
        node.appendChild(astToNode(type));
        parent.appendChild(node);
      }
    }
    else if (decl instanceof ValueDeclaration)
    {
      ValueDeclaration vdecl = (ValueDeclaration)decl;
      Element node = createNode("symbol");
      String name =  vdecl.getIdentifier().getName();
      node.setAttribute("name", name);
      node.setAttribute("kind", "object");
      parent.appendChild(node);
      Type type = vdecl.getType();
      Element node0 = createNode("type");
      node0.setAttributeNS(State.XMLURI, State.XMLPrefix + ":id", 
        name + "_type");
      node0.setAttribute("system", "simply_typed"); // ??? from OMDoc manual
      node0.appendChild(astToNode(type));
      node.appendChild(node0);
      AST value = vdecl.getValue();
      if (value != null)
      {
        node = createNode("definition");
        node.setAttribute("name", name + "_def");
        node.setAttribute("for", "#" + name);
        node.setAttribute("type", "simple");
        node.appendChild(astToNode(value));
        parent.appendChild(node);
      }
    }
    return parent;
  }
  
  /***************************************************************************
   * Convert OMDoc representation to declaration.
   * @param node DOM object holding the OMDoc representation of a declaration.
   * @return the corresponding declaration (null in case of error)
   *         where the body of the declaration is already type-checked.
   **************************************************************************/
  public static Declaration toDeclaration(Node node)
  {
    if (!(node instanceof Element))
    {
      State.setError("element expected: " + node.toString());
      return null;
    }
    String name = node.getLocalName();
    if (!name.equals("omgroup"))
    {
      State.setError("omgroup expected: " + node.toString());
      return null;
    }
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    if (n == 0)
    {
      State.setError("child expected: " + node.toString());
      return null;
    }
    Node child0 = children.item(0);
    String name0 = child0.getLocalName();
    
    if (name0.equals("assertion"))
    {
      String tfield = ((Element)node).getAttributeNS(State.XMLURI, "type");
      if (tfield == null)
      {
        State.setError("attribute 'type' expected: " + node.toString());
        return null;
      }
      String ident = 
        ((Element)node).getAttributeNS(State.XMLURI, State.XMLPrefix + ":id");
      if (ident == null)
      {
        State.setError("attribute '" + State.XMLPrefix + ":id' expected: " +
         node.toString());
        return null;
      }
      Element node0 = (Element)child0;
      NodeList children0 = child0.getChildNodes();
      if (children0.getLength() != 1)
      {
        State.setError("one child expected: " + node.toString());
        return null;
      }
      Node node1 = children0.item(0);
      NodeList children1 = node1.getChildNodes();
      if (children1.getLength() != 1)
      {
        State.setError("one child expected: " + node0.toString());
        return null;
      }
      Expression formula = Store.toExpression(children1.item(0));
      if (formula == null) return null;
      Type type = Checking.processFormula(formula);
      if (type == null) return null;
      FormulaDeclaration decl;
      if (tfield.equals("formula"))
        decl = Construct.formulaDeclaration(
          Construct.formulaDeclIdentifier(ident), formula);
      else if (tfield.equals("axiom"))
        decl = Construct.axiomDeclaration(
          Construct.formulaDeclIdentifier(ident), formula);
      else
      {
        State.setError("Field 'type' must have value \"formula\" " +
          "or \"axiom\":" + node.toString());
        return null;
      }
      return decl;
    }
    
    if (name0.equals("symbol"))
    {
      if (!(child0 instanceof Element))
      {
        State.setError("element expected: " + child0.toString());
        return null;
      }
      Element node0 = (Element)child0;
      String ident = node0.getAttribute("name");
      if (ident == null)
      {
        State.setError("attribute 'name' expected: " + child0.toString());
        return null;
      }
      String kind = node0.getAttribute("kind");
      if (kind == null)
      {
        State.setError("attribute 'kind' expected: " + child0.toString());
        return null;
      }
      if (kind.equals("type"))
      {
        if (n == 1) 
          return Construct.typeDeclaration(
            Construct.typeDeclIdentifier(ident), null);
        if (n == 2)
        {
          Node child1 = children.item(1);
          if (!(child1 instanceof Element))
          {
            State.setError("element expected: " + child1.toString());
            return null;
          }
          Element node1 = (Element)child1;
          NodeList children1 = node1.getChildNodes();
          if (children1.getLength() != 1)
          {
            State.setError("One child expected: " + node1.toString());
            return null;
          }
          Type type0 = Store.toType(children1.item(0));
          if (type0 == null) return null;
          Type type1 = Checking.canonical(type0);
          if (type1 == null) return null;
          return Construct.typeDeclaration(
              Construct.typeDeclIdentifier(ident), type1);
        }
        State.setError("One or two children expected: " + node.toString());
        return null;
      }
      if (kind.equals("object"))
      {
        if (n != 1 && n != 2)
        {
          State.setError("One or two children expected: " + node.toString());
          return null;  
        }
        NodeList children0 = node0.getChildNodes();
        if (children0.getLength() != 1)
        {
          State.setError("One child expected: " + node0.toString());
          return null;
        }
        Node child1 = children0.item(0);
        if (!(child1 instanceof Element))
        {
          State.setError("element expected: " + child1.toString());
          return null;
        }
        Element node1 = (Element)child1;
        NodeList children1 = node1.getChildNodes();
        if (children1.getLength() != 1)
        {
          State.setError("One child expected: " + node1.toString());
          return null;
        }
        Type type = Store.toType(children1.item(0));
        if (type == null) return null;
        type = Checking.canonical(type);
        if (type == null) return null;
        if (n == 1)
          return Construct.valueDeclaration(Construct.valueDeclIdentifier(ident),
            type, null);
        Node child3 = children.item(1);
        if (!(child3 instanceof Element))
        {
          State.setError("element expected: " + child3.toString());
          return null;
        }
        Element node3 = (Element)child3;
        NodeList children3 = node3.getChildNodes();
        if (children3.getLength() != 1)
        {
          State.setError("One child expected: " + node3.toString());
          return null;
        }
        Expression exp = Store.toExpression(children3.item(0));
        if (exp == null) return null;
        // Type etype = Checking.process(exp);
        // if (etype == null) return null;
        return Construct.valueDeclaration(Construct.valueDeclIdentifier(ident),
          type, exp);
      }
      State.setError("Unknown kind: " + child0.toString());
      return null;
    }
    
    State.setError("unknown declaration: " + node.toString());
    return null;
  }
}
