// ----------------------------------------------------------------------------
// OpenMath.java
// Conversion of AST objects from/to OpenMath representation.
// $Id: OpenMath.java,v 1.15 2012/10/16 15:49:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Communication;

import nl.tue.win.riaca.openmath.lang.*;
import java.util.*;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Syntax.*;

/**
 * 
 * Conversion of AST objects (excluding declarations) 
 * from/to OpenMath representation.
 * 
 * Uses the following OM content dictionaries:
 * - sts, arith1, arith2, relation1, logic1, quant1, fns1,
 *   set1, setname1, setname2, interval1.
 * 
 * Introduces the non-standard content dictionary "fmrisc" with the
 * following symbols (for which no adequate OM CD symbol seems to exist):
 * - if: conditional expression
 * - let: local binding expression
 * - ldecl: local value declaration
 * - array: array quantifier
 * - record: record expression
 * - select: select expression
 * - tuple: tuple expression
 * - update: update expression
 * - arraytype: array type
 * - recordtype: record type
 * - subtype: subtype.
 * - selident: selection identifier.
 * - selindex: selection index.
 * - selnumber: selection number.
 */
public final class OpenMath
{      
  /***************************************************************************
   * Convert AST to OpenMath object.
   * @param ast the object in AST representation.
   * @return the OpenMath representation of the object.
   **************************************************************************/
  public static OMObject toOpenMath(AST ast)
  {
    if (ast instanceof fmrisc.ProofNavigator.Syntax.Number)
    {
      fmrisc.ProofNavigator.Syntax.Number object = (fmrisc.ProofNavigator.Syntax.Number)ast; 
      return new OMInteger(object.getDigits());
    }
    if (ast instanceof Identifier)
    {
      Identifier object = (Identifier)ast;
      // collect symbols referenced in proof
      Proof proof = State.getProof();
      if (proof != null && !object.getName().equals("_")) proof.referenced(object);
      return new OMVariable(object.getName());
    }
    if (ast instanceof TypedIdentifier)
    {
      TypedIdentifier object = (TypedIdentifier)ast;
      OMAttribution om = new OMAttribution();
      om.setConstructor(new OMVariable(object.getIdentifier().getName()));
      om.put(new OMSymbol("sts", "type"), toOpenMath(object.getType()));
      return om;
    }
    if (ast instanceof ValuedIdentifier)
    {
      ValuedIdentifier object = (ValuedIdentifier)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("relation1", "eq"));
      om.addElement(toOpenMath(object.getIdentifier()));
      om.addElement(toOpenMath(object.getValue()));
      return om;
    }
     
    /*************************************************************************
     * expressions
     ************************************************************************/
    if (ast instanceof ApplicationExpression)
    {
      ApplicationExpression object = (ApplicationExpression)ast;
      OMApplication om = new OMApplication();
      om.addElement(toOpenMath(object.getFunction()));
      Expression[] args = object.getArguments();
      int n = args.length;
      for (int i=0; i<n; i++)
        om.addElement(toOpenMath(args[i]));
      return om;
    }
    if (ast instanceof BinaryExpression)
    {
      BinaryExpression object = (BinaryExpression)ast;
      OMObject fun = null;
      if (object instanceof AndFormula)
        fun = new OMSymbol("logic1", "and");
      else if (object instanceof DividesTerm)
        fun = new OMSymbol("arith1", "divide");
      else if (object instanceof EqualsFormula)
        fun = new OMSymbol("relation1", "eq");
      else if (object instanceof EquivalentFormula)
        fun = new OMSymbol("logic1", "equivalent");
      else if (object instanceof GreaterEqualFormula)
        fun = new OMSymbol("relation1", "geq");
      else if (object instanceof GreaterFormula)
        fun = new OMSymbol("relation1", "gt");
      else if (object instanceof ImpliesFormula)
        fun = new OMSymbol("logic1", "implies");
      else if (object instanceof LessEqualFormula)
        fun = new OMSymbol("relation1", "leq");
      else if (object instanceof LessFormula)
        fun = new OMSymbol("relation1", "lt");
      else if (object instanceof MinusTerm)
        fun = new OMSymbol("arith1", "minus");
      else if (object instanceof NotEqualsFormula)
        fun = new OMSymbol("relation1", "neq");
      else if (object instanceof NotEquivalentFormula)
        fun = new OMSymbol("logic1", "xor");
      else if (object instanceof OrFormula)
        fun = new OMSymbol("logic1", "or");
      else if (object instanceof PlusTerm)
        fun = new OMSymbol("arith1", "plus");
      else if (object instanceof BitOrTerm)
        fun = new OMSymbol("arith1", "plus");
      else if (object instanceof PowerTerm)
        fun = new OMSymbol("arith1", "power");
      else if (object instanceof TimesTerm)
        fun = new OMSymbol("arith2", "times"); // commutative times
      else if (object instanceof BitAndTerm)
        fun = new OMSymbol("arith2", "times"); // commutative times
      else 
      {
        System.out.println("Unknown binary expression in OpenMath conversion: " + object.toString());
        System.exit(-1);
      }
      OMApplication om = new OMApplication();
      om.addElement(fun);
      om.addElement(toOpenMath(object.getFirst()));
      om.addElement(toOpenMath(object.getSecond()));
      return om;
    }
    if (ast instanceof IfThenElseExpression)
    {
      IfThenElseExpression object = (IfThenElseExpression)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("fmrisc", "if"));
      om.addElement(toOpenMath(object.getCondition()));
      om.addElement(toOpenMath(object.getThenBranch()));
      om.addElement(toOpenMath(object.getElseBranch()));
      return om;
    }
    if (ast instanceof LetExpression)
    {
      LetExpression object = (LetExpression)ast;
      OMApplication om =
        new OMApplication();
      om.addElement(new OMSymbol("fmrisc", "let"));
      ValueDeclaration[] decls = object.getDeclarations();
      int n = decls.length;
      for (int i=0; i<n; i++)
        om.addElement(toOpenMathValueDecl(decls[i]));
      om.addElement(toOpenMath(object.getBase()));
      return om;
    }
    if (ast instanceof Logical)
    {
      Logical object = (Logical)ast;
      if (object.getValue())
        return new OMSymbol("logic1", "true");
      else
        return new OMSymbol("logic1", "false");
    }
    if (ast instanceof BitLogical)
    {
      Logical object = (Logical)ast;
      if (object.getValue())
        return new OMInteger("1");
      else
        return new OMInteger("0");
    }
    if (ast instanceof QuantifiedExpression)
    {
      QuantifiedExpression object = (QuantifiedExpression)ast;
      OMObject quantifier = null;
      if (object instanceof ForallFormula)
        quantifier = new OMSymbol("quant1", "forall");
      else if (object instanceof ExistsFormula)
        quantifier = new OMSymbol("quant1", "exists");
      else if (object instanceof LambdaTerm)
        quantifier = new OMSymbol("fns1", "lambda");
      else if (object instanceof ArrayTerm)
        quantifier = new OMSymbol("fmrisc", "array");
      else
      {
        System.out.println("unknown quantifier: " + object.toString());
        System.exit(-1);
      }
      OMBinding om = new OMBinding();
      om.setBinder(quantifier);
      TypedIdentifier[] vars = object.getVariables();
      int n = vars.length;
      for (int i=0; i<n; i++)
        om.addVariable(toOpenMath(vars[i]));
      om.setBody(toOpenMath(object.getBase()));
      return om;
    }
    if (ast instanceof RecordTerm)
    {
      RecordTerm object = (RecordTerm)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("fmrisc", "record"));
      ValuedIdentifier[] base = object.getBase();
      int n = base.length;
      for (int i=0; i<n; i++)
        om.addElement(toOpenMath(base[i]));
      return om;
    }
    if (ast instanceof SelectionTerm)
    {
      SelectionTerm object = (SelectionTerm)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("fmrisc", "select"));
      om.addElement(toOpenMath(object.getBase()));
      om.addElement(toOpenMath(object.getSelector()));
      return om;
    }
    if (ast instanceof TupleTerm)
    {
      TupleTerm object = (TupleTerm)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("fmrisc", "tuple"));
      Expression[] base = object.getBase();
      int n = base.length;
      for (int i=0; i<n; i++)
        om.addElement(toOpenMath(base[i]));
      return om;
    }
    if (ast instanceof UnaryExpression)
    {
      UnaryExpression object = (UnaryExpression)ast;
      OMSymbol fun = null;
      if (object instanceof NegationTerm)
        fun = new OMSymbol("arith1", "unary_minus");
      else if (object instanceof NotFormula)
        fun = new OMSymbol("logic1", "not");
      else if (object instanceof BitNotTerm)
      {
        OMApplication om = new OMApplication();
        om.addElement(new OMSymbol("arith1", "minus"));
        om.addElement(new OMInteger("1"));
        om.addElement(toOpenMath(object.getBase()));
        return om;
      }
      else
      {
        System.out.println("unknown unary operation: " + object.toString());
        System.exit(-1);
      }
      OMApplication om = new OMApplication();
      om.addElement(fun);
      om.addElement(toOpenMath(object.getBase()));
      return om;
    }
    if (ast instanceof UpdateTerm)
    {
      UpdateTerm object = (UpdateTerm)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("fmrisc", "update"));
      om.addElement(toOpenMath(object.getBase()));
      Selector[] sels = object.getSelectors();
      int n = sels.length;
      for (int i=0; i<n; i++)
        om.addElement(toOpenMath(sels[i]));
      om.addElement(toOpenMath(object.getUpdate()));
      return om;
    } 
    
    /*************************************************************************
     * types
     ************************************************************************/
    if (ast instanceof ArrayType)
    {
      ArrayType object = (ArrayType)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("fmrisc", "arraytype"));
      om.addElement(toOpenMath(object.getIndex()));
      om.addElement(toOpenMath(object.getBase()));
      return om;
    }
    if (ast instanceof AtomicType)
    {
      AtomicType object = (AtomicType)ast;
      String name = object.getName();
      if (name.equals("BOOLEAN"))
        return new OMSymbol("setname2", "Boolean");
      if (name.equals("BIT"))
        return new OMSymbol("setname1", "N");
      if (name.equals("NAT"))
        return new OMSymbol("setname1", "N");
      if (name.equals("INT"))
        return new OMSymbol("setname1", "Z");
      if (name.equals("REAL"))
        return new OMSymbol("setname1", "R");
      return new OMVariable(name);
    }
    if (ast instanceof FunctionType)
    {
      FunctionType object = (FunctionType)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("sts", "mapsto"));
      OMApplication om0 = new OMApplication();
      om0.addElement(new OMSymbol("set1", "cartesian_product"));
      Type[] domain = object.getDomain();
      int n = domain.length;
      for (int i=0; i<n; i++)
        om0.addElement(toOpenMath(domain[i]));
      om.addElement(om0);
      om.addElement(toOpenMath(object.getRange()));
      return om;
    }
    if (ast instanceof RecordType)
    {
      RecordType object = (RecordType)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("fmrisc", "recordtype"));
      TypedIdentifier[] base = object.getBase();
      int n = base.length;
      for (int i=0; i<n; i++)
        om.addElement(toOpenMath(base[i]));
      return om;
    }
    if (ast instanceof SubrangeType)
    {
      SubrangeType object = (SubrangeType)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("interval1", "integer_interval"));
      om.addElement(toOpenMath(object.getFrom()));
      om.addElement(toOpenMath(object.getTo()));
      return om;
    }
    if (ast instanceof SubType)
    {
      SubType object = (SubType)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("fmrisc", "subtype"));
      om.addElement(toOpenMath(object.getPredicate()));
      return om;  
    }
    if (ast instanceof TupleType)
    {
      TupleType object = (TupleType)ast;
      OMApplication om = new OMApplication();
      om.addElement(new OMSymbol("set1", "cartesian_product"));
      Type[] base = object.getBase();
      int n = base.length;
      for (int i=0; i<n; i++)
        om.addElement(toOpenMath(base[i]));
      return om;
    }
    
    /*************************************************************************
     * selectors
     ************************************************************************/
    if (ast instanceof Selector)
    {
      Selector selector= (Selector)ast;
      OMApplication om = new OMApplication();
      if (selector instanceof SelectorIdentifier)
      {
        SelectorIdentifier sel = (SelectorIdentifier)selector;
        om.addElement(new OMSymbol("fmrisc", "selident"));
        om.addElement(toOpenMath(sel.getIdentifier()));
      }
      else if (selector instanceof SelectorIndex)
      {
        SelectorIndex sel = (SelectorIndex)selector;
        om.addElement(new OMSymbol("fmrisc", "selindex"));
        om.addElement(toOpenMath(sel.getIndex()));
      }
      else if (selector instanceof SelectorNumber)
      {
        SelectorNumber sel = (SelectorNumber)selector;
        om.addElement(new OMSymbol("fmrisc", "selnumber"));
        om.addElement(toOpenMath(sel.getNumber()));
      }
      return om;
    }
    
    /*************************************************************************
     * error case
     ************************************************************************/
    System.out.println("Unknown kind of object: " + ast.toString());
    System.exit(-1);
    return null;
  }
  
  /***************************************************************************
   * Convert OpenMath expression to AST.
   * @param om the expression in OpenMath representatin.
   * @return the expression as an AST (null in case of error).
   **************************************************************************/
  public static Expression toExpression(OMObject om)
  {
    Expression exp = toExpressionCore(om);
    if (State.getError() != null) return null;
    return exp;
  }
  
  /***************************************************************************
   * Convert OpenMath type to AST.
   * @param om the type in OpenMath representatin.
   * @return the type as an AST (null in case of error).
   **************************************************************************/
  public static Type toType(OMObject om)
  {
    Type type = toTypeCore(om);
    if (State.getError() != null) return null;
    return type;
  }
  
  /***************************************************************************
   * Convert OpenMath object to AST.
   * @param om the object in AST representation.
   * @return the AST representation of the object.
   **************************************************************************/
  private static AST toAST(OMObject om)
  {
    if (om == null)
    {
      State.setError("Null OpenMath object encountered");
      return null;
    }
    
    /**************************************************************************
     * top-level classes
     *************************************************************************/
    if (om instanceof OMInteger)
    {
      OMInteger object = (OMInteger)om;
      return Construct.number(object.getInteger());
    }
    if (om instanceof OMVariable)
    {
      OMVariable object = (OMVariable)om;
      return Construct.reference(object.getName());
    }
    if (om instanceof OMAttribution)
    {
      OMAttribution object = (OMAttribution)om;
      OMObject om0 = object.getConstructor();
      if (om0 == null)
      {
        State.setError("Empty constructor encountered: " + object.toString());
        return null;
      }
      if (!(om0 instanceof OMVariable))
      {
        State.setError("Variable expected: " + om0.toString());
        return null;
      }
      String name = ((OMVariable)om0).getName();
      ValueDeclIdentifier ident = Construct.valueDeclIdentifier(name);
      Enumeration e = object.getValues();
      if (!e.hasMoreElements())
      {
        State.setError("empty type: " + object.toString());
        return null;
      }
      OMObject omt = (OMObject)e.nextElement();
      Type type = toTypeCore(omt);
      return Construct.typedIdentifier(ident, type);
    }
    if (om instanceof OMSymbol)
    {
      OMSymbol object = (OMSymbol)om;
      String cd = object.getCd();
      if (cd == null)
      {
        State.setError("empty CD: " + object.toString());
        return null;
      }
      String name = object.getName();
      if (name == null)
      {
        State.setError("empty name: " + object.toString());
        return null;
      } 
      if (cd.equals("logic1") && name.equals("true"))
        return Construct.logical(true);
      if (cd.equals("logic1") && name.equals("false"))
        return Construct.logical(true);
      State.setError("Unknown symbol: " + om.toString());
      return null;
    }
    if (om instanceof OMBinding)
    {
      OMBinding object = (OMBinding)om;
      OMObject om0 = object.getBinder();
      if (om0 == null)
      {
        State.setError("binder expected: " + object.toString());
        return null;
      }
      if (!(om0 instanceof OMSymbol))
      {
        State.setError("symbol expected: " + om0.toString());
        return null;
      }
      OMSymbol object0 = (OMSymbol)om0;
      String cd = object0.getCd();
      if (cd == null)
      {
        State.setError("empty CD: " + object0.toString());
        return null;
      }
      String name = object0.getName();
      if (name == null)
      {
        State.setError("empty name: " + object0.toString());
        return null;
      } 
      Vector v = object.getVariables();
      int n = v.size();
      TypedIdentifier[] vars = new TypedIdentifier[n];
      for (int i=0; i<n; i++)
        vars[i] = toTypedIdentifier((OMObject)v.elementAt(i));
      OMObject om1 = object.getBody();
      Expression body = toExpressionCore(om1);
      if (cd.equals("quant1") && name.equals("forall"))
        return Construct.forallFormula(vars, (Expression)body);
      if (cd.equals("quant1") && name.equals("exists"))
        return Construct.existsFormula(vars, (Expression)body); 
      if (cd.equals("fns1") && name.equals("lambda"))
        return Construct.lambdaTerm(vars, (Expression)body);
      if (cd.equals("fmrisc") && name.equals("array"))
        return Construct.arrayTerm(vars, (Expression)body);
      State.setError("Unknown quantifier: " + object.toString());
      return null;
    }
    
    /**************************************************************************
     * applications
     *************************************************************************/
    if (om instanceof OMApplication)
    {
      OMApplication object = (OMApplication)om;
      Vector args = object.getElements();
      int n = args.size();
      if (n == 0)
      {
        State.setError("empty application: " + object.toString());
        return null;
      }
      OMObject arg0 = (OMObject)args.firstElement();
      if (arg0 instanceof OMSymbol)
      {
        OMSymbol fun = (OMSymbol)arg0;
        String cd = fun.getCd();
        if (cd == null)
        {
          State.setError("empty CD: " + fun.toString());
          return null;
        }
        String name = fun.getName();
        if (name == null)
        {
          State.setError("empty name: " + fun.toString());
          return null;
        } 
        
        /**********************************************************************
         * tuple terms
         *********************************************************************/
        if (cd.equals("fmrisc") && name.equals("tuple"))
        {
          Expression[] exps = new Expression[n-1];
          for (int i=1; i<n; i++)
          {
            OMObject arg1 = (OMObject)args.elementAt(i);
            exps[i-1] = toExpressionCore(arg1);
          }
          return Construct.tupleTerm(exps);
        }
        
        /**********************************************************************
         * record terms
         *********************************************************************/
        if (cd.equals("fmrisc") && name.equals("record"))
        {
          ValuedIdentifier[] valident = new ValuedIdentifier[n-1];
          for (int i=1; i<n; i++)
          {
            OMObject arg1 = (OMObject)args.elementAt(i);
            valident[i-1] = toValuedIdentifier(arg1);
          }
          return Construct.recordTerm(valident);
        }
        
        /**********************************************************************
         * update terms
         *********************************************************************/
        if (cd.equals("fmrisc") && name.equals("update"))
        {
          OMObject arg1 = (OMObject)args.elementAt(1);
          Expression base = toExpressionCore(arg1);
          Selector[] selector = new Selector[n-3];
          for (int i=2; i<n-1; i++)
          {
            OMObject arg2 = (OMObject)args.elementAt(i);
            selector[i-2] = toSelector(arg2);
          }
          OMObject arg3 = (OMObject)args.elementAt(n-1);
          Expression update = toExpressionCore(arg3);
          return Construct.updateTerm(base, selector, update);
        }
        
        /**********************************************************************
         * let terms
         *********************************************************************/
        if (cd.equals("fmrisc") && name.equals("let"))
        {
          if (n == 1)
          {
            State.setError("malformed let expression: " + object.toString());
            return null;
          }
          ValueDeclaration[] vdecl = new ValueDeclaration[n-2];
          for (int i=1; i<n-1; i++)
            vdecl[i-1] = toValueDeclaration((OMObject)args.elementAt(i));
          Expression body = toExpressionCore((OMObject)args.elementAt(n-1));
          return Construct.letExpression(vdecl, body);
        }
        
        /**********************************************************************
         * special applications with one argument
         *********************************************************************/
        if (n == 2)
        {
          OMObject child0 = (OMObject)args.elementAt(1);
          Expression exp0 = toExpressionCore(child0);
          if (cd.equals("arith1") && name.equals("unary_minus"))
            return Construct.negationTerm(exp0);
          if (cd.equals("logic1") && name.equals("not"))
            return Construct.notFormula(exp0);
        }
        /**********************************************************************
         * special applications with two arguments
         *********************************************************************/
        else if (n == 3)
        {
          OMObject child0 = (OMObject)args.elementAt(1);
          OMObject child1 = (OMObject)args.elementAt(2);
          
          Expression exp0 = toExpressionCore(child0);
          if (cd.equals("fmrisc") && name.equals("select"))
          {
            Selector selector = toSelector(child1);
            if (selector == null) return null;
            return Construct.selectionTerm(exp0, selector);
          }
          
          Expression exp1 = toExpressionCore(child1);
          if (cd.equals("logic1") && name.equals("and"))
            return Construct.andFormula(exp0, exp1);
          if (cd.equals("arith1") && name.equals("divide"))
            return Construct.dividesTerm(exp0, exp1);
          if (cd.equals("logic1") && name.equals("equivalent"))
            return Construct.equivalentFormula(exp0, exp1);
          if (cd.equals("relation1") && name.equals("eq"))
            return Construct.equalsFormula(exp0, exp1);  
          if (cd.equals("relation1") && name.equals("geq"))
            return Construct.greaterEqualFormula(exp0, exp1);  
          if (cd.equals("relation1") && name.equals("gt"))
            return Construct.greaterFormula(exp0, exp1);
          if (cd.equals("logic1") && name.equals("implies"))
            return Construct.impliesFormula(exp0, exp1);
          if (cd.equals("relation1") && name.equals("leq"))
            return Construct.lessEqualFormula(exp0, exp1);
          if (cd.equals("relation1") && name.equals("lt"))
            return Construct.lessFormula(exp0, exp1);
          if (cd.equals("arith1") && name.equals("minus"))
            return Construct.minusTerm(exp0, exp1);
          if (cd.equals("logic1") && name.equals("xor"))
            return Construct.notEquivalentFormula(exp0, exp1);
          if (cd.equals("relation1") && name.equals("neq"))
            return Construct.notEqualsFormula(exp0, exp1);
          if (cd.equals("logic1") && name.equals("or"))
            return Construct.orFormula(exp0, exp1);
          if (cd.equals("arith1") && name.equals("plus"))
            return Construct.plusTerm(exp0, exp1);          
          if (cd.equals("arith1") && name.equals("power"))
              return Construct.powerTerm(exp0, exp1);
          if (cd.equals("arith2") && name.equals("times"))
            return Construct.timesTerm(exp0, exp1);
        }
        /**********************************************************************
         * special applications with one argument
         *********************************************************************/
        else if (n == 4)
        {
          Expression exp0 = toExpressionCore((OMObject)args.elementAt(1));
          Expression exp1 = toExpressionCore((OMObject)args.elementAt(2));
          Expression exp2 = toExpressionCore((OMObject)args.elementAt(3));
          if (cd.equals("fmrisc") && name.equals("if"))
            return Construct.ifThenElseExpression(exp0, exp1, exp2);
        }
        State.setError("Unknown function symbol: " + object.toString());
        return null;
      }
      /**********************************************************************
       * general applications
       *********************************************************************/
      Expression fexp = toExpressionCore(arg0);
      Expression[] exps = new Expression[n-1];
      for (int i=1; i<n; i++)
      {
        OMObject arg1 = (OMObject)args.elementAt(i);
        exps[i-1] = toExpressionCore(arg1);
      }
      return Construct.applicationExpression((Expression)fexp, exps);
    }
    
    State.setError("Unknown object: " + om.toString());
    return null;
  }
 
  /***************************************************************************
   * Convert OpenMath type to AST.
   * @param om the object in AST representation.
   * @return the OpenMath representation of the object.
   **************************************************************************/
  private static Type toTypeCore(OMObject om)
  {
    if (om instanceof OMSymbol)
    {
      OMSymbol object = (OMSymbol)om;
      String cd = object.getCd();
      if (cd == null)
      {
        State.setError("empty CD: " + object.toString());
        return null;
      }
      String name = object.getName();
      if (name == null)
      {
        State.setError("empty name: " + object.toString());
        return null;
      } 
      if (cd.equals("setname2") && name.equals("Boolean"))
        return Construct.atomicType("BOOLEAN");
      if (cd.equals("setname1") && name.equals("N"))
        return Construct.atomicType("NAT");
      if (cd.equals("setname1") && name.equals("Z"))
        return Construct.atomicType("INT");
      if (cd.equals("setname1") && name.equals("R"))
        return Construct.atomicType("REAL");
    }
    else if (om instanceof OMVariable)
    {
      OMVariable object = (OMVariable)om;
      return Construct.reference(object.getName());
    }
    else if (om instanceof OMApplication)
    {
      OMApplication object = (OMApplication)om;
      Vector args = object.getElements();
      int n = args.size();
      if (n == 0)
      {
        State.setError("Empty application: " + object.toString());
        return null;
      }
      OMObject arg0 = (OMObject)args.firstElement();
      if (!(arg0 instanceof OMSymbol))
      {
        State.setError("Symbol expected: " + arg0.toString());
        return null;
      }
      OMSymbol fun = (OMSymbol)arg0;
      String cd = fun.getCd();
      if (cd == null)
      {
        State.setError("empty CD: " + fun.toString());
        return null;
      }
      String name = fun.getName();
      if (name == null)
      {
        State.setError("empty name: " + fun.toString());
        return null;
      } 
      if (cd.equals("fmrisc") && name.equals("arraytype"))
      {
        if (n != 3)
        {
          State.setError("3 arguments expected: " + object.toString());
          return null;
        }
        return Construct.arrayType(toTypeCore((OMObject)args.elementAt(1)),
          toTypeCore((OMObject)args.elementAt(2)));
      }
      if (cd.equals("sts") && name.equals("mapsto"))
      {
        if (n != 3)
        {
          State.setError("3 arguments expected: " + object.toString());
          return null;
        }
        OMObject om0 = (OMObject)args.elementAt(1);
        if (!(om0 instanceof OMApplication))
        {
          State.setError("application expected: " + om0.toString());
          return null;
        }
        OMApplication om1 = (OMApplication)om0;
        Vector children = om1.getElements();
        int l = children.size();
        if (l == 0)
        {
          State.setError("children expected: " + om1.toString());
          return null;
        }
        Type[] domain = new Type[l-1];
        for (int i=0; i<l-1; i++)
          domain[i] = toTypeCore((OMObject)children.elementAt(i+1));
        return Construct.functionType(domain,
          toTypeCore((OMObject)args.elementAt(2)));
      }
      if (cd.equals("fmrisc") && name.equals("recordtype"))
      {
        TypedIdentifier[] base = new TypedIdentifier[n-1];
        for (int i=1; i<n; i++)
          base[i-1] = toTypedIdentifier((OMObject)args.elementAt(i));
        return Construct.recordType(base);
      }
      if (cd.equals("interval1") && name.equals("integer_interval"))
      {
        if (n != 3)
        {
          State.setError("3 arguments expected: " + object.toString());
          return null;
        }
        OMObject arg1 = (OMObject)args.elementAt(1);
        if (!(arg1 instanceof OMInteger))
        {
          State.setError("Integer expected: " + arg1);
          return null;
        }
        OMObject arg2 = (OMObject)args.elementAt(2);
        if (!(arg2 instanceof OMInteger))
        {
          State.setError("Integer expected: " + arg2);
          return null;
        }
        OMInteger i1 = (OMInteger)arg1;
        OMInteger i2 = (OMInteger)arg2;
        return Construct.subrangeType(Construct.number(i1.getInteger()),
          Construct.number(i2.getInteger()));
      }
      if (cd.equals("fmrisc") && name.equals("subtype"))
      {
        if (n != 2)
        {
          State.setError("2 arguments expected: " + object.toString());
          return null;
        }
        Expression exp = toExpressionCore((OMObject)args.elementAt(1));
        return Construct.subType(exp);
      }
      if (cd.equals("set1") && name.equals("cartesian_product"))
      {
        Type[] base = new Type[n-1];
        for (int i=1; i<n; i++)
          base[i-1] = toTypeCore((OMObject)args.elementAt(i));
        return Construct.tupleType(base);
      }
    }
    State.setError("unknown type: " + om.toString());
    return null;
  }
  
  /***************************************************************************
   * Convert OpenMath expression to AST.
   * @param om the object in AST representation.
   * @return the OpenMath representation of the object.
   **************************************************************************/
  private static Expression toExpressionCore(OMObject om)
  {
    AST ast = toAST(om);
    if (!(ast instanceof Expression))
    {
      State.setError("expression expected: " + om.toString());
      return null;
    }
    return (Expression)(ast);
  }
  
  /***************************************************************************
   * Convert OpenMath expression to identifier..
   * @param om the identifier in AST representation.
   * @return the OpenMath representation of the identifier.
   **************************************************************************/
  public static Identifier toIdentifier(OMObject om)
  {
    AST ast = toAST(om);
    if (!(ast instanceof Identifier))
    {
      State.setError("identifier expected: " + om.toString());
      return null;
    }
    return (Identifier)(ast);
  }
  
  /***************************************************************************
   * Convert OpenMath type to AST.
   * @param om the object in AST representation.
   * @return the OpenMath representation of the object.
   **************************************************************************/
  private static TypedIdentifier toTypedIdentifier(OMObject om)
  {
    AST ast = toAST(om);
    if (!(ast instanceof TypedIdentifier))
    {
      State.setError("typed identifier expected: " + om.toString());
      return null;
    }
    return (TypedIdentifier)(ast);
  }
  
  /****************************************************************************
   * Convert value declaration to OpenMath representation.
   * @param decl a value declaration.
   * @return the corresponding OpenMath representation.
   ***************************************************************************/
  private static OMObject toOpenMathValueDecl(ValueDeclaration decl)
  {
    OMApplication om = new OMApplication();
    om.addElement(new OMSymbol("fmrisc", "ldecl"));
    Type type = decl.getType();
    if (type == null)
      om.addElement(toOpenMath(decl.getIdentifier()));
    else
    {
      OMAttribution om0 = new OMAttribution();
      om0.setConstructor(toOpenMath(decl.getIdentifier()));
      om0.put(new OMSymbol("sts", "type"), toOpenMath(decl.getType()));
      om.addElement(om0);
    }
    Expression value = decl.getValue();
    if (value != null) om.addElement(toOpenMath(value));
    return om;
  }
  
  /****************************************************************************
   * Convert value declaration from OpenMath representation to AST.
   * @param om a value declaration in OpenMath representation.
   * @return the corresponding AST representation (null if error)
   ***************************************************************************/
  private static ValueDeclaration toValueDeclaration(OMObject om)
  {
    if (!(om instanceof OMApplication))
    {
      State.setError("application expected: " + om.toString());
      return null;
    }
    OMApplication object = (OMApplication)om;
    Vector v = object.getElements();
    int n = v.size();
    if (n != 3)
    {
      State.setError("malformed local declaration: " + om.toString());
      return null;
    }
    
    OMObject oms = (OMObject)v.elementAt(0);
    if (!(oms instanceof OMSymbol))
    {
      State.setError("symbol expected: " + oms.toString());
      return null;
    }
    OMSymbol oms0 = (OMSymbol)oms;
    if (!oms0.getCd().equals("fmrisc") || !oms0.getName().equals("ldecl"))
    {
      State.setError("malformed local declaration: " + om.toString());
      return null;
    }
    
    OMObject om0 = (OMObject)v.elementAt(1);
    ValueDeclIdentifier ident = null;
    Type type = null;
    if (om0 instanceof OMVariable)
      ident = Construct.valueDeclIdentifier(((OMVariable)om0).getName());
    else if (om0 instanceof OMAttribution)
    {
      OMAttribution oma = (OMAttribution)om0;
      OMObject om1 = oma.getConstructor();
      Enumeration e = oma.getValues();
      if (!(om1 instanceof OMVariable) || !e.hasMoreElements())
      {
        State.setError("malformed attribution: " + oma.toString());
        return null;
      }
      ident = Construct.valueDeclIdentifier(((OMVariable)om1).getName());
      type = toTypeCore((OMObject)e.nextElement());
    }
    else
    {
      State.setError("Variable or attribution expected: " + om0.toString());
      return null;
    }
    OMObject om1 = (OMObject)v.elementAt(2);
    AST body = toAST(om1);
    if (!(body instanceof Expression))
    {
      State.setError("malformed expression: " + om1.toString());
      return null;
    }
    
    return Construct.valueDeclaration(ident, type, (Expression)body);
  }
  
  /****************************************************************************
   * Convert valued identifier from OpenMath representation to AST.
   * @param vident a valued identifier
   * @return the corresponding AST representation (null if error)
   ***************************************************************************/
  private static ValuedIdentifier toValuedIdentifier(OMObject om)
  {
    if (!(om instanceof OMApplication))
    {
      State.setError("application expected: " + om.toString());
      return null;
    }
    OMApplication object = (OMApplication)om;
    Vector args = object.getElements();
    int n = args.size();
    if (n != 3)
    {
      State.setError("malformed selector: " + object.toString());
      return null;
    }
    OMObject arg0 = (OMObject)args.elementAt(0);
    if (!(arg0 instanceof OMSymbol))
    {
      State.setError("symbol expected: " + arg0.toString());
      return null;
    }
    OMSymbol symbol = (OMSymbol)arg0;
    String cd = symbol.getCd();
    String name = symbol.getName();
    if (!cd.equals("relation1") || !name.equals("eq"))
    {
      State.setError("malformed valued identifier: " + om.toString());
      return null;
    }
    OMObject arg1 = (OMObject)args.elementAt(1);
    if (!(arg1 instanceof OMVariable))
    {
      State.setError("variable expected: " + arg1.toString());
      return null;
    }
    OMVariable var = (OMVariable)arg1;
    Reference ident = Construct.reference(var.getName());
    OMObject arg2 = (OMObject)args.elementAt(2);
    Expression value = toExpressionCore(arg2);
    return Construct.valuedIdentifier(ident, value);
  }
  
  /****************************************************************************
   * Convert selector from OpenMath representation to AST.
   * @param decl a value declaration in OpenMath representation.
   * @return the corresponding AST representation (null if error)
   ***************************************************************************/
  private static Selector toSelector(OMObject om)
  {
    if (!(om instanceof OMApplication))
    {
      State.setError("application expected: " + om.toString());
      return null;
    }
    OMApplication object = (OMApplication)om;
    Vector args = object.getElements();
    int n = args.size();
    if (n != 2)
    {
      State.setError("malformed selector: " + object.toString());
      return null;
    }
    OMObject arg0 = (OMObject)args.firstElement();
    if (!(arg0 instanceof OMSymbol))
    {
      State.setError("malformed selector: " + object.toString());
      return null;
    }  
    OMSymbol fun = (OMSymbol)arg0;
    String cd = fun.getCd();
    if (cd == null)
    {
      State.setError("empty CD: " + fun.toString());
      return null;
    }
    String name = fun.getName();
    if (name == null)
    {
      State.setError("empty name: " + fun.toString());
      return null;
    } 
    OMObject child0 = (OMObject)args.elementAt(1);
    if (cd.equals("fmrisc") && name.equals("selident"))
    {
      if (!(child0 instanceof OMVariable))
      {
        State.setError("variable expected: " + child0.toString());
        return null;
      }
      OMVariable omv = (OMVariable)child0;
      return Construct.selectorIdentifier(
        Construct.reference(omv.getName()));
    }
    if (cd.equals("fmrisc") && name.equals("selindex"))
    {
      Expression exp0 = toExpressionCore(child0);
      return Construct.selectorIndex(exp0);
    }
    if (cd.equals("fmrisc") && name.equals("selnumber"))
    {
      if (!(child0 instanceof OMInteger))
      {
        State.setError("integer expected: " + child0.toString());
        return null;
      }
      OMInteger omi = (OMInteger)child0;
      return Construct.selectorNumber(
        Construct.number(omi.getInteger()));
    }
    State.setError("unknown selector: " + om.toString());
    return null;
  }
}
