// ----------------------------------------------------------------------------
// ProofUtil.java
// Proving utilities.
// $Id: ProofUtil.java,v 1.10 2012/10/15 15:16:47 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Proving;

import java.util.Vector;

import fmrisc.External.GeneralHashFunctionLibrary;
import fmrisc.ProofNavigator.Syntax.*;

/**
 * 
 * Proving utilities.
 *
 */
public final class ProofUtil
{  
  /***************************************************************************
   * converts vector v of Formula objects to Formula array
   * @param v a vector holding Formula objects
   * @return an array holding these objects
   **************************************************************************/
  public static Formula[] toFormulaArray(Vector<Formula> v)
  {
    return (Formula[])v.toArray(new Formula[v.size()]);
  }
  
  /***************************************************************************
   * converts vector v of ProofState objects to ProofState array
   * @param v a vector holding ProofState objects
   * @return an array holding these objects
   **************************************************************************/
  public static ProofState[] toProofStateArray(Vector<ProofState> v)
  {
    return (ProofState[])v.toArray(new ProofState[v.size()]);
  }
  
  /***************************************************************************
   * hash string representation of an object to a short word
   * @param rep the string representation of the object
   * @param len the length of the result word
   * @return a hashed version of rep of length len consisting of letters and
   * decimal digits.
   **************************************************************************/
  public static String hash(String rep, int len)
  {
    StringBuffer buffer = new StringBuffer(len);
    long code = GeneralHashFunctionLibrary.RSHash(rep);
    for (int i=0; i<len; i++)
    {
      long c = code%32;
      if (c < 26)
        buffer.append((char)('a'+c));
      else
        buffer.append((char)('1'+c-26));
      code = code/32;
    }
    return buffer.toString();
  }
  
  /***************************************************************************
   * return negated version of formula where negation is drawn innermost
   * (to atomic formulas)
   * @param formula a formula
   * @return its negated and simplified version
   *************************************************************************/
  public static Expression negate(Expression formula)
  {
    if (formula instanceof Logical)
    {
      Logical f = (Logical)formula;
      return Construct.logical(!f.getValue());
    }
    if (formula instanceof LessFormula)
    {
      LessFormula f = (LessFormula)formula;
      return Construct.greaterEqualFormula(f.getFirst(), f.getSecond());
    }
    if (formula instanceof LessEqualFormula)
    {
      LessEqualFormula f = (LessEqualFormula)formula;
      return Construct.greaterFormula(f.getFirst(), f.getSecond());
    }
    if (formula instanceof GreaterFormula)
    {
      GreaterFormula f = (GreaterFormula)formula;
      return Construct.lessEqualFormula(f.getFirst(), f.getSecond());
    }
    if (formula instanceof GreaterEqualFormula)
    {
      GreaterEqualFormula f = (GreaterEqualFormula)formula;
      return Construct.lessFormula(f.getFirst(), f.getSecond());
    }
    if (formula instanceof NotFormula)
    {
      NotFormula f = (NotFormula)formula;
      return f.getBase();
    }
    if (formula instanceof AndFormula)
    {
      AndFormula f = (AndFormula)formula;
      Expression f1 = f.getFirst();
      Expression f2 = f.getSecond();
      return Construct.orFormula(negate(f1), negate(f2));
    }
    if (formula instanceof OrFormula)
    {
      OrFormula f = (OrFormula)formula;
      Expression f1 = f.getFirst();
      Expression f2 = f.getSecond();
      return Construct.andFormula(negate(f1), negate(f2));
    }
    if (formula instanceof EquivalentFormula)
    {
      EquivalentFormula f = (EquivalentFormula)formula;
      Expression f1 = f.getFirst();
      Expression f2 = f.getSecond();
      return Construct.notEquivalentFormula(f1, f2);
    }
    if (formula instanceof NotEquivalentFormula)
    {
      NotEquivalentFormula f = (NotEquivalentFormula)formula;
      Expression f1 = f.getFirst();
      Expression f2 = f.getSecond();
      return Construct.equivalentFormula(f1, f2);
    }
    if (formula instanceof ImpliesFormula)
    {
      ImpliesFormula f = (ImpliesFormula)formula;
      Expression f1 = f.getFirst();
      Expression f2 = f.getSecond();
      return Construct.andFormula(f1, negate(f2));
    }
    if (formula instanceof EqualsFormula)
    {
      EqualsFormula f = (EqualsFormula)formula;
      Expression f1 = f.getFirst();
      Expression f2 = f.getSecond();
      return Construct.notEqualsFormula(f1, f2);
    }
    if (formula instanceof NotEqualsFormula)
    {
      NotEqualsFormula f = (NotEqualsFormula)formula;
      Expression f1 = f.getFirst();
      Expression f2 = f.getSecond();
      return Construct.equalsFormula(f1, f2);
    }
    if (formula instanceof LetExpression)
    {
      LetExpression f = (LetExpression)formula;
      ValueDeclaration[] d = f.getDeclarations();
      Expression b = f.getBase();
      return Construct.letExpression(d, negate(b)).instantiate();
    }
    if (formula instanceof ForallFormula)
    {
      ForallFormula f = (ForallFormula)formula;
      TypedIdentifier[] vars = f.getVariables();
      Expression b = f.getBase();
      return Construct.existsFormula(vars, negate(b)).instantiate();
    }
    if (formula instanceof ExistsFormula)
    {
      ExistsFormula f = (ExistsFormula)formula;
      TypedIdentifier[] vars = f.getVariables();
      Expression b = f.getBase();
      return Construct.forallFormula(vars, negate(b)).instantiate();
    }
    return Construct.notFormula(formula);
  }
  
  /***************************************************************************
   * Construct expression that binds ident to value in body.
   * @param ident an identifier potentially referenced in body.
   * @param value the value to bind the identifier to.
   * @param body the body expression.
   * @return a copy of body with the appropriate binding.
   **************************************************************************/
  public static Expression bind(ValueDeclIdentifier ident,
    Expression value, Expression body)
  {
    return (Expression)Substitute.substitute(body, 
      new ValueDeclIdentifier[] { ident }, new Expression[] { value },
      false);
  }
}
