// ----------------------------------------------------------------------------
// ASTVisitorBase.java
// Base class of visitors of abstract syntax trees
// $Id: ASTVisitorBase.java,v 1.6 2012/10/16 15:49:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2006, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;

/**
 * 
 * Base class of visitors of abstract syntax trees.
 *
 */
public class ASTVisitorBase implements ASTVisitor
{
  /****************************************************************************
   * Signal whether tree is to be cloned.
   * @return false, signalling that the tree is not to be cloned.
   ***************************************************************************/
  public boolean isCloned()
  {
    return false;
  }
  
  /****************************************************************************
   * Visit an abstract syntax tree.
   * @param tree the abstract syntax tree that is visited.
   * @return the result of the visit.
   ***************************************************************************/
  public AST visit(AST tree)
  {
    if (tree instanceof Identifier) return visit((Identifier)tree);
    if (tree instanceof Declaration) return visit((Declaration)tree);
    if (tree instanceof Expression) return visit((Expression)tree);
    if (tree instanceof Selector) return visit((Selector)tree);
    if (tree instanceof Type) return visit((Type)tree);
    if (tree instanceof TypedIdentifier) return visit((TypedIdentifier)tree);
    if (tree instanceof ValuedIdentifier) return visit((ValuedIdentifier)tree);
    System.out.println("Unknown syntax tree visited: " + tree.toString());
    System.exit(-1);
    return null;
  }
  
  /****************************************************************************
   * 
   * Visit identifiers.
   * 
   ***************************************************************************/
  protected AST visit(Identifier tree)
  {
    if (tree instanceof Reference) return visit((Reference)tree);
    if (tree instanceof TypeDeclIdentifier) return visit((TypeDeclIdentifier)tree);
    if (tree instanceof ValueDeclIdentifier) return visit((ValueDeclIdentifier)tree);
    if (tree instanceof FormulaDeclIdentifier) return visit((FormulaDeclIdentifier)tree);
    System.out.println("Unknown identifier visited: " + tree.toString());
    System.exit(-1);
    return null;
  }
  
  protected AST visit(Reference tree) { return null; }
  protected AST visit(TypeDeclIdentifier tree) { return null; }
  protected AST visit(ValueDeclIdentifier tree) { return null; }
  protected AST visit(FormulaDeclIdentifier tree) { return null; }
  
  /****************************************************************************
   * 
   * Visit declarations.
   * 
   ***************************************************************************/
  protected AST visit(Declaration tree) 
  { 
    if (tree instanceof AxiomDeclaration) return visit((AxiomDeclaration)tree);
    if (tree instanceof FormulaDeclaration) return visit((FormulaDeclaration)tree);
    if (tree instanceof TypeDeclaration) return visit((TypeDeclaration)tree);
    if (tree instanceof ValueDeclaration) return visit((ValueDeclaration)tree); 
    System.out.println("Unknown declaration visited: " + tree.toString());
    System.exit(-1);
    return null;
  }
 
  protected AST visit(AxiomDeclaration tree) { return null; }
  protected AST visit(FormulaDeclaration tree) { return null; }
  protected AST visit(TypeDeclaration tree) { return null; }
  protected AST visit(ValueDeclaration tree) { return null; }
  
  /****************************************************************************
   * 
   * Visit expressions.
   * 
   ***************************************************************************/
  protected AST visit(Expression tree) 
  { 
    if (tree instanceof ApplicationExpression) return visit((ApplicationExpression)tree);
    if (tree instanceof BinaryExpression) return visit((BinaryExpression)tree);
    if (tree instanceof IfThenElseExpression) return visit((IfThenElseExpression)tree);
    if (tree instanceof LetExpression) return visit((LetExpression)tree);
    if (tree instanceof Logical) return visit((Logical)tree); 
    if (tree instanceof BitLogical) return visit((BitLogical)tree); 
    if (tree instanceof Number) return visit((Number)tree); 
    if (tree instanceof QuantifiedExpression) return visit((QuantifiedExpression)tree); 
    if (tree instanceof RecordTerm) return visit((RecordTerm)tree); 
    if (tree instanceof SelectionTerm) return visit((SelectionTerm)tree); 
    if (tree instanceof TupleTerm) return visit((TupleTerm)tree); 
    if (tree instanceof UnaryExpression) return visit((UnaryExpression)tree); 
    if (tree instanceof UpdateTerm) return visit((UpdateTerm)tree); 
    System.out.println("Unknown expression visited: " + tree.toString());
    System.exit(-1);
    return null;
  }
  
  protected AST visit(BinaryExpression tree) 
  { 
    if (tree instanceof AndFormula) return visit((AndFormula)tree); 
    if (tree instanceof DividesTerm) return visit((DividesTerm)tree); 
    if (tree instanceof EqualsFormula) return visit((EqualsFormula)tree); 
    if (tree instanceof EquivalentFormula) return visit((EquivalentFormula)tree); 
    if (tree instanceof GreaterEqualFormula) return visit((GreaterEqualFormula)tree); 
    if (tree instanceof GreaterFormula) return visit((GreaterFormula)tree); 
    if (tree instanceof ImpliesFormula) return visit((ImpliesFormula)tree); 
    if (tree instanceof LessEqualFormula) return visit((LessEqualFormula)tree); 
    if (tree instanceof LessFormula) return visit((LessFormula)tree); 
    if (tree instanceof MinusTerm) return visit((MinusTerm)tree); 
    if (tree instanceof NotEqualsFormula) return visit((NotEqualsFormula)tree); 
    if (tree instanceof NotEquivalentFormula) return visit((NotEquivalentFormula)tree); 
    if (tree instanceof OrFormula) return visit((OrFormula)tree); 
    if (tree instanceof PlusTerm) return visit((PlusTerm)tree); 
    if (tree instanceof BitOrTerm) return visit((BitOrTerm)tree); 
    if (tree instanceof TimesTerm) return visit((TimesTerm)tree); 
    if (tree instanceof BitAndTerm) return visit((BitAndTerm)tree); 
    if (tree instanceof PowerTerm) return visit((PowerTerm)tree); 
    System.out.println("Unknown binary expression visited: " + tree.toString());
    System.exit(-1);
    return null;
  }
 
  protected AST visit(UnaryExpression tree) 
  {  
    if (tree instanceof NegationTerm) return visit((NegationTerm)tree); 
    if (tree instanceof NotFormula) return visit((NotFormula)tree);
    if (tree instanceof BitNotTerm) return visit((BitNotTerm)tree); 
    System.out.println("Unknown unary expression visited: " + tree.toString());
    System.exit(-1);
    return null;
  }
  
  protected AST visit(QuantifiedExpression tree) 
  { 
    if (tree instanceof ArrayTerm) return visit((ArrayTerm)tree); 
    if (tree instanceof ExistsFormula) return visit((ExistsFormula)tree); 
    if (tree instanceof ForallFormula) return visit((ForallFormula)tree); 
    if (tree instanceof LambdaTerm) return visit((LambdaTerm)tree); 
    System.out.println("Unknown quantified expression visited: " + tree.toString());
    System.exit(-1);
    return null;
  }
  
  protected AST visit(AndFormula tree) { return null; }
  protected AST visit(DividesTerm tree) { return null; }
  protected AST visit(EqualsFormula tree) { return null; }
  protected AST visit(EquivalentFormula tree) { return null; }
  protected AST visit(GreaterEqualFormula tree) { return null; }
  protected AST visit(GreaterFormula tree) { return null; }
  protected AST visit(ImpliesFormula tree) { return null; }
  protected AST visit(LessEqualFormula tree) { return null; }
  protected AST visit(LessFormula tree) { return null; }
  protected AST visit(MinusTerm tree) { return null; }
  protected AST visit(NotEqualsFormula tree) { return null; }
  protected AST visit(NotEquivalentFormula tree) { return null; }
  protected AST visit(OrFormula tree) { return null; }
  protected AST visit(PlusTerm tree) { return null; }
  protected AST visit(BitOrTerm tree) { return null; }
  protected AST visit(PowerTerm tree) { return null; }
  protected AST visit(TimesTerm tree) { return null; }
  protected AST visit(BitAndTerm tree) { return null; }
  
  protected AST visit(NegationTerm tree) { return null; }
  protected AST visit(NotFormula tree) { return null; }
  protected AST visit(BitNotTerm tree) { return null; }
  
  protected AST visit(ArrayTerm tree) { return null; }
  protected AST visit(ExistsFormula tree) { return null; }
  protected AST visit(ForallFormula tree) { return null; }
  protected AST visit(LambdaTerm tree) { return null; }
  
  protected AST visit(ApplicationExpression tree) { return null; }
  protected AST visit(IfThenElseExpression tree) { return null; }
  protected AST visit(LetExpression tree) { return null; }
  protected AST visit(Logical tree) { return null; }
  protected AST visit(BitLogical tree) { return null; }
  protected AST visit(Number tree) { return null; }
  protected AST visit(RecordTerm tree) { return null; }
  protected AST visit(SelectionTerm tree) { return null; }
  protected AST visit(TupleTerm tree) { return null; }
  protected AST visit(UpdateTerm tree) { return null; }
  
  /****************************************************************************
   * 
   * Visit selectors.
   * 
   ***************************************************************************/
  protected AST visit(Selector tree) 
  { 
    if (tree instanceof SelectorIdentifier) return visit((SelectorIdentifier)tree);
    if (tree instanceof SelectorIndex) return visit((SelectorIndex)tree);
    if (tree instanceof SelectorNumber) return visit((SelectorNumber)tree);
    System.out.println("Unknown selector visited: " + tree.toString());
    System.exit(-1);
    return null;
  }
  
  protected AST visit(SelectorIdentifier tree) { return null; }
  protected AST visit(SelectorIndex tree) { return null; }
  protected AST visit(SelectorNumber tree) { return null; }
  
  /****************************************************************************
   * 
   * Visit types.
   * 
   ***************************************************************************/
  protected AST visit(Type tree) 
  { 
    if (tree instanceof ArrayType) return visit((ArrayType)tree);
    if (tree instanceof AtomicType) return visit((AtomicType)tree);
    if (tree instanceof FunctionType) return visit((FunctionType)tree);
    if (tree instanceof LetType) return visit((LetType)tree);
    if (tree instanceof RecordType) return visit((RecordType)tree);
    if (tree instanceof SubrangeType) return visit((SubrangeType)tree);
    if (tree instanceof SubType) return visit((SubType)tree);
    if (tree instanceof TupleType) return visit((TupleType)tree);
    System.out.println("Unknown type visited: " + tree.toString());
    System.exit(-1);
    return null;
  }
  
  protected AST visit(ArrayType tree) { return null; }
  protected AST visit(AtomicType tree) { return null; }
  protected AST visit(FunctionType tree) { return null; }
  protected AST visit(LetType tree) { return null; }
  protected AST visit(RecordType tree) { return null; }
  protected AST visit(SubrangeType tree) { return null; }
  protected AST visit(SubType tree) { return null; }
  protected AST visit(TupleType tree) { return null; }
  
  /****************************************************************************
   * 
   * Visit special nodes.
   * 
   ***************************************************************************/
  protected AST visit(TypedIdentifier tree) { return null; }
  protected AST visit(ValuedIdentifier tree) { return null; }
}
