// ----------------------------------------------------------------------------
// QuantifiedExpression.java
// Abstract base class of quantified expressions.
// $Id: QuantifiedExpression.java,v 1.23 2015/09/09 15:45:53 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;

import java.io.*;

import fmrisc.ProofNavigator.Semantics.*;

/**
 * Abstract base class of quantified expressions.
 */
public abstract class QuantifiedExpression extends ExpressionBase
{
  protected String quantifier;           // the name of the quantifier
  protected TypedIdentifier[] variables; // the bound variables
  protected Expression base;             // the base expression

  /****************************************************************************
   * constructs quantified expression from quantifier, variables, and base
   * @param quantifier the name of the quantifier used for printing
   * @param variables the variable sequence
   * @param base the base expression
   ***************************************************************************/
  public QuantifiedExpression(String quantifier, 
    TypedIdentifier[] variables, Expression base)
  {
    this.quantifier = quantifier;
    this.variables = variables;
    this.base = base;
  }
  
  /***************************************************************************
   * get binding level for printing (lower numbers mean greater binding power)
   * @return the priority level
   **************************************************************************/
  public int getPriority()
  {
    return 100;
  }
  
  /***************************************************************************
   * get quantified variables
   * @return the variable sequence
   *************************************************************************/
  public TypedIdentifier[] getVariables()
  {
    return variables;
  }
  
  /***************************************************************************
   * get base expression
   * @return the base expression
   **************************************************************************/
  public Expression getBase()
  {
    return base;
  }
  
  /** result of substitution **/
  protected final static class Substitution
  {
    TypedIdentifier[] vars; /* the new variables */
    Expression        base; /* the new base using these variables */
    Substitution(TypedIdentifier[] vars, Expression base)
    {
      this.vars = vars;
      this.base = base;
    }
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Substitution substitution()
  {
    TypedIdentifier[] vars = getVariables();
    int n = vars.length;
    TypedIdentifier[] vars0 = new TypedIdentifier[n];
    for (int i=0; i<n; i++)
    {
      vars0[i] = vars[i].instantiate();
      vars[i].setSubstitution(vars0[i]);
    }
    Expression base = getBase().instantiate();
    for (int i=0; i<n; i++)
      vars[i].resetSubstitution(); 
    return new Substitution(vars0, base);
  }
    
  /***************************************************************************
   * Prints quantifier/variables part of expression.
   * If in CVCL mode, returns collected constraint (null, otherwise).
   * @param out the stream on which the text is written
   * @return the constaint.
   **************************************************************************/
  protected Expression printHead(PrintWriter out)
  {
    out.print(quantifier);
    out.print('(');
    int n = variables.length;
    Expression constraint = null;
    for (int i=0; i<n; i++)
    {
      int v = ASTUtil.incVarNumber();
      ValueDeclIdentifier ident = variables[i].getIdentifier();
      ValueSymbol symbol = ident.getSymbol();
      if (symbol != null) symbol.setVarNumber(v);
      AST type0 = variables[i].getType();
      AST type1 = null;
      if (i+1 < n) type1 = variables[i+1].getType();
      if (type0 == type1)
        ident.print(out);
      else  
        variables[i].print(out);
      if (i < n-1)
        out.print(", ");
      else
        out.print("): "); 
    }
    return constraint;
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    int ov = ASTUtil.getVarNumber(); 
    printHead(out);
    base.printPriority(getPriority(), out);
    ASTUtil.setVarNumber(ov);
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Expression instantiate()
  {
    Substitution s = substitution();
    return construct(s.vars, s.base);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null).
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    int n = variables.length;
    TypedIdentifier[] variables0 = new TypedIdentifier[n];
    for (int i=0; i<n; i++)
    {
      variables0[i] = (TypedIdentifier)variables[i].accept(visitor);
      if (variables0[i] != null)
        variables[i].setSubstitution(variables0[i]);
    }
    Expression base0 = (Expression)base.accept(visitor);
    for (int i=0; i<n; i++)
      variables[i].resetSubstitution();
    if (!visitor.isCloned()) return null; 
    return construct(variables0, base0);
  }
  
  /***************************************************************************
   * Construct quantified expression.
   * @param variables typed variables.
   * @param base an expression.
   * @return a quantified expression of the same kind as this expression
   *         with the denoted variables and base expression.
   **************************************************************************/
  public abstract QuantifiedExpression construct(
    TypedIdentifier[] variables, Expression base);
}
