// ----------------------------------------------------------------------------
// ValueDeclIdentifier.java
// Handling of identifiers used in value declarations
// $Id: ValueDeclIdentifier.java,v 1.19 2015/09/09 15:45:53 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;

import fmrisc.ProofNavigator.Semantics.*;

import java.io.*;

/**
 * Handling of identifiers used in value declarations
 */
public final class ValueDeclIdentifier extends Identifier
{
  private String uniqueName;      // unique name
  private ValueSymbol symbol;  // symbol in value table linked to identifier
  private ValueDeclIdentifier substitution; // substitution for instantiation
  
  /****************************************************************************
   * construct identifier from name
   * @param name the text representation of the identifier
   ***************************************************************************/
  public ValueDeclIdentifier(String name)
  {
    super(name);
    uniqueName = ASTUtil.getUniqueName(name);
    symbol = new ValueSymbol(this);
  }

  /****************************************************************************
   * Replace name of identifier by unique name derived from table. 
   * This change is automatically propagated to all references 
   * to this identifier.
   * @param table the name table to be used for making the name unique
   ***************************************************************************/
  public void makeNameUnique(UniqueNameTable table)
  {
    setName(table.getUniqueName(getName()));
  }
  
  /****************************************************************************
   * returns unique name of identifier
   * @return the unique name of the identifier
   ***************************************************************************/
  public String getUniqueName()
  {
    return uniqueName;
  }
  
  /****************************************************************************
   * returns value symbol
   * @return the value symbol
   ***************************************************************************/
  public ValueSymbol getSymbol()
  {
    return symbol;
  }
  
  /****************************************************************************
   * returns substitution identifier for instantiation
   * @return the substitution identifier
   ***************************************************************************/
  public ValueDeclIdentifier getSubstitution()
  {
    return substitution;
  }
  
  /****************************************************************************
   * set substitution expression for instantiation
   * @param substitution the substitution expression
   ***************************************************************************/
  public void setSubstitution(ValueDeclIdentifier substitution)
  {
    this.substitution = substitution;
  }
  
  /***************************************************************************
   * reset substitution field
   **************************************************************************/
  public void resetSubstitution()
  {
    substitution = null;
  }
  
  /***************************************************************************
   * returns instantiated copy of identifier
   * @return the copy of the identifier
   **************************************************************************/
  public ValueDeclIdentifier instantiate()
  {
    ValueDeclIdentifier ident = Construct.valueDeclIdentifier(getName());
    ValueSymbol symbol0 = ident.getSymbol();
    symbol0.instantiate(symbol);
    TypeExpression te = symbol.getClearType();
    if (te != null) 
    {
      ValueDeclIdentifier osubstitution = substitution; 
      substitution = ident;
      te = te.instantiate();
      substitution = osubstitution;
    }
    symbol0.setClearType(te);
    return ident;
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    if (ASTUtil.getPrintVarNumber())
    {
      if (symbol.getDepth() > 0)
        out.print("_" + symbol.getVarNumber());
      else
        out.print(getName());
    }
    else if (ASTUtil.getPrintUnique())
      out.print(getUniqueName());
    else
      out.print(getName());
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * Overrides ASTBase.print such that parentheses are never printed
   * around the number.
   * @param out the stream on which the text is written
   **************************************************************************/
  public void print(PrintWriter out)
  {
    super.print(out);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null).
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    if (!visitor.isCloned()) return null;
    return this.instantiate();
  }
}
